/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Globals.h"

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <GL/freeglut_ext.h> // for glutMouseWheelFunc()
#endif

#include "RtAudio.h"

#include <iostream>
#include <vector>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <cstdio>
#include <cstdlib>
#include <sstream>

#ifdef __linux__
// For usleep
#include <unistd.h>
#endif

#include "MyRtAudio.h"
#include "AudioFile.h"
#include "AudioFileSet.h"
#include "Envelope.h"

#include "GrainCloud.h"
#include "SingleGrain.h"
#include "LoopSource.h"

#include "UISoundRect.h"
#include "UIScreen.h"
#include "UIStyle.h"
#include "UIStyleManager.h"
#include "UIParameter.h"
#include "UIPieMenu.h"
#include "UIWindow.h"
#include "UILFOCanvas.h"
#include "UIEnvelopeTypeCanvas.h"
#include "UIEffectsCanvas.h"

#include "App.h"
#include "AppControl.h"
#include "Parameter.h"
#include "Serializer.h"
#include "SamplesToSpectro.h"

#include "MyRtMidiIn.h"
#include "MidiQueue.h"
#include "MyRtMidiOut.h"

#include "EffectsBox.h"

using namespace std;

void idleFunc();
void displayFunc();
void reshapeFunc(int w, int h);
void specialFunc(int key, int x, int y);
void specialUpFunc(int key, int x, int y);
void keyboardFunc(unsigned char key, int x, int y);
void keyUpFunc(unsigned char key, int x, int y);
void mouseFunc(int button, int state, int x, int y);
void mouseDragFunc(int x, int y);
void mousePassiveMotionFunc(int x, int y);
void mouseWheelFunc(int button, int dir, int x, int y);

App::App()
{
    _mutex = new Mutex();
    
    _control = new AppControl(this);
    _screen = new UIScreen(this);
    
    _audio = NULL;
    
    _audioPath = "./loops/";
    
    _blockSize = BLOCK_SIZE;
    
    _sampTimeSec = (double)1.0/(double)SRATE;

    _position = Pt3d(0.0, 0.0, 0.0);

    _zoom = 1.0;
    
    _selectionMode = SOURCE;
    _dragMode = MOVE;

    _selectedSource = -1;
    _selectedRect = -1;
    _menuFlag = true;

    _paramValueString = "";

    _isEditingParamString = false;
    
    _isFullScreen = false;
    
    _xMoveSpeed = 10.0f;
    _yMoveSpeed = 10.0f;
    
    _audioBounce = false;
    _bounceFile = NULL;
    
    _fps = FPS;

    _displayHelp = false;

    _midiIn = new MyRtMidiIn(_mutex);
    _midiQueue = _midiIn->GetMidiQueue();

    _midiOut = new MyRtMidiOut();
}

App::~App()
{
    delete _mutex;

    if (_fileSet != NULL)
        delete _fileSet;
}

App _app;

int audioCallback(void *outputBuffer, void *inputBuffer,
                  unsigned int nFrames, double streamTime,
                  RtAudioStreamStatus status, void *userData);


void
App::Clean()
{
    try
    {
        if (_audio != NULL)
        {       
            _audio->StopStream();
            _audio->CloseStream();
        }
    }
    catch (RtAudioError &err)
    {
        err.printMessage();
    }
    
    if (_audio != NULL)
        delete _audio;

    if (_styleManager != NULL)
    {
        delete _styleManager;
        _styleManager = NULL;
        _style = NULL;
    }

    delete _pieMenu;

    delete _midiIn;
    delete _midiOut;

    delete _masterEffectsBox;
}

int
audioCallback(void *outputBuffer, void *inputBuffer,
              unsigned int nFrames, double streamTime,
              RtAudioStreamStatus status, void *userData)
{
    int res = _app.OnAudio(outputBuffer, inputBuffer,
                           nFrames, streamTime,
                           status, userData);

    return res;
}
    
int
App::OnAudio(void *outputBuffer, void *inputBuffer,
                     unsigned int nFrames, double streamTime,
                     RtAudioStreamStatus status, void *userData)
{
    _mutex->lock();
    
    SAMPLE *out = (SAMPLE *)outputBuffer;
    SAMPLE *in = (SAMPLE *)inputBuffer;
    
    memset(out, 0, sizeof(SAMPLE)*nFrames*CHANNELS);
    _tmpBuf.resize(nFrames*CHANNELS);
    
    if (!_menuFlag)
    {
        for (int i = 0; i < _sources.size(); i++)
        {
            memset(_tmpBuf.data(), 0, sizeof(SAMPLE)*nFrames*CHANNELS);
            
            // Mute the original note if we are playing midi notes
            if (_noteSources[i].empty())
                _sources[i]->NextBuffer(&_tmpBuf, nFrames);
            else
                // Even if not playing source, make it progress in time
                _sources[i]->NextBuffer(NULL, nFrames);

            for (int j = 0; j < _noteSources[i].size(); j++)
                _noteSources[i][j]->NextBuffer(&_tmpBuf, nFrames);

            // Process effects once for a give source and its derivative
            // played midi notes
            EffectsBox *effectsBox = _sources[i]->GetEffectsBox();
            effectsBox->Process(&_tmpBuf, nFrames, CHANNELS);

            for (int j = 0; j < _tmpBuf.size(); j++)
                out[j] += _tmpBuf[j];
        }
    }

    memcpy(_tmpBuf.data(), out, nFrames*CHANNELS*sizeof(SAMPLE));
           
    _masterEffectsBox->Process(&_tmpBuf, nFrames, CHANNELS);

    memcpy(out, _tmpBuf.data(), nFrames*CHANNELS*sizeof(SAMPLE));
    
    Time::Instance().AddSeconds(nFrames*_sampTimeSec);

    if (_bounceFile != NULL)
        fwrite(out, sizeof(SAMPLE), nFrames*CHANNELS, _bounceFile);

    _mutex->unlock();
    
    return 0;
}

void
App::ToggleDrawHelp(bool flag)
{
    _displayHelp = flag;
}

void
App::UpdateStyle()
{
    _style = _styleManager->GetCurrentStyle();

    double bgColor[4];
    _style->GetBackgroundColor(bgColor);
    glClearColor(bgColor[0], bgColor[1], bgColor[2], bgColor[3]);

    for (int i = 0; i < _soundRects.size(); i++)
        _soundRects[i]->GetUI()->UpdateStyle(_style);
    
    for (int i = 0; i < _sources.size(); i++)
        _sources[i]->GetUI()->UpdateStyle(_style);

    if (_grainLFOWindow != NULL)
        _grainLFOWindow->UpdateStyle(_style);
    if (_envelopeTypeWindow != NULL)
        _envelopeTypeWindow->UpdateStyle(_style);

    if (_effectsWindow != NULL)
        _effectsWindow->UpdateStyle(_style);
    
    glutPostRedisplay();
}

void
App::ToggleAudioBounce()
{
    if (!_audioBounce)
        _audioBounce = true;
    else
    {
        _audioBounce = false;

        memset(_bounceFileName, 0, 255);
    }

    if (_audioBounce)
    {
        long tm = time(NULL);
        
        int rnd = rand()%1000000;
        sprintf(_bounceFileName,
                "./bounces/bounce-%ld-%d.raw", tm, rnd);

        if (_bounceFile != NULL)
            fclose(_bounceFile);
        
        _bounceFile = fopen(_bounceFileName, "wb");
    }
    else
    {
        if (_bounceFile != NULL)
        {
            fclose(_bounceFile);
            _bounceFile = NULL;
        }
    }
}

void
App::WindowInit()
{    
    glutIdleFunc(idleFunc);
    glutReshapeFunc(reshapeFunc);
    glutDisplayFunc(displayFunc);
    glutMouseFunc(mouseFunc);
    glutMotionFunc(mouseDragFunc);
    glutPassiveMotionFunc(mousePassiveMotionFunc);
#ifndef __MACOSX_CORE__
    glutMouseWheelFunc(mouseWheelFunc);
#endif
    glutKeyboardFunc(keyboardFunc);
    glutKeyboardUpFunc(keyUpFunc);
    glutSpecialFunc(specialFunc);
    glutSpecialUpFunc(specialUpFunc);
}

void
App::SetWindowedMode()
{
    _winWidth = 0.6*_screenWidth;
    _winHeight = 0.6*_screenHeight;
    glutReshapeWindow(_winWidth,_winHeight);
    
    int centerXCorner = _screenWidth/2 - _winWidth/2;
    int centerYCorner = _screenHeight/2 - _winHeight/2;
    glutPositionWindow(centerXCorner, centerYCorner);
}

void
App::ToggleFullScreen()
{
    if (_isFullScreen)
    {
        SetWindowedMode();
        
        _isFullScreen = false;
    }
    else
    {    
        glutFullScreen();
        
        _isFullScreen = true;
    }
}

bool
App::Init(int midiPort)
{
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);    
    
    _screenWidth = glutGet(GLUT_SCREEN_WIDTH);
    _screenHeight = glutGet(GLUT_SCREEN_HEIGHT);

    _winWidth = 0.6*_screenWidth;
    _winHeight = 0.6*_screenHeight;

    _styleManager = new UIStyleManager();
    _style = _styleManager->GetCurrentStyle();

    char appName[512];
    _style->GetAppName(appName);
    
    glutInitWindowSize(_winWidth,_winHeight);
    glutInitWindowPosition (0, 0);
    glutCreateWindow(appName);

    // Do ingore key repeat!
    glutIgnoreKeyRepeat(1);
    
    SetWindowedMode();
    
    double bgColor[4];
    _style->GetBackgroundColor(bgColor);

    _pieMenu = new UIPieMenu(this, _screen);

    // Grain LFO window
    _grainLFOWindow = new UIWindow(_screen, _style,
                                   "LFO",
                                   0.01, 0.01, 0.25, 0.25);
    _grainLFOWindow->SetVisible(false);
    
    double cx0;
    double cy0;
    double cw0;
    double ch0;
    _grainLFOWindow->GetContentsRect(&cx0, &cy0, &cw0, &ch0);
    _grainLFOCanvas = new UILFOCanvas(_screen, _style, "",
                                      cx0, cy0, cw0, ch0);
    _grainLFOWindow->AddCanvas(_grainLFOCanvas);

    // Window type
    _envelopeTypeWindow = new UIWindow(_screen, _style,
                                     "Envelope",
                                     0.01, 0.01, 0.25, 0.25);
    _envelopeTypeWindow->SetVisible(false);
    
    double cx1;
    double cy1;
    double cw1;
    double ch1;
    _envelopeTypeWindow->GetContentsRect(&cx1, &cy1, &cw1, &ch1);
    _envelopeTypeCanvas = new UIEnvelopeTypeCanvas(this, _screen, _style, "",
                                                   cx1, cy1, cw1, ch1);
    _envelopeTypeWindow->AddCanvas(_envelopeTypeCanvas);

    // Effects window
    _effectsWindow = new UIWindow(_screen, _style,
                                  "Effects",
                                  0.1, 0.1, 0.8, 0.8);
    _effectsWindow->SetVisible(false);
    
    double cx2;
    double cy2;
    double cw2;
    double ch2;
    _effectsWindow->GetContentsRect(&cx2, &cy2, &cw2, &ch2);
    _effectsCanvas = new UIEffectsCanvas(_screen, _style, "",
                                         cx2, cy2, cw2, ch2);
    _effectsWindow->AddCanvas(_effectsCanvas);

    _masterEffectsBox = new EffectsBox();
    
    glClearColor(bgColor[0], bgColor[1], bgColor[2], bgColor[3]);
    
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glFrontFace(GL_CCW);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glPolygonMode(GL_FRONT_AND_BACK, GL_POLYGON_SMOOTH);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POINT_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
    glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    
    glutIdleFunc(idleFunc);
    glutReshapeFunc(reshapeFunc);
    glutDisplayFunc(displayFunc);
    glutMouseFunc(mouseFunc);
    glutMotionFunc(mouseDragFunc);
    glutPassiveMotionFunc(mousePassiveMotionFunc);
#ifndef __MACOSX_CORE__
    glutMouseWheelFunc(mouseWheelFunc);
#endif
    glutKeyboardFunc(keyboardFunc);
    glutKeyboardUpFunc(keyUpFunc);
    glutSpecialFunc(specialFunc);
    glutSpecialUpFunc(specialUpFunc);

    _fileSet = new AudioFileSet();
    if (_fileSet->Load(GetAudioPath()) == 1)
        return false;

    _fileSet->Stereoize();
    _fileSet->Normalize();
    _fileSet->Fade();
    
    _sounds = _fileSet->GetAudioFiles();
    cout << "Sounds loaded successfully..." << endl;
    
    _soundRects.resize(_sounds.size());
    for (int i = 0; i < _sounds.size(); i++)
    {
        SoundRect *rect = new SoundRect();
        
        UISoundRect *uiRect =
            new UISoundRect(rect, &_app, _style, _screen);
        rect->RegisterUI(uiRect);
        
        const char *fileName = _sounds[i]->GetName().c_str();
        uiRect->SetName(fileName);
            
        _soundRects[i] = rect;
        _soundRects[i]->AssociateSamples(*_sounds[i]->GetBuffer(),
                                         _sounds[i]->GetNumFrames(),
                                         _sounds[i]->GetNumChannels());

        for (int k = 0; k < _sounds[i]->GetNumChannels(); k++)
        {
            Spectrogram *spectro =
                SamplesToSpectro::SamplesToSpectrogram(*_sounds[i]->GetBuffer(),
                                                       _sounds[i]->GetNumFrames(),
                                                       _sounds[i]->GetNumChannels(), k);
            _soundRects[i]->AssociateSpectrogram(k, spectro);
        }

        for (int k = 0; k < _sounds[i]->GetNumChannels(); k++)
        {
            Spectrogram *chroma =
                SamplesToSpectro::SamplesToChromagram(*_sounds[i]->GetBuffer(),
                                                      _sounds[i]->GetNumFrames(),
                                                      _sounds[i]->GetNumChannels(), k);
            _soundRects[i]->AssociateChromagram(k, chroma);
        }
        
        _selectedRect = i;
        _pieMenu->InitFromParameters();
    }

    _selectedRect = -1;
    
    try
    {
        _audio = new MyRtAudio(1,
                               CHANNELS,
                               SRATE,
                               &_app._blockSize,
                               FORMAT,
                               true);
    }
    catch (RtAudioError & err)
    {
        err.printMessage();
        exit(1);
    }    
    try
    {
        _audio->OpenStream(&audioCallback);
        
        _blockSize = _audio->GetBufferSize();
        
        _audio->StartStream();
        
        _audio->ReportStreamLatency();        
        
    }
    catch(RtAudioError & err)
    {
        err.printMessage();
        return false;
    }

    if (midiPort == 0)
    {
        int port = _midiOut->Open(midiPort);
        _midiIn->Open(port);
    }
    else
    {
        _midiIn->Open(midiPort);
    }

    _virtualPitch = 0.0;
    
    UnSerializeSession();
        
    return true;
}

void
displayFunc()
{
    _app.OnDisplay();
}

void
App::OnDisplay()
{
    //_mutex->lock();
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glClearDepth(1.0);
    
    glPushMatrix();
    
    glLoadIdentity();
    
    glTranslatef(-_position._x*_winWidth,
                 -_position._y*_winHeight,
                 -_position._z);

    glScalef(_zoom, _zoom, 1.0);
    
    if (!_menuFlag)
    {
        for (int i = 0; i < _soundRects.size(); i++)
            _soundRects[i]->GetUI()->Draw();
        
        for (int i = 0; i < _sources.size(); i++)
            _sources[i]->GetUI()->Draw();

        if (_bounceFile != NULL)
            DrawBounceFile();
    }
    else
    {
        DrawTitle();
        DrawHelp(false);
    }

    if (_displayHelp && !_menuFlag)
        DrawHelp(true);
    
    //drawFps();

    if (!_menuFlag)
    {
        _grainLFOWindow->Draw();
        _envelopeTypeWindow->Draw();

        _effectsWindow->Draw();
        
        _pieMenu->Draw();
    }
    
    glPopMatrix();
    
    glFlush();
    glutSwapBuffers();

    //_mutex->unlock();
}

void
reshapeFunc(int w, int h)
{
    _app.OnReshape(w, h);
}

void
App::OnReshape(int w, int h)
{
    _mutex->lock();
    
    glViewport(0, 0, (GLsizei)w, (GLsizei)h);
    
    _winWidth = w;
    _winHeight = h;
        
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, (GLdouble)_winWidth,
            0.0, (GLdouble)_winHeight, -10.0, 10.0);
        
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    glTranslatef(-_position._x*_winWidth,
                 -_position._y*_winHeight,
                 -_position._z);
    
    glutPostRedisplay();

    _mutex->unlock();
}

void
idleFunc()
{
    _app.OnIdle();
}

void
App::UpdateLFOWindow()
{
    Source *source = GetCurrentSource();
    int paramIdx = GetCurrentSourceParamIdx();

    if (source->GetSourceType() == Source::GRAIN_CLOUD)
    {
        // Grain cloud
        
        // Grain LFO window
        if ((paramIdx == (int)GrainCloud::P_LFO_FREQ) ||
            (paramIdx == (int)GrainCloud::P_LFO_AMP))
        {
            _grainLFOWindow->SetVisible(true);
            
            double lfoFreq =
                ((GrainCloud *)source)->GetParameter(GrainCloud::P_LFO_FREQ)->GetValue();
            _grainLFOCanvas->SetFreq(lfoFreq);
        
            double lfoAmp =
                ((GrainCloud *)source)->GetParameter(GrainCloud::P_LFO_AMP)->GetValue();
            _grainLFOCanvas->SetAmp(lfoAmp);
        }
        else
        {
            _grainLFOWindow->SetVisible(false);
        }
    }
    else if (source->GetSourceType() == Source::SINGLE_GRAIN)
    {
        // Single grain
        
        // Grain LFO window
        if ((paramIdx == (int)SingleGrain::P_LFO_FREQ) ||
            (paramIdx == (int)SingleGrain::P_LFO_AMP))
        {
            _grainLFOWindow->SetVisible(true);
            
            double lfoFreq =
                ((SingleGrain *)source)->GetParameter(SingleGrain::P_LFO_FREQ)->GetValue();
            _grainLFOCanvas->SetFreq(lfoFreq);
        
            double lfoAmp =
                ((SingleGrain *)source)->GetParameter(SingleGrain::P_LFO_AMP)->GetValue();
            _grainLFOCanvas->SetAmp(lfoAmp);
        }
        else
        {
            _grainLFOWindow->SetVisible(false);
        }
    }
    else if (source->GetSourceType() == Source::LOOP_SOURCE)
    {
        // Loop source
        
        // Grain LFO window
        if ((paramIdx == (int)LoopSource::P_LFO_FREQ) ||
            (paramIdx == (int)LoopSource::P_LFO_AMP))
        {
            _grainLFOWindow->SetVisible(true);
            
            double lfoFreq =
                ((LoopSource *)source)->GetParameter(LoopSource::P_LFO_FREQ)->GetValue();
            _grainLFOCanvas->SetFreq(lfoFreq);
        
            double lfoAmp =
                ((LoopSource *)source)->GetParameter(LoopSource::P_LFO_AMP)->GetValue();
            _grainLFOCanvas->SetAmp(lfoAmp);
        }
        else
        {
            _grainLFOWindow->SetVisible(false);
        }
    }
}

void
App::UpdateEnvelopeWindow()
{
    Source *source = GetCurrentSource();
    int paramIdx = GetCurrentSourceParamIdx();
    
    if (source->GetSourceType() == Source::GRAIN_CLOUD)
    {
        if (paramIdx == GrainCloud::ENVELOPE)
        {
            _envelopeTypeWindow->SetVisible(true);
            
            int envelopeType =
                ((GrainCloud *)source)->GetParameter(GrainCloud::ENVELOPE)->GetValue();
            _envelopeTypeCanvas->SetEnvelopeType(envelopeType);
        }
        else
        {
            _envelopeTypeWindow->SetVisible(false);
        }
    }
    else if (source->GetSourceType() == Source::SINGLE_GRAIN)
    {
        if (paramIdx == SingleGrain::ENVELOPE)
        {
            _envelopeTypeWindow->SetVisible(true);
            
            int envelopeType =
                ((SingleGrain *)source)->GetParameter(SingleGrain::ENVELOPE)->GetValue();
            _envelopeTypeCanvas->SetEnvelopeType(envelopeType);
        }
        else
        {
            _envelopeTypeWindow->SetVisible(false);
        }
    }
    else if (source->GetSourceType() == Source::LOOP_SOURCE)
    {
        if (paramIdx == LoopSource::ENVELOPE)
        {
            _envelopeTypeWindow->SetVisible(true);
            
            int envelopeType =
                ((LoopSource *)source)->GetParameter(LoopSource::ENVELOPE)->GetValue();
            _envelopeTypeCanvas->SetEnvelopeType(envelopeType);
        }
        else
        {
            _envelopeTypeWindow->SetVisible(false);
        }
    }
}

void
App::UpdateMidiIn()
{
    while(_midiQueue->GetSize() > 0)
    {
        MidiQueue::MidiMessage msg;
        bool success = _midiQueue->PopMessage(&msg);
        if (success)
        {
#if 0 // Debug
            if (msg.IsNoteOn())
            {
                int noteNum = msg.GetNoteNumber();
                int chan = msg.GetChannel();
                fprintf(stderr, "Channel: %d Midi note on: %d\n", chan, noteNum);
            }

            if (msg.IsNoteOff())
            {
                int noteNum = msg.GetNoteNumber();
                int chan = msg.GetChannel();
                fprintf(stderr, "Channel: %d Midi note off: %d\n", chan, noteNum);
            }

            if (msg.IsPitchWheel())
            {
                double pitchWheel = msg.GetPitchWheel();
                int chan = msg.GetChannel();
                fprintf(stderr, "Channel: %d Midi pitch wheel: %g\n", chan, pitchWheel);
            }
#endif

            if (msg.IsNoteOn())
            {                
                int noteNum = msg.GetNoteNumber();
                if (_selectedSource >= 0)
                {
                    Source *noteSource = _sources[_selectedSource]->Clone();
                    noteSource->SetMidiNote(noteNum);
                    _noteSources[_selectedSource].push_back(noteSource);
                }
            }

            if (msg.IsNoteOff())
            {    
                int noteNum = msg.GetNoteNumber();
                vector<vector<Source *> > sourcesToKeep;
                sourcesToKeep.resize(_noteSources.size());
                for (int i = 0; i < _noteSources.size(); i++)
                {
                    for (int j = 0; j < _noteSources[i].size(); j++)
                    {
                        Source *source = _noteSources[i][j];
                        if (source->GetMidiNote() != noteNum)
                            sourcesToKeep[i].push_back(source);
                        else
                        {
                            delete source;
                        }
                    }
                }

                _noteSources = sourcesToKeep;
            }

            if (msg.IsPitchWheel())
            {
                double pitch = msg.GetPitchWheel();

                // Played notes
                for (int i = 0; i < _noteSources.size(); i++)
                    for (int j = 0; j < _noteSources[i].size(); j++)
                    {
                        Source *source = _noteSources[i][j];
                        source->SetMidiPitchWheelValue(pitch);
                    }

                // But also currently selected source
                if (_selectedSource >= 0)
                {
                    _sources[_selectedSource]->SetMidiPitchWheelValue(pitch);
                }
            }
        }
    }

    if (!IsPlayingNoteSource())
        // Reset virtual pitch value
        _virtualPitch = 0.0;
}

void
App::OnIdle()
{
    _mutex->lock();
    
#ifdef __linux__
    if (_fps > 0.0)
        usleep(1000000/_fps);
#endif

    UpdateMidiIn();
        
    Source *source = GetCurrentSource();
    if (source == NULL)
    {
        _mutex->unlock();
        
        glutPostRedisplay();
        
        return;
    }

    UpdateLFOWindow();

    UpdateEnvelopeWindow();
    
    glutPostRedisplay();
    
    _mutex->unlock();
}

void
App::DrawFps()
{
    static int frame = 0;
    static int timebase = 0;
    
    frame++;
	int time = glutGet(GLUT_ELAPSED_TIME);

    static double fps = 0.0;
    
	if (time - timebase > 1000) {
		fps = frame*1000.0/(time-timebase);
	 	timebase = time;
		frame = 0;
    }
    
    // Draw
    char fpsStr[255];
    sprintf(fpsStr, "%d fps", (int)fps);

    double a = 0.6f + 0.2*sin(0.8*PI*Time::Instance().GetSeconds());
        
    double paramColor[4];
    _style->GetParamColor(paramColor);
    glColor4f(paramColor[0], paramColor[1],
              paramColor[2], a*paramColor[3]);
    
    glLineWidth(2.0f);

    glPushMatrix();
    glTranslatef(_position._x*_winWidth,
                 _position._y*_winHeight, 0.0);
    
    const double scale = 0.06;
    DrawString(0.0 + (double)_winWidth*scale*0.25,
               0.0 + (double)_winWidth*scale*0.25,
               0.5f,
               fpsStr,
               (double)_winWidth*scale);

    glPopMatrix();
}

void
App::DrawString(GLdouble x, GLdouble y, GLdouble z,
                        const char * str,
                        GLdouble scale)
{
    GLint len = strlen(str);
    GLint i;
    
    glPushMatrix();
    glTranslatef(x, _winHeight - y, z);
    glScalef(.001f*scale, .001f*scale, .001f);
    
    for(i = 0; i < len; i++)
        glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, str[i]);
    
    glPopMatrix();
}

void
App::DrawString2(GLdouble x, GLdouble y, GLdouble z,
                         const char * str,
                         GLdouble scale)
{
    GLint len = strlen(str);
    GLint i;
    
    glPushMatrix();
    
    glTranslatef(x, _winHeight - y, z);
    glScalef(scale, scale, 1.0);    
    
    for(i = 0; i < len; i++)
        glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, str[i]);
    
    glPopMatrix();
}

void
App::DrawTitle()
{
    glPushMatrix();
    glLoadIdentity();
    
    double titleLineWidth = _style->GetTitleLineWidth();
    glLineWidth(titleLineWidth);
    
    double a = 0.6f + 0.2*sin(0.8*PI*Time::Instance().GetSeconds());

    double colCoeffA[4];
    _style->GetTitleAColorCoeff(colCoeffA);
    glColor4f(colCoeffA[0], colCoeffA[1],
              colCoeffA[2], a*colCoeffA[3]);

    char appName[512];
    _style->GetAppName(appName);
        
    DrawString(_winWidth/2.0f + 0.2f*(double)_winWidth,
               (double)_winHeight/2.0f,
               0.5f,
               appName,
               (double)_winWidth*0.1f);
   
    a = 0.6f + 0.2*sin(0.9*PI*Time::Instance().GetSeconds());
    double insColor = a*0.4f;

    double colCoeffIns[4];
    _style->GetTitleInsColorCoeff(colCoeffIns);
    glColor4f(colCoeffIns[0],
              colCoeffIns[1],
              colCoeffIns[2],
              a*colCoeffIns[3]);

    DrawString(_winWidth/2.0f + 0.2f*(double)_winWidth + 10.0,
               (double)_winHeight/2.0f + 30.0, 0.5f,"CLICK TO START",
               (double)_winWidth*0.04f);

    a = 0.6f + 0.2*sin(1.0*PI*Time::Instance().GetSeconds());
    insColor = a*0.4f;
    glColor4f(colCoeffIns[0], colCoeffIns[1],
              colCoeffIns[2], a*colCoeffIns[3]);
    
    DrawString(_winWidth/2.0f + 0.2f*(double)_winWidth+10.0,
               (double)_winHeight/2.0f + 50.0, 0.5f,
               "ESCAPE TO QUIT",
               (double)_winWidth*0.04f);

    glPopMatrix();
}

void
App::DrawHelp(bool shortHelp)
{
    if (shortHelp)
    {
        // Draw a dark transparent overlay
        double col[4];
        _style->GetHelpBackgroundColor(col);
        glColor4f(col[0], col[1], col[2], col[3]);

        glPushMatrix();

        glLoadIdentity();
    
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
        glBegin(GL_QUADS);
        glVertex3f(0.0,       0.0,     0.0);
        glVertex3f(_winWidth, 0.0,     0.0);
        glVertex3f(_winWidth, _winHeight, 0.0);
        glVertex3f(0.0,       _winHeight, 0.0);
        glEnd();
        
        glPopMatrix();
    }
    
    double helpLineWidth = _style->GetHelpLineWidth();
    glLineWidth(helpLineWidth);

    double insColor[4];
    _style->GetHelpInsColor(insColor);
    glColor4f(insColor[0], insColor[1],
              insColor[2], insColor[3]);

    const char **commands = NULL;
    if (!shortHelp)
        commands = _control->GetCommands();
    else
        commands = _control->GetCommandsShort();
            
    double x = (double)_winWidth*0.1;
    double y = (double)_winHeight*0.15;

    if (shortHelp)
        // Align more on the left
        x = (double)_winWidth*0.02;
    
    double yStep = (double)_winWidth*0.01f;
    double size = (double)_winWidth*0.06f;

    glPushMatrix();
    glLoadIdentity();
    
    int numCommands = !shortHelp ? _control->GetNumCommands() : 
        _control->GetNumCommandsShort();

    for (int i = 0; i < numCommands; i++)
    {
        const char *cmd = commands[i];
        DrawString(x, y + i*yStep, 0.5, cmd, size);
    }

    glPopMatrix();
}

void
App::DrawBounceFile()
{
    if (_bounceFile == NULL)
        return;

    double a = 0.6f + 0.2*sin(0.8*PI*Time::Instance().GetSeconds());

    double paramColor[4];
    _style->GetParamColor(paramColor);
    glColor4f(paramColor[0], paramColor[1],
              paramColor[2], a*paramColor[3]);

    char str[255];
    sprintf(str, "*** Bouncing to: %s ***",
            _bounceFileName);

    glPushMatrix();
    glTranslatef(_position._x*_winWidth,
                 _position._y*_winHeight, 0.0);
    
#define BORDER 0.02
    DrawString(BORDER*(double)_winWidth,
               (double)_winHeight - BORDER*_winHeight,
               0.5f,
               str,
               (double)_winWidth*0.1f);

    glPopMatrix();
}

void
specialFunc(int key, int x, int y)
{
    _app.OnSpecialKey(key, x, y);
}

void
App::OnSpecialKey(int key, int x, int y)
{
    _mutex->lock();
    
    _control->OnSpecialKey(key, x, y);

    _mutex->unlock();
}

void
specialUpFunc(int key, int x, int y)
{
    _app.OnSpecialKeyUp(key, x, y);
}

void
App::OnSpecialKeyUp(int key, int x, int y)
{
    _mutex->lock();
    
    _control->OnSpecialKeyUp(key, x, y);

    _mutex->unlock();
}

void
keyUpFunc(unsigned char key, int x, int y)
{
    _app.OnKeyUp(key, x, y);
}

void
App::OnKeyUp(unsigned char key, int x, int y)
{
    _mutex->lock();
    
    _control->OnKeyUp(key, x, y);

    _mutex->unlock();
}

void
keyboardFunc(unsigned char key, int x, int y)
{
    _app.OnKeyboard(key, x, y);
}

void
App::OnKeyboard(unsigned char key, int x, int y)
{
    _mutex->lock();
    
    _control->OnKeyboard(key, x, y);

    _mutex->unlock();
}

void
App::Deselect(int shapeType)
{
    switch (shapeType)
    {
        case SOURCE:
            if (_selectedSource >=0)
            {
                _sources[_selectedSource]->GetUI()->SetSelectState(false);
                _selectedSource = -1;
            }
            break;
            
        case RECT:
            if (_selectedRect >= 0)
            {
                _soundRects[_selectedRect]->GetUI()->SetSelectState(false);
                _selectedRect = -1;
            }
            break;

        default:
            break;
    }
}

void
mouseFunc(int button, int state, int x, int y)
{
    _app.OnMouse(button, state, x, y);
}

void
App::OnMouse(int button, int state, int x, int y)
{
    _mutex->lock();

    bool overWindow = false;
    if (_grainLFOCanvas->Pick(((double)x)/_winWidth,
                              ((double)y)/_winHeight) ||
        _envelopeTypeCanvas->Pick(((double)x)/_winWidth,
                                  ((double)y)/_winHeight) ||
        _effectsCanvas->Pick(((double)x)/_winWidth,
                             ((double)y)/_winHeight))
        overWindow = true;

    // Don't interact with sources and rects when picking on a window
    if (!overWindow)
        _control->OnMouse(button, state, x, y);
    
    if (_envelopeTypeCanvas != NULL)
        _envelopeTypeCanvas->OnMouse(button, state,
                                     ((double)x)/_winWidth,
                                     ((double)y)/_winHeight);

    if (_effectsCanvas != NULL)
        _effectsCanvas->OnMouse(button, state,
                                ((double)x)/_winWidth,
                                ((double)y)/_winHeight);
    
    _mutex->unlock();
}

void
mouseDragFunc(int x, int y)
{
    _app.OnMouseDrag(x, y);
}

void
App::OnMouseDrag(int x, int y)
{
    _mutex->lock();
    
    _control->OnMouseDrag(x, y);
    
    if (_envelopeTypeCanvas != NULL)
        _envelopeTypeCanvas->OnMouseDrag(((double)x)/_winWidth,
                                         ((double)y)/_winHeight);

    double fineDragFactor = _control->GetFineDragFactor();
        
    if (_effectsCanvas != NULL)
        _effectsCanvas->OnMouseDrag(((double)x)/_winWidth,
                                    ((double)y)/_winHeight,
                                    fineDragFactor);
    
    _mutex->unlock();
}

void
mousePassiveMotionFunc(int x, int y)
{
    _app.OnMousePassiveMotion(x, y);
}

void
App::OnMousePassiveMotion(int x, int y)
{
    _mutex->lock();
    
    _control->OnMousePassiveMotion(x, y);

    _mutex->unlock();
}

void
mouseWheelFunc(int button, int dir, int x, int y)
{
    _app.OnMouseWheel(button, dir, x, y);
}

void
App::OnMouseWheel(int button, int dir, int x, int y)
{
    _mutex->lock();
    
    _control->OnMouseWheel(button, dir, x, y);

    _mutex->unlock();
}

void
App::AddGrainCloud(double x, double y)
{
    int numVoices = 6;
    int idx = _sources.size();
    if (_selectedSource >=0)
    {
        if (!_sources.empty())
            _sources[_selectedSource]->GetUI()->SetSelectState(false);
    }

    Source *source0 = GetCurrentSource();
    if (source0 != NULL)
        source0->GetUI()->SetSelectState(false);
    
    _selectedSource = idx;

    GrainCloud *cloud = new GrainCloud(this,
                                       (const vector<SamplesProvider *> &)_soundRects,
                                       numVoices);
    _sources.push_back(cloud);
    _noteSources.resize(_sources.size());
    
    UIGrainCloud *ui = new UIGrainCloud(cloud, _style, _screen,
                                        x/_winWidth,
                                        y/_winHeight,
                                        numVoices,
                                        _soundRects);

    ui->SetSelectState(true);
    
    _sources[idx]->RegisterUI(ui);

    _pieMenu->InitFromParameters();
}

void
App::AddSingleGrain(double x, double y)
{
    int idx = _sources.size();
    if (_selectedSource >=0)
    {
        if (!_sources.empty())
            _sources[_selectedSource]->GetUI()->SetSelectState(false);
    }

    Source *source0 = GetCurrentSource();
    if (source0 != NULL)
        source0->GetUI()->SetSelectState(false);
    
    _selectedSource = idx;

    SingleGrain *grain = new SingleGrain(this,
                                         (const vector<SamplesProvider *> &)_soundRects);
    _sources.push_back(grain);
    _noteSources.resize(_sources.size());
    
    UISingleGrain *ui = new UISingleGrain(grain, _style, _screen,
                                          x/_winWidth,
                                          y/_winHeight,
                                          &_soundRects);

    ui->SetSelectState(true);
    
    _sources[idx]->RegisterUI(ui);

    _pieMenu->InitFromParameters();
}

void
App::AddLoopSource(double x, double y)
{
    int idx = _sources.size();
    if (_selectedSource >=0)
    {
        if (!_sources.empty())
            _sources[_selectedSource]->GetUI()->SetSelectState(false);
    }

    Source *source0 = GetCurrentSource();
    if (source0 != NULL)
        source0->GetUI()->SetSelectState(false);
    
    _selectedSource = idx;

    LoopSource *source = new LoopSource(this,
                                        (const vector<SamplesProvider *> &)_soundRects);
    _sources.push_back(source);
    _noteSources.resize(_sources.size());
    
    UILoopSource *ui = new UILoopSource(source, _style, _screen,
                                        x/_winWidth,
                                        y/_winHeight,
                                        &_soundRects);

    ui->SetSelectState(true);
    
    _sources[idx]->RegisterUI(ui);

    _pieMenu->InitFromParameters();
}

void
App::SaveCurrentSession()
{
    SerializeSession();
}

ParameterProvider *
App::GetCurrentParamProvider()
{
    Source *source = GetCurrentSource();
    if (source != NULL)
        return source;

    UISoundRect *rect = GetCurrentSoundRect();
    if (rect != NULL)
        return rect;

    return NULL;
}

void
App::SendNoteOn(int midiNote)
{
    _midiOut->SendNoteOn(midiNote);
}

void
App::SendNoteOff(int midiNote)
{
    _midiOut->SendNoteOff(midiNote);
}

bool
App::IsPlayingNoteSource() const
{
    bool noteSourcesEmpty = true;
    for (int i = 0; i < _noteSources.size(); i++)
    {
        if (!_noteSources[i].empty())
        {
            noteSourcesEmpty = false;
            
            break;
        }
    }

    return (!noteSourcesEmpty);
}

void
App::ApplyVirtualPitchWheel(double dx, double dy)
{
#define PITCH_WHEEL_SENSIVITY 0.002

    double d = (fabs(dx) > fabs(dy)) ? dx : dy;
    double dpitch = d*PITCH_WHEEL_SENSIVITY;

    _virtualPitch += dpitch;
        
    if (_virtualPitch > 1.0)
        _virtualPitch = 1.0;
    if (_virtualPitch < -1.0)
        _virtualPitch = -1.0;
    
    for (int i = 0; i < _noteSources.size(); i++)
        for (int j = 0; j < _noteSources[i].size(); j++)
        {
            Source *source = _noteSources[i][j];
            source->SetMidiPitchWheelValue(_virtualPitch);
        }
}

const string
App::GetAudioPath() const
{
    return _audioPath;
}

void
App::PrevStyle()
{
    int styleNum = _styleManager->GetCurrentStyleNum();
    styleNum = (styleNum - 1) % _styleManager->GetNumStyles();
    if (styleNum < 0)
        styleNum += _styleManager->GetNumStyles();
    _styleManager->SetStyleNum(styleNum);
    UpdateStyle();
}

void
App::NextStyle()
{
    int styleNum = _styleManager->GetCurrentStyleNum();
    styleNum = (styleNum + 1) % _styleManager->GetNumStyles();
    _styleManager->SetStyleNum(styleNum);
    UpdateStyle();
}

double
App::GetPositionX() const
{
    return _position._x*_winWidth;
}

double
App::GetXMoveSpeed() const
{
    return _xMoveSpeed;
}

void
App::ApplyZoom(double dzoom,
                       double zoomCenterX, double zoomCenterY)
{
    zoomCenterX /= _winWidth;
    zoomCenterY /= _winHeight;
        
    double prevZoom = _zoom;
    _zoom *= dzoom;

    if (_zoom < MIN_APP_ZOOM)
        _zoom = prevZoom;
        
    double a = ((double)(zoomCenterX + _position._x))/prevZoom;
    double b = ((double)(zoomCenterY + _position._y))/prevZoom;
    
    _position._x = a * _zoom - zoomCenterX;
    _position._y = b * _zoom - zoomCenterY;
}

double
App::GetZoom() const
{
    return _zoom;
}

void
App::TranslatePositionX(double dx)
{
    _position._x += dx/_winWidth;
}

double
App::GetPositionY() const
{
    return _position._y*_winHeight;
}

double
App::GetYMoveSpeed() const
{
    return _yMoveSpeed;
}

void
App::TranslatePositionY(double dy)
{
    _position._y += dy/_winHeight;
}

bool
App::GetMenuFlag() const
{
    return _menuFlag;
}

void
App::SetMenuFlag(bool flag)
{
    _menuFlag = flag;
}

int
App::GetWinWidth() const
{
    return _winWidth;
}

int
App::GetWinHeight() const
{
    return _winHeight;
}

void
App::SetDragMode(int mode)
{
    _dragMode = mode;
}

int
App::GetDragMode() const
{
    return _dragMode;
}

void
App::OpenPieMenu()
{
    if ((_selectedSource >= 0) ||
        (_selectedRect >= 0))
    {
        _pieMenu->SetVisible(true);
    }
}

void
App::ClosePieMenu()
{
    if (!_isEditingParamString)
    {
        _pieMenu->SetVisible(false);

        Source *source = GetCurrentSource();
        if (source != NULL)
            source->ResetCurrentParameterIndex();

        UISoundRect *rect = GetCurrentSoundRect();
        if (rect != NULL)
            rect->ResetCurrentParameterIndex();
    }
}

bool
App::IsPieMenuOpen() const
{
    bool open = _pieMenu->IsVisible();
    return open;
}

void
App::PieMenuMouseDrag(double dx, double dy)
{
    _pieMenu->OnMouseDrag(dx/_winWidth, dy/_winHeight);
}

void
App::RightPickPieMenu(int button, int state, double x, double y)
{
    if (_pieMenu->IsVisible())
    {
        _pieMenu->OnMouse(button, state,
                          x/_winWidth, y/_winHeight);
            
        _isEditingParamString = true;
    }
}

void
App::LeftPickPieMenu(int button, int state, double x, double y)
{
    bool piePicked = _pieMenu->OnMouse(button, state,
                                       x/_winWidth, y/_winHeight);
    if (piePicked)
        return;

    Deselect(SOURCE);
    //Deselect(RECT);
 
    UpdateSourcePicking(x, y);
    UpdateRectPicking(x, y);
}

void
App::OpenEffectsWindow(bool master)
{
    if (_effectsWindow != NULL)
    {
        _effectsWindow->SetVisible(true);

        if (master)
        {
            _masterEffectsBox->RegisterUI(_effectsCanvas);
            _effectsWindow->SetTitle("Effects | Master");
        }
        else
        {
            Source *source = GetCurrentSource();
            if (source != NULL)
            {
                EffectsBox *effectsBox = source->GetEffectsBox();
                effectsBox->RegisterUI(_effectsCanvas);
            }
            
            _effectsWindow->SetTitle("Effects | Source");
        }
    }
}

void
App::CloseEffectsWindow()
{
    if (_effectsWindow != NULL)
        _effectsWindow->SetVisible(false);
}

UISoundRect *
App::GetCurrentSoundRect()
{
    if (_selectedRect < 0)
        return NULL;
    
    return _soundRects[_selectedRect]->GetUI();
}

void
App::TranslateSoundRect(double dx, double dy)
{
    if (_selectedRect >=0)
        _soundRects[_selectedRect]->GetUI()->Move(dx/_winWidth, dy/_winHeight);
}

void
App::ScaleSoundRect(double x, double y, double dx, double dy)
{
    if (_selectedRect >= 0)
    {
        // Center
        double cx = _soundRects[_selectedRect]->GetUI()->GetX();
        double cy = _soundRects[_selectedRect]->GetUI()->GetY();

        cx *= _winWidth;
        cy *= _winHeight;
        
        // On which side of the center?
        if (x < cx)
            dx = -dx;
        if (y < cy)
            dy = -dy;
        
        // Rectangle dimensions
        double rw = _soundRects[_selectedRect]->GetUI()->GetWidth();
        double rh = _soundRects[_selectedRect]->GetUI()->GetHeight();

        rw *= _winWidth;
        rh *= _winHeight;
        
        // Distance to the center
        double dxc = fabs(x - cx);
        double dyc = fabs(y - cy);

#define DIST_THRS 10.0
        double wfact = (dxc > DIST_THRS) ?
            2.0*(rw*0.5)/dxc : 0.0;
        double hfact = (dyc > DIST_THRS) ?
            2.0*(rh*0.5)/dyc : 0.0;
        
        rw += 2.0*dx*wfact;
        rh += 2.0*dy*hfact;

        rw /= _winWidth;
        rh /= _winHeight;
        
        _soundRects[_selectedRect]->GetUI()->SetDimensions(rw, rh);
    }
}

void
App::ToggleSoundRectOrientation()
{
    if (_selectedRect >=0)
        _soundRects[_selectedRect]->GetUI()->ToggleOrientation();
}

void
App::SetSoundRectResizeMode()
{
    _dragMode = RESIZE;
}

void
App::SetSoundRectMoveMode()
{
    _dragMode = MOVE;
}

bool
App::IsSelectedRect() const
{
    bool selected = (_selectedRect >= 0);

    return selected;
}

void
App::ResetSelectedRect()
{
    UISoundRect *rect = GetCurrentSoundRect();
    if (rect != NULL)
        rect->SetSelectState(false);
    
    _selectedRect = -1;
}

bool
App::IsEditingParamString() const
{
    return _isEditingParamString;
}

void
App::AddParamStringChar(char c)
{
    if (_isEditingParamString)
        _paramValueString.push_back(c);
}

void
App::DelLastParamStringChar()
{
    if (_paramValueString.size () > 0)
        _paramValueString.resize(_paramValueString.size () - 1);
}

void
App::InsertParamStringChar(int index, const char *c)
{
    if (_isEditingParamString)
        _paramValueString.insert(index, c);
}

void
App::ApplyParamStringGrainCloud(GrainCloud *cloud, double value)
{
    GrainCloud::ParamIndex paramIdx =
        (GrainCloud::ParamIndex)GetCurrentSourceParamIdx();
    switch (paramIdx)
    {
        case GrainCloud::DURATION:
            if (_selectedSource >=0)
            {
                if (value < 1.0)
                    value = 1.0;
                
                cloud->SetDurationMs(value);
            }
            break;
                        
        case GrainCloud::OVERLAP:
            if (_selectedSource >=0)
                cloud->SetOverlap(value);
            break;
                        
        case GrainCloud::PITCH:
            if (_selectedSource >=0)
                cloud->SetPitch(value);
            break;
            
        case GrainCloud::P_LFO_FREQ:
            if (_selectedSource >=0)
                cloud->SetPitchLFOFreq(value);
            break;
            
        case GrainCloud::P_LFO_AMP:
            if (_selectedSource >=0)
                cloud->SetPitchLFOAmount(value);
            break;
            
        case GrainCloud::VOLUME:
            if (_selectedSource >=0)
                cloud->SetVolumeDb(value);

        case GrainCloud::TIME_STRETCH:
            if (_selectedSource >=0)
                cloud->SetTimeStretch(value);
            break;
                        
        default:
            break;
    }
    
    if (paramIdx != GrainCloud::NO_PARAMETER)
    {
        cloud->GetParameter(paramIdx)->SetValue(value);
        _pieMenu->UpdateCurrentParamPos();
    }
}

void
App::ApplyParamStringSingleGrain(SingleGrain *grain, double value)
{
    SingleGrain::ParamIndex paramIdx =
        (SingleGrain::ParamIndex)GetCurrentSourceParamIdx();
    switch (paramIdx)
    {
        case SingleGrain::DURATION:
            if (_selectedSource >=0)
            {
                if (value < 1.0)
                    value = 1.0;
                
                grain->SetDurationMs(value);
            }
            break;
                        
        case SingleGrain::PITCH:
            if (_selectedSource >=0)
                grain->SetPitch(value);
            break;
            
        case SingleGrain::P_LFO_FREQ:
            if (_selectedSource >=0)
                grain->SetPitchLFOFreq(value);
            break;
            
        case SingleGrain::P_LFO_AMP:
            if (_selectedSource >=0)
                grain->SetPitchLFOAmount(value);
            break;
            
        case SingleGrain::VOLUME:
            if (_selectedSource >=0)
                grain->SetVolumeDb(value);

        case SingleGrain::TIME_STRETCH:
            if (_selectedSource >=0)
                grain->SetTimeStretch(value);
            break;
            
        default:
            break;
    }
    
    if (paramIdx != SingleGrain::NO_PARAMETER)
    {
        grain->GetParameter(paramIdx)->SetValue(value);
        _pieMenu->UpdateCurrentParamPos();
    }
}

void
App::ApplyParamStringLoopSource(LoopSource *source, double value)
{
    LoopSource::ParamIndex paramIdx =
        (LoopSource::ParamIndex)GetCurrentSourceParamIdx();
    switch (paramIdx)
    {
        case LoopSource::DURATION:
            if (_selectedSource >=0)
            {
                if (value < 1.0)
                    value = 1.0;
                
                source->SetDurationMs(value);
            }
            break;
                        
        case LoopSource::PITCH:
            if (_selectedSource >=0)
                source->SetPitch(value);
            break;
            
        case LoopSource::P_LFO_FREQ:
            if (_selectedSource >=0)
                source->SetPitchLFOFreq(value);
            break;
            
        case LoopSource::P_LFO_AMP:
            if (_selectedSource >=0)
                source->SetPitchLFOAmount(value);
            break;
            
        case LoopSource::VOLUME:
            if (_selectedSource >=0)
                source->SetVolumeDb(value);

        case LoopSource::TIME_STRETCH:
            if (_selectedSource >=0)
                source->SetTimeStretch(value);
            break;
            
        default:
            break;
    }
    
    if (paramIdx != LoopSource::NO_PARAMETER)
    {
        source->GetParameter(paramIdx)->SetValue(value);
        _pieMenu->UpdateCurrentParamPos();
    }
}

void
App::ApplyParamString()
{
    if (_paramValueString != "")
    {
        if (_isEditingParamString)
        {
            double value = atof(_paramValueString.c_str());

            Source *source = GetCurrentSource();
            if (source != NULL)
            {
                if (source->GetSourceType() == Source::GRAIN_CLOUD)
                    ApplyParamStringGrainCloud((GrainCloud *)source, value);
                if (source->GetSourceType() == Source::SINGLE_GRAIN)
                    ApplyParamStringSingleGrain((SingleGrain *)source, value);
                if (source->GetSourceType() == Source::LOOP_SOURCE)
                    ApplyParamStringLoopSource((LoopSource *)source, value);
            }

            UISoundRect *rect = GetCurrentSoundRect();
            if (rect != NULL)
            {
                UISoundRect::ParamIndex paramIdx = GetCurrentRectParamIdx();
                switch (paramIdx)
                {
                    case UISoundRect::TYPE:
                        if (_selectedRect >=0)
                        {
                            _soundRects[_selectedRect]->GetUI()->
                                SetRectType((UISoundRect::RectType)value);
                        }
                        break;

                    case UISoundRect::RANGE:
                        if (_selectedRect >=0)
                        {
                            _soundRects[_selectedRect]->GetUI()->
                                SetColorRange(value);
                        }
                        break;

                    case UISoundRect::CONTRAST:
                        if (_selectedRect >=0)
                        {
                            _soundRects[_selectedRect]->GetUI()->
                                SetColorContrast(value);
                        }
                        break;

                    case UISoundRect::ORIENTATION:
                        if (_selectedRect >=0)
                        {
                            _soundRects[_selectedRect]->GetUI()->
                                SetOrientation((UISoundRect::Orientation)value);
                        }
                        break;
                        
                    default:
                        break;
                }

                if (paramIdx != UISoundRect::NO_PARAMETER)
                {
                    rect->GetParameter(paramIdx)->SetValue(value);
                    _pieMenu->UpdateCurrentParamPos();
                }
            }
        }
    }

    ResetParamString();
}

void
App::ResetParamString()
{
    _paramValueString = "";
    _isEditingParamString = false;
}

const char *
App::GetParamValueString() const
{
    const char *paramString = _paramValueString.c_str();
    if (strlen(paramString) == 0)
        return NULL;

    return paramString;
}

void
App::ResetCurrentParam()
{
    _paramValueString = "";
    _isEditingParamString = false;
    
    Source *source = GetCurrentSource();
    if (source != NULL)
        source->ResetCurrentParameterIndex();

    UISoundRect *rect = GetCurrentSoundRect();
    if (rect != NULL)
        rect->ResetCurrentParameterIndex();
}

int
App::GetCurrentSourceParamIdx() const
{
    int paramIdx = -1;
    Source *source = ((App *)this)->GetCurrentSource();
    if (source != NULL)
    {
        Parameter *param = source->GetCurrentParameter();
        if (param != NULL)
            paramIdx = param->GetIndex();
    }
    
    return paramIdx;
}

UISoundRect::ParamIndex
App::GetCurrentRectParamIdx() const
{
    UISoundRect::ParamIndex paramIdx =
        UISoundRect::NO_PARAMETER;
    UISoundRect *rect = ((App *)this)->GetCurrentSoundRect();
    if (rect != NULL)
    {
        Parameter *param = rect->GetCurrentParameter();
        if (param != NULL)
            paramIdx = (UISoundRect::ParamIndex)param->GetIndex();
    }
    
    return paramIdx;
}

void
App::OnCurrentParamValueUpdated()
{
    UpdateParamWindows();
    UpdateParamNoteSources();
}

void
App::UpdateParamWindows()
{
    Source *source = GetCurrentSource();
    if (source == NULL)
        return;
    int paramIdx = GetCurrentSourceParamIdx();

    if (source->GetSourceType() == Source::GRAIN_CLOUD)
    {
        if ((paramIdx == GrainCloud::ENVELOPE) ||
            (paramIdx == GrainCloud::P_LFO_FREQ) ||
            (paramIdx == GrainCloud::P_LFO_AMP))
        {        
            const char *name =
                source->GetParameter(paramIdx)->GetUI()->GetName();
            const char *valueStr =
                source->GetParameter(paramIdx)->GetUI()->GetValueStr();
        
            char label[255];
            sprintf(label, "%s: %s", name, valueStr);

            if (paramIdx == GrainCloud::ENVELOPE)
                _envelopeTypeWindow->SetTitle(label);
            else if ((paramIdx == GrainCloud::P_LFO_FREQ) ||
                     (paramIdx == GrainCloud::P_LFO_AMP))
                _grainLFOWindow->SetTitle(label);
        }
    }
    else if (source->GetSourceType() == Source::SINGLE_GRAIN)
    {
        if ((paramIdx == SingleGrain::ENVELOPE) ||
            (paramIdx == SingleGrain::P_LFO_FREQ) ||
            (paramIdx == SingleGrain::P_LFO_AMP))
        {        
            const char *name =
                source->GetParameter(paramIdx)->GetUI()->GetName();
            const char *valueStr =
                source->GetParameter(paramIdx)->GetUI()->GetValueStr();
            
            char label[255];
            sprintf(label, "%s: %s", name, valueStr);

            if (paramIdx == SingleGrain::ENVELOPE)
                _envelopeTypeWindow->SetTitle(label);
            else if ((paramIdx == SingleGrain::P_LFO_FREQ) ||
                     (paramIdx == SingleGrain::P_LFO_AMP))
                _grainLFOWindow->SetTitle(label);
        }
    }
    else if (source->GetSourceType() == Source::LOOP_SOURCE)
    {
        if ((paramIdx == LoopSource::ENVELOPE) ||
            (paramIdx == LoopSource::P_LFO_FREQ) ||
            (paramIdx == LoopSource::P_LFO_AMP))
        {        
            const char *name =
                source->GetParameter(paramIdx)->GetUI()->GetName();
            const char *valueStr =
                source->GetParameter(paramIdx)->GetUI()->GetValueStr();
            
            char label[255];
            sprintf(label, "%s: %s", name, valueStr);

            if (paramIdx == LoopSource::ENVELOPE)
                _envelopeTypeWindow->SetTitle(label);
            else if ((paramIdx == LoopSource::P_LFO_FREQ) ||
                     (paramIdx == LoopSource::P_LFO_AMP))
                _grainLFOWindow->SetTitle(label);
        }
    }
}

void
App::UpdateParamNoteSources()
{
    Source *source = GetCurrentSource();
    if (source == NULL)
        return;
    
    int paramIdx = GetCurrentSourceParamIdx();
    Parameter *param = source->GetParameter(paramIdx);

    for (int i = 0; i < _noteSources.size(); i++)
        for (int j = 0; j < _noteSources[i].size(); j++)
        {
            Source *noteSource = _noteSources[i][j];
            if (noteSource->GetId() == source->GetId())
                // This is one of the clones
            {
                double val = param->GetValue();
                param->SetValue(noteSource, val);
            }
        }
}

void
App::UpdateSourcePicking(double x, double y)
{
    x /= _winWidth;
    y /= _winHeight;

    // Unselect current source
    Source *source = GetCurrentSource();
    if (source != NULL)
        source->GetUI()->SetSelectState(false);
            
    _selectedSource = -1;
    
    // Check sources to see if we have selection
    for (int i = 0; i < _sources.size(); i++)
    {
        if (_sources[i]->GetUI()->Pick(x, y))
        {
            // Select picked source
            _sources[i]->GetUI()->SetSelectState(true);
            _selectedSource = i;
            
            break;
        }            
    }
}

void
App::DelSelectedSource()
{
    if (!_isEditingParamString && (_paramValueString == ""))
    {
        if (_selectedSource >=0)
        {
            _sources.erase(_sources.begin() + _selectedSource);
            _noteSources.erase(_noteSources.begin() + _selectedSource);
            _selectedSource = -1;
        }
    }
}

bool
App::IsSelectedSource() const
{
    bool selected = (_selectedSource >= 0);

    return selected;
}

void
App::ResetSelectedSource()
{
    Source *source = GetCurrentSource();
    if (source != NULL)
        source->GetUI()->SetSelectState(false);
    
    _selectedSource = -1;
}

Source *
App::GetCurrentSource()
{
    if (_selectedSource < 0)
        return NULL;

    Source *source = _sources[_selectedSource];
    
    return source;
}

void
App::SourceMouseDrag(double dx, double dy)
{
    dx /= _winWidth;
    dy /= _winHeight;
    
    if (_selectedSource >= 0)
    {
        int paramIdx = GetCurrentSourceParamIdx();
            
        if (paramIdx == -1)
            // Drag could only if we are not dragging a param value
        {
            double x = _sources[_selectedSource]->GetUI()->GetX();
            double y = _sources[_selectedSource]->GetUI()->GetY();

            x += dx;
            y += dy;
            
            _sources[_selectedSource]->GetUI()->UpdatePosition(x, y);
        }
    }
}

void
App::RectMouseDrag(double dx, double dy)
{
    dx /= _winWidth;
    dy /= _winHeight;
    
    if (_selectedRect >= 0)
    {
        UISoundRect::ParamIndex paramIdx = GetCurrentRectParamIdx();
            
        if (paramIdx == UISoundRect::NO_PARAMETER)
            // Drag could only if we are not dragging a param value
        {
            double x = _soundRects[_selectedRect]->GetUI()->GetX();
            double y = _soundRects[_selectedRect]->GetUI()->GetY();

            x += dx;
            y += dy;
            
            _soundRects[_selectedRect]->GetUI()->UpdateRectPosition(x, y);
        }
    }
}

void
App::UpdateRectPicking(double x, double y)
{
    x /= _winWidth;
    y /= _winHeight;

    int prevRect = _selectedRect;
    
    // Unselect current rect
    UISoundRect *rect = GetCurrentSoundRect();
    if (rect != NULL)
    {
        if (!rect->GetSelectState())
            prevRect = false;
        
        rect->SetSelectState(false);
    }
    
    _selectedRect = -1;

    // Check if we are about to select a source
    int selectedsource = -1;
    for (int i = 0; i < _sources.size(); i++)
    {
        if (_sources[i]->GetUI()->Pick(x, y))
        {
            _selectedSource = i;
            break;
        }
    }

    // Search for rectangle selection only if source is not under the mouse
    if (_selectedSource < 0)
    {
        // Search for selections
        // Set resize direction to horizontal

        int selectedRect = -1;
        for (int i = 0; i < _soundRects.size(); i++)
        {
            // Pick sound rects.
            // Take the already selected one if any.
            if (_soundRects[i]->GetUI()->Pick(x, y) &&
                ((selectedRect == -1) ||
                 (i == prevRect)))
            {                
                selectedRect = i;
            }
        }

        if (selectedRect != -1)
        {
            _selectedRect = selectedRect;
            _soundRects[_selectedRect]->GetUI()->SetSelectState(true);
        }
    }
}

UIScreen *
App::GetScreen() const
{
    return _screen;
}

const UIStyle *
App::GetStyle() const
{
    return _style;
}

bool
App::PickParameters(double x, double y)
{
    Source *source = GetCurrentSource();
    UISoundRect *rect = GetCurrentSoundRect();
    if (source != NULL)
    {
        bool picked = source->GetUI()->PickParameters(x, y);
        return picked;
    }
    else if (rect != NULL)
    {
        bool picked = rect->PickParameters(x, y);
        return picked;
    }

    return false;
}

void
App::SerializeSession()
{
    Serializer serializer;
    bool success = serializer.OpenWriteFile(SESSION_FILENAME);
    if (!success)
        return;

    SerializeSoundRects(&serializer);

    SerializeSources(&serializer);

    serializer.PutDouble(_sampTimeSec);

    serializer.PutDouble(_position._x);
    serializer.PutDouble(_position._y);

    serializer.PutDouble(_zoom);

    serializer.PutInt(_selectionMode);
    serializer.PutInt(_dragMode);

    serializer.PutInt(_selectedSource);
    serializer.PutInt(_selectedRect);

    serializer.PutInt(_winWidth);
    serializer.PutInt(_winHeight);

    serializer.PutBool(_isFullScreen);

    int styleNum = _styleManager->GetCurrentStyleNum();
    serializer.PutInt(styleNum);

    _masterEffectsBox->Serialize(&serializer);
    
    serializer.CloseFile();
}

void
App::SerializeSoundRects(Serializer *serializer)
{
    int size = _soundRects.size();
    serializer->PutInt(size);
    for (int i = 0; i < _soundRects.size(); i++)
        _soundRects[i]->Serialize(serializer);
}

void
App::SerializeSources(Serializer *serializer)
{
    int size = _sources.size();
    serializer->PutInt(size);
    
    for (int i = 0; i < _sources.size(); i++)
    {
        int type = _sources[i]->GetSourceType();
        serializer->PutInt(type);
            
        _sources[i]->Serialize(serializer);
        _sources[i]->GetUI()->Serialize(serializer);
    }
}

void
App::UnSerializeSession()
{
    Serializer serializer;
    bool success = serializer.OpenReadFile(SESSION_FILENAME);
    if (!success)
        return;

    UnSerializeSoundRects(&serializer);
    
    UnSerializeSources(&serializer);
 
    serializer.GetDouble(&_sampTimeSec);

    serializer.GetDouble(&_position._x);
    serializer.GetDouble(&_position._y);

    serializer.GetDouble(&_zoom);

    serializer.GetInt(&_selectionMode);
    serializer.GetInt(&_dragMode);

    serializer.GetInt(&_selectedSource);
    serializer.GetInt(&_selectedRect);

    serializer.GetInt(&_winWidth);
    serializer.GetInt(&_winHeight);

    serializer.GetBool(&_isFullScreen);

    int styleNum;
    serializer.GetInt(&styleNum);
    _styleManager->SetStyleNum(styleNum);
    UpdateStyle();

    _masterEffectsBox->UnSerialize(&serializer);
    
    serializer.CloseFile();
}

void
App::UnSerializeSoundRects(Serializer *serializer)
{
    int size;
    serializer->GetInt(&size);
    for (int i = 0; i < _soundRects.size(); i++)
        _soundRects[i]->UnSerialize(serializer);
}

void
App::UnSerializeSources(Serializer *serializer)
{
    int size;
    serializer->GetInt(&size);
    
    for (int i = 0; i < size; i++)
    {
        Source::SourceType type;
        serializer->GetInt((int *)&type);

        if (type == Source::GRAIN_CLOUD)
            AddGrainCloud(0, 0);
        else if (type == Source::SINGLE_GRAIN)
            AddSingleGrain(0, 0);
        else if (type == Source::LOOP_SOURCE)
            AddLoopSource(0, 0);
        
        _sources[i]->UnSerialize(serializer);
        _sources[i]->GetUI()->UnSerialize(serializer);
    }
}

int
main(int argc, char ** argv)
{
    fprintf(stderr, "usage: %s <midi port #>\n", argv[0]);

    int midiPort = 0;
    if (argc > 1)
        midiPort = atoi(argv[1]);
    
    srand(time(NULL));
    
    glutInit(&argc, argv);
    
    bool appInit = _app.Init(midiPort);
    if (!appInit)
    {
        cout << "Something went wrong...shouldn't be here" << endl;
        
        _app.Clean();
        return 0;
    }
    
    glutMainLoop();
    
    _app.Clean();
    
    return 0;
}
