/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef APP_H
#define APP_H

#include "RtAudio.h"
#include "MyRtAudio.h"
#include "AudioFileSet.h"
#include "Envelope.h"
#include "Thread.h"
#include "SoundRect.h"
#include "UISoundRect.h"
#include "Source.h"
#include "Parameter.h"
#include "ParameterProvider.h"

struct Pt3d
{ 
    Pt3d(GLdouble x, GLdouble y, GLdouble z)
    : _x(x), _y(y), _z(z) {};

    Pt3d()
    : _x(0.0), _y(0.0), _z(0.0) {};
    
    double _x;
    double _y;
    double _z;
};

class AppControl;
class UIScreen;
class UIStyleManager;
class UIStyle;
class UISource;
class UIPieMenu;
class UIWindow;
class UILFOCanvas;
class UIEnvelopeTypeCanvas;
class UIEffectsCanvas;
class Serializer;
class GrainCloud;
class SingleGrain;
class LoopSource;
class MyRtMidiIn;
class MyRtMidiOut;
class MidiQueue;
class EffectsBox;
class App
{
public:
    // Selection mode
    enum
    {
        RECT = 0,
        SOURCE
    };
    
    enum
    {
        MOVE = 0,
        RESIZE
    };
    
    App();
    virtual ~App();
    
    bool Init(int midiPort);
    void Clean();
 
    void WindowInit();

    void SetWindowedMode();
    void ToggleFullScreen();
             
    void Deselect(int mode);
    
    void DrawString(GLdouble x, GLdouble y, GLdouble z,
                    const char * str,
                    GLdouble scale);
    void DrawString2(GLdouble x, GLdouble y, GLdouble z,
                     const char * str,
                     GLdouble scale);
    void DrawTitle();
    void DrawHelp(bool shortHelp);
    
    void DrawParamMeter(GLdouble x, GLdouble y, GLdouble z,
                        double value, double minVal, double maxVal);
        
    void DrawBounceFile();
    void DrawFps();

    void ToggleAudioBounce();

    // Forwarded glut callbacks
    void OnIdle();
    void OnDisplay();
    void OnReshape(int w, int h);
    void OnSpecialKey(int key, int x, int y);
    void OnSpecialKeyUp(int key, int x, int y);
    void OnKeyboard(unsigned char key, int x, int y);
    void OnKeyUp(unsigned char key, int x, int y);
    void OnMouse(int button, int state, int x, int y);
    void OnMouseDrag(int x, int y);
    void OnMousePassiveMotion(int x, int y);
    void OnMouseWheel(int button, int dir, int x, int y);
    
    int OnAudio(void *outputBuffer, void *inputBuffer,
                unsigned int nFrames, double streamTime,
                RtAudioStreamStatus status, void *userData);

    void ToggleDrawHelp(bool flag);

    void UpdateStyle();

    const string GetAudioPath() const;

    void PrevStyle();
    void NextStyle();
    
    double GetPositionX() const;
    void TranslatePositionX(double dx);
    double GetXMoveSpeed() const;
    
    double GetPositionY() const;
    void TranslatePositionY(double dx);
    double GetYMoveSpeed() const;

    void ApplyZoom(double dzoom, double zoomCenterX, double zoomCenterY);
    double GetZoom() const;
    
    bool GetMenuFlag() const;
    void SetMenuFlag(bool flag);

    int GetWinWidth() const;
    int GetWinHeight() const;

    void SetDragMode(int mode);
    int GetDragMode() const;
    
    void OpenPieMenu();
    void ClosePieMenu();
    bool IsPieMenuOpen() const;
    void PieMenuMouseDrag(double dx, double dy);

    void OpenEffectsWindow(bool master);
    void CloseEffectsWindow();
    
    void LeftPickPieMenu(int button, int state, double x, double y);
    void RightPickPieMenu(int button, int state, double x, double y);

    UISoundRect *GetCurrentSoundRect();
    void TranslateSoundRect(double dx, double dy);
    void ScaleSoundRect(double x, double y, double dx, double dy);
    void ToggleSoundRectOrientation();
    void SetSoundRectResizeMode();
    void SetSoundRectMoveMode();
    bool IsSelectedRect() const;
    void ResetSelectedRect();
    void RectMouseDrag(double dx, double dy);
    void UpdateRectPicking(double x, double y);
    
    bool IsEditingParamString() const;
    void AddParamStringChar(char c);
    void DelLastParamStringChar();
    void InsertParamStringChar(int index, const char *c);
    void ApplyParamString();
    void ResetParamString();
    const char *GetParamValueString() const;
    
    int GetCurrentSourceParamIdx() const;
    UISoundRect::ParamIndex GetCurrentRectParamIdx() const;
    
    void ResetCurrentParam();
    void OnCurrentParamValueUpdated();
    void UpdateParamWindows();
    void UpdateParamNoteSources();
    bool PickParameters(double x, double y);

    Source *GetCurrentSource();
    bool IsSelectedSource() const;
    void ResetSelectedSource();
    void SourceMouseDrag(double dx, double dy);
    void UpdateSourcePicking(double x, double y);
    void DelSelectedSource();
    
    void AddGrainCloud(double x, double y);
    void AddSingleGrain(double x, double y);
    void AddLoopSource(double x, double y);

    UIScreen *GetScreen() const;
    const UIStyle *GetStyle() const;

    void SaveCurrentSession();

    ParameterProvider *GetCurrentParamProvider();

    void SendNoteOn(int midiNote);
    void SendNoteOff(int midiNote);
    
    bool IsPlayingNoteSource() const;
    void ApplyVirtualPitchWheel(double dx, double dy);
        
protected:
    void SerializeSession();
    void SerializeSoundRects(Serializer *serializer);
    void SerializeSources(Serializer *serializer);
    
    void UnSerializeSession();
    void UnSerializeSoundRects(Serializer *serializer);
    void UnSerializeSources(Serializer *serializer);

    void ApplyParamStringGrainCloud(GrainCloud *cloud, double value);
    void ApplyParamStringSingleGrain(SingleGrain *grain, double value);
    void ApplyParamStringLoopSource(LoopSource *source, double value);
    
    void UpdateLFOWindow();
    void UpdateEnvelopeWindow();

    void UpdateMidiIn();
    
    MyRtAudio * _audio;

    string _audioPath;

    string _paramValueString;
    bool _isEditingParamString;

    unsigned int _blockSize;
    
    vector <AudioFile *> _sounds;

    vector<SoundRect *> _soundRects;
    vector<Source *> _sources;
    vector<vector<Source *> > _noteSources;
    
    double _sampTimeSec;
    
    Pt3d _position;
    
    double _zoom;
    
    int _selectionMode;
    int _dragMode;

    int _selectedSource;
    int _selectedRect;
    bool _menuFlag;
    
    unsigned int _screenWidth;
    unsigned int _screenHeight;

    int _winWidth;
    int _winHeight;

    bool _isFullScreen;

    UIScreen *_screen;
    UIStyleManager *_styleManager;
    UIStyle *_style;

    UIPieMenu *_pieMenu;
    
    UIWindow *_grainLFOWindow;
    UILFOCanvas *_grainLFOCanvas;

    UIWindow *_envelopeTypeWindow;
    UIEnvelopeTypeCanvas *_envelopeTypeCanvas;

    UIWindow *_effectsWindow;
    UIEffectsCanvas *_effectsCanvas;
    
    double _xMoveSpeed;
    double _yMoveSpeed;

    double _virtualPitch;
    
    bool _audioBounce;
    FILE *_bounceFile;
    char _bounceFileName[255];
    
    int _fps;

    bool _displayHelp;
    
    AppControl *_control;

    AudioFileSet *_fileSet;

    MyRtMidiIn *_midiIn;
    MidiQueue *_midiQueue;

    MyRtMidiOut *_midiOut;

    EffectsBox *_masterEffectsBox;

    vector<SAMPLE> _tmpBuf;
    
    Mutex *_mutex;
};

#endif
