/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "App.h"
#include "Source.h"
#include "AppControl.h"
#include "KeyMap.h"
#include "UIPieMenu.h"
#include "UIStyleManager.h"
#include "UIStyle.h"

#define FINE_DRAG_FACTOR 0.1

static const char *_commands[] =
{
    "--------------------------------------------",
    "Esc + Shift      Quit",
    "F1 key           Display help over the track",
    "F5               Toggle fullscreen",
    "F10 key          Hide/Show Title Screen",
    "F11              Toggle audio bounce",
    "F12              Save current session",
    " ",
    
    "Rectangles",
    "------------",
    "Left click       Select",
    "Drag             Move",
    "2/~ key + drag   Resize",
    " ",
    
    "Source Addition and Selection",
    "------------",
    "1 key            Add cloud",
    "2 key            Add single grain",
    "3 key            Add loop source",
    "Delete + Shift   Remove selected source",
    "Left click       Select",
    "Drag             Move",
    " ",
    
    "------------",
    "Spacebar        Pie Menu",
    "Spacebar + Alt  Effects menu",
    "Right + Pie     Enter Param Edit Mode",
    "Ctrl + arrows   Switch Style",
    " ",
    
    "------------",
    "Ctrl + Drag      Drag Background",
    "Middle + drag    Drag Background",
    "Alt + Drag       Zoom",
    "Middle           Zoom",
    "--------------------------------------------"
};

static const char *_commandsShort[] =
{
    "------------------------------------",
    "Esc + Shift   Quit",
    "F1            Help",
    "F5            Fullscreen",
    "F10           Title Screen",
    "F11           Bounce Audio",
    "F12           Save Session",
    " ",
    
    "Rectangles",
    "------------",
    "Left Click    Select",
    "Drag          Move",
    "2/~           Resize",
    " ",
    
    "Cloud Addition + Selection",
    "------------",
    "1             Add Cloud",
    "2             Add Single Grain",
    "3             Add Loop Source",
    "Del + Shift   Remove Source",
    "Left          Select Source",
    "Drag          Move Source",
    " ",
    
    "------------",
    "Space          Pie Menu",
    "Space + Alt    Effects Menu",
    "Right + Pie    Enter Param Edit Mode",
    "Ctrl + Arrows  Switch Style",
    " ",
    
    "------------",
    "Ctrl + Drag    Drag Background",
    "Middle + drag  Drag Background",
    "Alt + Drag     Zoom",
    "Middle         Zoom",
    "------------------------------------"
};

AppControl::AppControl(App *app)
{
    _app = app;

    _leftMouseDown = false;
    _rightMouseDown = false;
    _middleMouseDown = false;

    _isDraggingBackground = false;
    _ctrlDown = false;

    _zoomCenterX = 0.0;
    _zoomCenterY = 0.0;
    _altPressed = false;
    _spacePressed = false;
    _shiftPressed = false;
    
    _mouseX = -1;
    _mouseY = -1;
    _veryHighNumber = 50000000;
    _lastDragX = _veryHighNumber;
    _lastDragY = _veryHighNumber;
    
    _lastMouseX = -1;
    _lastMouseY = -1;
}

AppControl::~AppControl() {}

const char **
AppControl::GetCommands() const
{
    return _commands;
}

int
AppControl::GetNumCommands() const
{
    int numCommands = sizeof(_commands)/sizeof(char *);
    
    return numCommands;
}

const char **
AppControl::GetCommandsShort() const
{
    return _commandsShort;
}

int
AppControl::GetNumCommandsShort() const
{
    int numCommands = sizeof(_commandsShort)/sizeof(char *);
    
    return numCommands;
}

void
AppControl::OnSpecialKey(int key, int x, int y)
{
    int modkey = glutGetModifiers();

    _app->SetMenuFlag(false);
        
    switch (key)
    {    
        case GLUT_KEY_LEFT:
            if (modkey == 0)
            {
                _app->TranslatePositionX(-_app->GetXMoveSpeed());
                _mouseX -= _app->GetXMoveSpeed();
            }
            else if ((modkey & GLUT_ACTIVE_CTRL) != 0)
            {
                _app->PrevStyle();
            }
            break;
            
        case GLUT_KEY_RIGHT:
            if (modkey == 0)
            {
                _app->TranslatePositionX(+_app->GetXMoveSpeed());
                _mouseX += _app->GetXMoveSpeed();
            }
            else if ((modkey & GLUT_ACTIVE_CTRL) != 0)
            {
                _app->NextStyle();
            }
            break;
            
        case GLUT_KEY_DOWN:
            _app->TranslatePositionY(-_app->GetYMoveSpeed());
            _mouseY += _app->GetYMoveSpeed();
            break;
            
        case GLUT_KEY_UP:
            _app->TranslatePositionY(_app->GetYMoveSpeed());
            _mouseY -= _app->GetYMoveSpeed();
            break;

        case GLUT_KEY_F1:
            _app->ToggleDrawHelp(true);
            break;

        case GLUT_KEY_F5:
            _app->ToggleFullScreen();
            break;

        case GLUT_KEY_F10:
        {
            bool menuFlag = _app->GetMenuFlag();
            menuFlag = !menuFlag;
            _app->SetMenuFlag(menuFlag);
        }
        break;
        
        case GLUT_KEY_F11:
            _app->ToggleAudioBounce();
            break;

        case GLUT_KEY_F12:
            _app->SaveCurrentSession();
            break;
            
        case 114: // Left Ctrl
        case 115: // Right Ctrl
        {
            _isDraggingBackground = true;
            _ctrlDown = true;
        }
        break;

        case 116: // Alt
        {
            _altPressed = true;

            _zoomCenterX = x;
            _zoomCenterY = _app->GetWinHeight() - y;
            
            _lastMouseX = x;
            _lastMouseY = y;
        }
        break;

        case 112: // Shift
            _shiftPressed = true;
            break;
                
        default:
            break;
    }
    
    glutPostRedisplay();
}

void
AppControl::OnSpecialKeyUp(int key, int x, int y)
{
    switch (key)
    {
        case GLUT_KEY_F1:
            _app->ToggleDrawHelp(false);
            break;
            
        case 114: // Left Ctrl
        case 115: // Right Ctrl
        {
            _isDraggingBackground = false;
            _ctrlDown = false;
        }
        break;

        case 116: // Alt
            _altPressed = false;
            break;

        case 112: // shift
            _shiftPressed = false;
            break;
            
        default:
            break;
    }
}

void
AppControl::OnKeyUp(unsigned char key, int x, int y)
{
    switch(key)
    {            
        case 178: // ²/~
        {
            _app->SetSoundRectMoveMode();
            _lastDragX = _veryHighNumber;
            _lastDragY = _veryHighNumber;
        }
        break;

        case ' ':
        {
            _spacePressed = false;
            
            _app->ClosePieMenu();
            _app->CloseEffectsWindow();
        }
        break;
            
        default:
            break;
    }

    CheckMidiNoteKey(key, false);
}

void
AppControl::OnKeyboard(unsigned char key, int x, int y)
{
    int modkey = glutGetModifiers();

    _app->SetMenuFlag(false);
        
    switch(key)
    {
        case 178: // ²/~
            _app->SetSoundRectResizeMode();
            break;
            
        case '1':
        case '&':
        {
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('1');
            else
            {
                _app->ResetParamString();
                _app->Deselect(App::RECT);
                
                _app->AddGrainCloud(_mouseX, _mouseY);
            }
        }
        break;
                
        case '2':
            //case 'é':
        {
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('2');
            else
            {
                _app->ResetParamString();
                _app->Deselect(App::RECT);
                
                _app->AddSingleGrain(_mouseX, _mouseY);
            }
        }
        break;
            
        case '3':
        case '"':
        {
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('3');
            else
            {
                _app->ResetParamString();
                _app->Deselect(App::RECT);
                
                _app->AddLoopSource(_mouseX, _mouseY);
            }
        }
        break;
            
        case '4':
        {
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('4');
        }
        break;
            
        case'5':
        {
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('5');
        }   
        break;

        case'6':
        {
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('6');
        }      
        break;
            
        case'7':
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('7');
            break;

        case'8':
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('8');
            break;

        case'9':
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('9');
            break;

        case '0':
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('0');
            break;

        case '.':
        case ';':
            if (_app->IsEditingParamString())
                _app->AddParamStringChar('.');
            break;
            
        case '-':
        case '_':
            if (_app->IsEditingParamString())
                _app->InsertParamStringChar(0, "-");
            break;
            
        case 13: // Enter
        {
            if (_app->IsEditingParamString())
                _app->ApplyParamString();

            _app->ClosePieMenu();
        }
        break;
            
        case 27: // Esc
        {
            if ((modkey & GLUT_ACTIVE_SHIFT) != 0)
            {
                _app->Clean();
                exit(1);
            }

            if (_app->IsEditingParamString())
            {
                _app->ResetParamString();
                _app->ClosePieMenu();
            }
        }
        break;    
            
        case 127: // Delete
        case 8: // Delete (tested on mac keyboard)
        {
            if (!_app->IsEditingParamString())
                _app->DelSelectedSource();
            else
                _app->DelLastParamStringChar();

#if 0 // With this, may have the risk to delete the source by mistake
            if (_app->_isEditingParamString &&
                (_app->_paramValueString == ""))
            {
                _app->_currentParam = -1;
                
                _app->_isEditingParamString = false;
            }
#endif
        }    
        break;

        case ' ':
        {
            _spacePressed = true;
            
            if ((modkey & GLUT_ACTIVE_ALT) == 0)
            {
                Source *source = _app->GetCurrentSource();
                UISoundRect *rect = _app->GetCurrentSoundRect();
                if ((source != NULL) || (rect != NULL))
                    _app->OpenPieMenu();
            }
            else
            {
                bool masterEffects = !_app->IsSelectedSource();
                _app->OpenEffectsWindow(masterEffects);
            }
        }
        break;
    
        default:
            break;
    }

    CheckMidiNoteKey(key, true);
    
    glutPostRedisplay();
}

void
AppControl::OnMouse(int button, int state, int x, int y)
{
    if ((button == GLUT_LEFT_BUTTON) &&
        (state == GLUT_DOWN))
        _leftMouseDown = true;
    if ((button == GLUT_RIGHT_BUTTON) &&
        (state == GLUT_DOWN))
        _rightMouseDown = true;
    if ((button == GLUT_MIDDLE_BUTTON) &&
        (state == GLUT_DOWN))
        _middleMouseDown = true;

    if ((button == GLUT_LEFT_BUTTON) &&
        (state == GLUT_UP))
        _leftMouseDown = false;
    if ((button == GLUT_RIGHT_BUTTON) &&
        (state == GLUT_UP))
        _rightMouseDown = false;
    if ((button == GLUT_MIDDLE_BUTTON) &&
        (state == GLUT_UP))
        _middleMouseDown = false;

    int modkey = glutGetModifiers();
    if ((modkey & GLUT_ACTIVE_ALT) == 0)
        _altPressed = false;
            
    if (_rightMouseDown)
    {
        if (_app->IsEditingParamString())
        {
            _app->ApplyParamString();
            
            _app->ClosePieMenu();
        }
    }
    
    if (_app->IsEditingParamString())
        return;
    
    if (_rightMouseDown)
    {
        if (_app->IsPieMenuOpen())
        {
            _app->RightPickPieMenu(button, state, _mouseX, _mouseY);
            
            return;
        }
    }
    
    UpdateMouseCoords(x, y);

    if (_leftMouseDown || _middleMouseDown)
    {
        _lastMouseX = x;
        _lastMouseY = y;
    }

    // Zoom with middle button
    _isDraggingBackground = (_middleMouseDown || _ctrlDown);
    
    if (_isDraggingBackground)
        return;

    // Update zoom center if mouse button is just pressed
    // and zoom key previously pressed.
    if (_altPressed && !_spacePressed)
    {
        if (_leftMouseDown)
        {
            _zoomCenterX = x;
            _zoomCenterY = _app->GetWinHeight() - y;
            
            _lastMouseX = x;
            _lastMouseY = y;
        }
    }

    if (_leftMouseDown && _app->IsPieMenuOpen())
    {    
        _app->LeftPickPieMenu(button, state, _mouseX, _mouseY);

        _lastDragX = _mouseX;
        _lastDragY = _mouseY;
    
        return;
    }
    
    if (_leftMouseDown)
    {
        _app->SetMenuFlag(false);
        
        _app->ResetCurrentParam();

        _app->Deselect(App::SOURCE);

        _lastDragX = _mouseX;
        _lastDragY = _mouseY;
                
        _app->UpdateSourcePicking(_mouseX, _mouseY);
        _app->UpdateRectPicking(_mouseX, _mouseY);
    }
}

void
AppControl::OnMouseDrag(int x, int y)
{    
    if (!_leftMouseDown && !_rightMouseDown && !_middleMouseDown)
        return;
        
    if (_app->IsEditingParamString())
        return;
        
    UpdateMouseCoords(x, y);

    if (_lastDragX == _veryHighNumber)
        _lastDragX = _mouseX;
    if (_lastDragY == _veryHighNumber)
        _lastDragY = _mouseY;
                   
    double dx = _mouseX - _lastDragX;
    double dy = _mouseY - _lastDragY;

    double fineDragFactor = GetFineDragFactor();
    dx *= fineDragFactor;
    dy *= fineDragFactor;
            
    _lastDragX = _mouseX;
    _lastDragY = _mouseY;
        
    if (_isDraggingBackground)
    {
        double dx = x - _lastMouseX;
        double dy = y - _lastMouseY;

        if (_shiftPressed)
        {
            dx *= FINE_DRAG_FACTOR;            
            dy *= FINE_DRAG_FACTOR;
        }
        
        _lastMouseX = x;
        _lastMouseY = y;
    
        // Must use a trick, because _mouseX depends on _position.x
        dy = -dy;

        _app->TranslatePositionX(-dx);
        _app->TranslatePositionY(-dy);
        
        return;
    }

    if (_altPressed && !_spacePressed)
    {
        // Update zoom
#define DRAG_ZOOM_COEFF 0.005
        double dm = dy;
        dm = -dm;

        _app->ApplyZoom(1.0 + dm*DRAG_ZOOM_COEFF,
                        _zoomCenterX, _zoomCenterY);

        return;
    }

    if (_leftMouseDown)
    {
        if (_app->IsPieMenuOpen())    
            _app->PieMenuMouseDrag(dx, dy);
    }
    
    if (_app->IsSelectedSource())
        _app->SourceMouseDrag(dx, dy);
    else if (_app->IsSelectedRect())
    {
        _app->RectMouseDrag(dx, dy);

        switch (_app->GetDragMode())
        {
          case App::RESIZE:
            if((_lastDragX != _veryHighNumber) &&
               (_lastDragY != _veryHighNumber))
              _app->ScaleSoundRect(_mouseX, _mouseY, dx, dy);
            break;
            
          default:
            break;
        }
    }
}

void
AppControl::OnMouseWheel(int button, int dir, int x, int y)
{
    if (_app->IsEditingParamString())
        return;
    
#define MOUSE_WHEEL_COEFF 0.1

    _zoomCenterX = x;
    _zoomCenterY = _app->GetWinHeight() - y;
    
    _app->ApplyZoom(1.0 + dir*MOUSE_WHEEL_COEFF,
                    _zoomCenterX, _zoomCenterY);
}

double
AppControl::GetFineDragFactor() const
{
    if (!_shiftPressed)
        return 1.0;
    
    return FINE_DRAG_FACTOR;
}

void
AppControl::OnMousePassiveMotion(int x, int y)
{
    UpdateMouseCoords(x,y);

    double dx = _mouseX - _lastDragX;
    double dy = _mouseY - _lastDragY;
    
    _lastDragX = _mouseX;
    _lastDragY = _mouseY;

    if (_app->IsPlayingNoteSource())
        // If playing notes from keyboard or midi device
    {
        // Use the computer mouse or trackpad to emulate pitch wheel
        _app->ApplyVirtualPitchWheel(dx, dy);
    }
}

void
AppControl::UpdateMouseCoords(int x, int y)
{
    _mouseX = x + _app->GetPositionX();
    _mouseY = (_app->GetWinHeight() - (y - _app->GetPositionY()));

    // Take the zoom into account
    double zoom = _app->GetZoom();
    _mouseX /= zoom;
    _mouseY /= zoom;
}

// 4th or 5th consistency between keyboard rows
void
AppControl::CheckMidiNoteKey(unsigned char key, bool keyDown)
{
    int intervalHt = 0;
    bool invalid = false;
    switch (key)
    {
        case 'w':
        case 'W':
            intervalHt = -12;
            break;
        case 'x':
        case 'X':
            intervalHt = -11;
            break;
        case 'c':
        case 'C':
            intervalHt = -10;
            break;
        case 'v':
        case 'V':
            intervalHt = -9;
            break;
        case 'b':
        case 'B':
            intervalHt = -8;
            break;
        case 'n':
        case 'N':
            intervalHt = -7;
            break;
            
        case 'q':
        case 'Q':
            intervalHt = -7;
            break;
        case 's':
        case 'S':
            intervalHt = -6;
            break;
        case 'd':
        case 'D':
            intervalHt = -5;
            break;
        case 'f':
        case 'F':
            intervalHt = -4;
            break;
        case 'g':
        case 'G':
            intervalHt = -3;
            break;
        case 'h':
        case 'H':
            intervalHt = -2;
            break;
        case 'j':
        case 'J':
            intervalHt = -1;
            break;
        case 'k':
        case 'K':
            intervalHt = 0;
            break;
        case 'l':
        case 'L':
            intervalHt = 1;
            break;
        case 'm':
        case 'M':
            intervalHt = 2;
            break;

            // Reference
        case 'a':
        case 'A':
            intervalHt = 0;
            break;
        case 'z':
        case 'Z':
            intervalHt = 1;
            break;
        case 'e':
        case 'E':
            intervalHt = 2;
            break;
        case 'r':
        case 'R':
            intervalHt = 3;
            break;
        case 't':
        case 'T':
            intervalHt = 4;
            break;
        case 'y':
        case 'Y':
            intervalHt = 5;
            break;
        case 'u':
        case 'U':
            intervalHt = 6;
            break;
        case 'i':
        case 'I':
            intervalHt = 7;
            break;
        case 'o':
        case 'O':
            intervalHt = 8;
            break;
        case 'p':
        case 'P':
            intervalHt = 9;
            break;

        default:
            invalid = true;
            break;
    }

    // Use 'a' as the C5 reference
    int midiNote = intervalHt + 72;
    
    if (!invalid)
    {
        if (keyDown)
            _app->SendNoteOn(midiNote);
        else
            _app->SendNoteOff(midiNote);
    }
}
