/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <string.h>

#include "Stk.h"
#include "DSP.h"
#include "Stereoizer.h"
#include "AudioFile.h"
#include "AudioFileSet.h"

AudioFileSet::AudioFileSet()
{
    _stereoizer = new Stereoizer();
}

AudioFileSet::~AudioFileSet()
{
    for (int i = 0; i < _fileSet.size(); i++)
        delete _fileSet[i];

    delete _stereoizer;
}

const vector<AudioFile *> &
AudioFileSet::GetAudioFiles()
{
    return _fileSet;
}

int
AudioFileSet::Load(string localPath)
{
    struct dirent *ent;
    int numFiles = 0;
    
    DIR *dir = opendir(localPath.c_str());   
    if (dir != NULL)
    {    
        while ((ent = readdir(dir)) != NULL)
        {
            string fileName = ent->d_name;
            if ((fileName == ".") || (fileName == "..") || (fileName == ".DS_Store"))
                continue;
            
            string path = localPath + fileName;
            
            SF_INFO sfinfo;
            SNDFILE *infile;
            if (!(infile = sf_open(path.c_str(), SFM_READ, &sfinfo)))
            {
                fprintf(stderr, "Not able to open input file %s.\n", fileName.c_str());

                puts(sf_strerror(NULL));
                
                continue;
            }
            
            fprintf(stderr, "Loading '%s'... \n", fileName.c_str());
            
            cout << "  Channels: " << sfinfo.channels << endl;
            cout << "  Frames: " << sfinfo.frames << endl;
            cout << "  Sample Rate: " << sfinfo.samplerate << endl;
            cout << "  Format: " << sfinfo.format << endl;
            
            if (sfinfo.samplerate != SRATE)
            {
                fprintf(stderr, "\nWARNING: '%s' is sampled at a different rate from \
the current sample rate of %i\n", fileName.c_str(), SRATE);
            }
            
            const int bufferSize = 512;
            
            SAMPLE stereoBuff[bufferSize];
            
            unsigned long fullSize = sfinfo.frames * sfinfo.channels;
            vector<SAMPLE> samples;
            samples.resize(fullSize);
            _fileSet.push_back(new AudioFile(fileName,
                                             path,
                                             sfinfo.channels,
                                             sfinfo.frames,
                                             sfinfo.samplerate,
                                             samples));

            unsigned long totalCount = 0;
            bool empty = false;
            do
            {
#if USE_SAMPLES_DOUBLE
                sf_count_t count =
                    sf_read_double(infile, &stereoBuff[0], bufferSize);
#else
                sf_count_t count =
                    sf_read_float(infile, &stereoBuff[0], bufferSize);
#endif
                
                for(int i = 0; i < bufferSize; i++)
                {
                    if (totalCount < fullSize)
                    {
                        vector<SAMPLE> *afBuf =
                            ((vector<SAMPLE> *)_fileSet[numFiles]->GetBuffer());
                        (*afBuf)[totalCount] = stereoBuff[i]*ATTENUATION;

                        totalCount++;
                    }     
                }
                
                if (count == 0) 
                {
                    empty = true;
                    
                    continue;
                }
                
            } while(!empty);

            cout << totalCount << endl;
            numFiles++;
            
            sf_close(infile);
        }
        
        closedir(dir);
    }
    else
    {
        perror("");
        return 1;
    }

    return 0;
}

void
AudioFileSet::Stereoize()
{
    for (int i = 0; i < _fileSet.size(); i++)
    {
        AudioFile *audioFile = _fileSet[i];

        if (audioFile->GetNumChannels() == 2)
        {
#if STEREOIZE_BI_MONO
            vector<SAMPLE> *stereoBuffer = (vector<SAMPLE> *)audioFile->GetBuffer();
            int nFrames = audioFile->GetNumFrames();

            SAMPLE diff = DSP::CmpChannels(*stereoBuffer);

            if (diff > 1e-4)
                // Almost true stereo
                continue;

            // Almost bi-mono
            vector<SAMPLE> monoL;
            monoL.resize(nFrames);
                         
            vector<SAMPLE> monoR;
            monoR.resize(nFrames);
                         
            DSP::DeInterleave(*stereoBuffer, &monoL, &monoR);

            vector<SAMPLE> l;
            vector<SAMPLE> r;
            _stereoizer->Reset();
            _stereoizer->Process(monoL, &l, &r);

            DSP::Interleave(stereoBuffer, l, r);
#endif
            
            continue;
        }

        // True mono
        int nFrames = audioFile->GetNumFrames();
        vector<SAMPLE> stereoBuffer;
        stereoBuffer.resize(nFrames*2);
        AudioFile *stereoFile = new AudioFile(audioFile->GetName(),
                                              audioFile->GetPath(),
                                              2, nFrames,
                                              audioFile->GetSampleRate(),
                                              stereoBuffer);

        vector<SAMPLE> mono;
        mono.resize(nFrames);
        memcpy(mono.data(), audioFile->GetBuffer()->data(), mono.size()*sizeof(SAMPLE));

        vector<SAMPLE> l;
        vector<SAMPLE> r;
        _stereoizer->Reset();
        _stereoizer->Process(mono, &l, &r);

        vector<SAMPLE> *afStereoBuffer = (vector<SAMPLE> *)audioFile->GetBuffer();
        DSP::Interleave(afStereoBuffer, l, r);
        
        delete audioFile;
        _fileSet[i] = stereoFile;
    }
}

void
AudioFileSet::Normalize()
{
    for (int i = 0; i < _fileSet.size(); i++)
    {
        AudioFile *audioFile = _fileSet[i];

        vector<SAMPLE> *buffer = (vector<SAMPLE> *)audioFile->GetBuffer();
        int size = audioFile->GetNumSamples();
        DSP::AutomaticClipping(buffer);
        DSP::Normalize(buffer);
    }
}

void
AudioFileSet::Fade()
{
#define NUM_SAMPLES_FADE 128
    
    for (int i = 0; i < _fileSet.size(); i++)
    {
        AudioFile *audioFile = _fileSet[i];

        vector<SAMPLE> *buffer = (vector<SAMPLE> *)audioFile->GetBuffer();
        int nChan = audioFile->GetNumChannels();
        
        DSP::FadeIn(buffer, nChan, NUM_SAMPLES_FADE);
        DSP::FadeOut(buffer, nChan, NUM_SAMPLES_FADE);
    }
}
