/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "DSP.h"
#include "Envelope.h"
#include "SamplesProvider.h"
#include "AudioSource.h"

AudioSource::AudioSource(const vector<SamplesProvider *> &sampProviders,
                         Envelope *envelope,
                         double durationMs,
                         double pitch, double timeStretch)
{
    _isClone = false;
    
    _sampProviders.resize(sampProviders.size());
    for (int i = 0; i < _sampProviders.size(); i++)
        _sampProviders[i] = sampProviders[i]->Clone();
    
    if (_sampProviders.size() > 0)
    {
        _positions.resize(_sampProviders.size());
        _startPositions.resize(_sampProviders.size());
        _endPositions.resize(_sampProviders.size());
        
        _paramY.resize(_sampProviders.size());

        _prevSamps[0].resize(_sampProviders.size());
        _prevSamps[1].resize(_sampProviders.size());
        
        for (int i = 0; i < sampProviders.size(); i++)
        {
            _positions[i] = 0.0;
            _startPositions[i] = 0.0;
            _endPositions[i] = 0.0;
            
            _paramY[i] = 0.0;

            _prevSamps[0][i] = 0.0;
            _prevSamps[1][i] = 0.0;
        }
    }
    
    _isPlaying = false;

    _loop = false;
    
    if (randf() < 0.5)
        _direction = 1.0;
    else
        _direction = -1.0;
    _queuedDirection = _direction;

    _envelope = envelope;
    
    _envelopeType = Envelope::HANNING;
    _queuedEnvelopeType = _envelopeType;

    _envelopeSamples = _envelope->GetEnvelope(_envelopeType);

    _range = 10.0;
    
    _localAtten = 1.0;
    _queuedLocalAtten = _localAtten;
    
    _durationMs = durationMs;
    _queuedDurationMs = durationMs; 
    
    _pitch = pitch;
    _queuedPitch = _pitch;

    _timeStretch = timeStretch;
    _queuedTimeStretch = timeStretch;
    
    _queuedChanMults.resize(CHANNELS);
    _chanMults.resize(CHANNELS);
    
    for (int i = 0; i < CHANNELS; i++)
    {
        _chanMults[i] = 1.0;
        _queuedChanMults[i] = 1.0;
    }
    
    _newParam = false;
    
    _step = _pitch*_direction;
    
    _envPos = 0.0;
    
    _durationSamps = ceil(_durationMs*SRATE*0.001);
    _envStep = (double)ENVELOPE_LEN/_durationSamps; 
}

AudioSource::~AudioSource()
{
    for (int i = 0; i < _sampProviders.size(); i++)
    {
        SamplesProvider *provider = _sampProviders[i];
        delete provider;
    }
}

AudioSource *
AudioSource::Clone() const
{
    AudioSource *clone = new AudioSource(_sampProviders, _envelope, _durationMs,
                                         _pitch, _timeStretch);
    clone->_isClone = true;
    
    clone->_loop = _loop;
    
    clone->_isPlaying = _isPlaying;
    
    clone->_newParam = _newParam;
    
    clone->_durationMs = _durationMs;;
    clone->_queuedDurationMs = _queuedDurationMs;
    clone->_durationSamps = _durationSamps;
    clone->_pitch = _pitch;
    clone->_queuedPitch = _queuedPitch;
    clone->_timeStretch = _timeStretch;
    clone->_queuedTimeStretch = _queuedTimeStretch;
    clone->_direction = _direction;
    clone->_queuedDirection = _queuedDirection; 
    clone->_step = _step;
    
    clone->_localAtten = _localAtten;
    clone->_queuedLocalAtten = _queuedLocalAtten;
    
    clone->_chanMults = _chanMults;
    clone->_queuedChanMults = _queuedChanMults;
    
    clone->_activeSampProviders = _activeSampProviders;
    
    clone->_envelopeType = _envelopeType;
    clone->_queuedEnvelopeType = _queuedEnvelopeType;
    
    clone->_envStep = _envStep;
    clone->_envPos = _envPos;

    clone->_envelope = _envelope;
    clone->_envelopeSamples = _envelopeSamples;
    
    clone->_paramY = _paramY;

    clone->_range = _range;
    
    if (clone->_sampProviders.size() > 0)
        clone->_positions = _positions;

    if (clone->_sampProviders.size() > 0)
        clone->_startPositions = _startPositions;

    if (clone->_sampProviders.size() > 0)
        clone->_endPositions = _endPositions;

    if (clone->_prevSamps[0].size() > 0)
        clone->_prevSamps[0] = _prevSamps[0];
    if (clone->_prevSamps[1].size() > 0)
        clone->_prevSamps[1] = _prevSamps[1];
    
    return clone;
}

const vector<SamplesProvider *> &
AudioSource::GetSampleProviders() const
{
    return _sampProviders;
}

void
AudioSource::SetLoop(bool loop)
{
    _loop = loop;
}

bool
AudioSource::Play(double *startPositions,
                  double *startParamY)
{
    if (!_isPlaying)
    { 
        _isPlaying = true;
        
        if (_newParam)
            UpdateParams();
        
        _activeSampProviders.clear();
        for (int i = 0; i < _sampProviders.size(); i++)
        {
            if (startPositions[i] != -1)
            {
                _activeSampProviders.push_back(i);

                _startPositions[i] =
                    floor(startPositions[i]*_sampProviders[i]->GetNumFrames() - 1);
            
                int channels = _sampProviders[i]->GetNumChannels();
                double durFrames = _durationSamps/channels;
                _endPositions[i] = _startPositions[i] + durFrames;
                if (_endPositions[i] > _sampProviders[i]->GetNumFrames())
                    _endPositions[i] = _sampProviders[i]->GetNumFrames();
                
                if (_direction > 0.0)
                    _positions[i] = _startPositions[i];
                else
                    _positions[i] = _endPositions[i];
                
                _paramY[i] = startParamY[i];

                _prevSamps[0][i] = 0.0;
                _prevSamps[1][i] = 0.0;
            }
        }
        
        _envPos = 0;
        
        return true;    
    }
    
    return false;
}

void
AudioSource::Update(double *startPositions,
                    double *startParamY)
{
    if (_newParam)
        UpdateParams();
    
    _activeSampProviders.clear();
        
    for (int i = 0; i < _sampProviders.size(); i++)
    {
        if (startPositions[i] != -1)
        {
            _activeSampProviders.push_back(i);

            _startPositions[i] =
                floor(startPositions[i]*_sampProviders[i]->GetNumFrames() - 1);

            int channels = _sampProviders[i]->GetNumChannels();
            double durFrames = _durationSamps/channels;
            _endPositions[i] = _startPositions[i] + durFrames;
            if (_endPositions[i] > _sampProviders[i]->GetNumFrames())
                _endPositions[i] = _sampProviders[i]->GetNumFrames();
            
            _paramY[i] = startParamY[i];
        }
    }
}

bool
AudioSource::IsPlaying()
{
    return _isPlaying;
}     

void
AudioSource::SetChannelMultipliers(const vector<SAMPLE> &multipliers)
{
    for (int i = 0; i < CHANNELS; i++)
        _queuedChanMults[i] = multipliers[i];
    
    _newParam = true;
}

void
AudioSource::SetVolume(double normVolume)
{
    double pVol = fabs(normVolume);
    _queuedLocalAtten = pVol;
}

double
AudioSource::GetVolume()
{
    return _localAtten;
}

void
AudioSource::SetDurationMs(double dur)
{
    _queuedDurationMs = fabs(dur);
    if (_queuedDurationMs != _durationMs)
        _newParam = true;
}     

void
AudioSource::SetPitch(double pitch)
{
    _queuedPitch = pitch;
    if (_queuedPitch != _pitch)
        _newParam = true;
}     

double
AudioSource::GetPitch()
{
    if (_queuedPitch != _pitch)
        return _queuedPitch;
    else
        return _pitch;
}

void
AudioSource::SetTimeStretch(double timeStretch)
{
    _queuedTimeStretch = timeStretch;
    if (_queuedTimeStretch != _timeStretch)
        _newParam = true;
}     

double
AudioSource::GetTimeStretch()
{
    if (_queuedTimeStretch != _timeStretch)
        return _queuedTimeStretch;
    else
        return _timeStretch;
}

void
AudioSource::UpdateParams()
{
    _localAtten = _queuedLocalAtten;
    
    _pitch = _queuedPitch;
    _timeStretch = _queuedTimeStretch;
    
    _durationMs = _queuedDurationMs;
    _durationSamps = ceil(_durationMs*SRATE*0.001*CHANNELS);
    _envStep = (double)ENVELOPE_LEN/_durationSamps;
    
    _direction = _queuedDirection;
    
    _step = _direction*_pitch;
    
    _envelopeType = _queuedEnvelopeType;
    _envelopeSamples = _envelope->GetEnvelope(_envelopeType);
    
    for (int i = 0; i < CHANNELS; i++)
        _chanMults[i] = _queuedChanMults[i];
    
    _newParam = false;
}     

void
AudioSource::SetEnvelopeType(unsigned int type)
{
    _queuedEnvelopeType = type;
    if (_queuedEnvelopeType != _envelopeType)
        _newParam = true;
}

void
AudioSource::SetRange(double range)
{
    _range = range;
}

void
AudioSource::SetDirection(double dir)
{
    _queuedDirection = dir;
    if (_queuedDirection != _direction)
        _newParam = true;
}

bool
AudioSource::NextBuffer(vector<SAMPLE> *outBuf,
                        unsigned int numFrames,
                        unsigned int bufOffset)
{
    bool newLoopCycle = false;
    
    if (!_isPlaying)
        return newLoopCycle;

    double numFrames0 = (int)(numFrames*fabs(_step));
    double step = numFrames0/numFrames;
    
    vector<vector<SAMPLE> > buffers;
    buffers.resize(_sampProviders.size());
    vector<double> initBufPos;
    initBufPos.resize(buffers.size());
    for (int i = 0; i < _activeSampProviders.size(); i++)
    {
        int provIdx = _activeSampProviders[i];
        
        double paramY = _paramY[provIdx];
        bool loop = GetSamplesFromProvider(_sampProviders[provIdx], paramY,
                                           &buffers[provIdx], numFrames*step,
                                           _positions[provIdx],
                                           _startPositions[provIdx],
                                           _endPositions[provIdx]);
        
        if (loop)
        {            
            newLoopCycle = true;
            
            if (_direction > 0.0)
                _positions[provIdx] = _startPositions[provIdx];
            else if (_direction < 0.0)
            {
                int blockMinSamples = _sampProviders[provIdx]->GetBlockMinSamples();
                
                _positions[provIdx] =
                    _endPositions[provIdx] - numFrames*step - 1 - blockMinSamples;
            }

            GetSamplesFromProvider(_sampProviders[provIdx], paramY,
                                   &buffers[provIdx], numFrames*step,
                                   _positions[provIdx],
                                   _startPositions[provIdx],
                                   _endPositions[provIdx]);
        }

        initBufPos[provIdx] = _positions[provIdx];
    }

    for (int i = 0; i < numFrames; i++)
    {
        bool envLoop;
        double envCoeff = GetEnvelopeCoeff(&envLoop);
        if (envLoop)
            return newLoopCycle;
        
        SAMPLE left = 0.0;
        SAMPLE right = 0.0;
        
        vector<int> sampProvidersToKeep;
        for (int j = 0; j < _activeSampProviders.size(); j++)
        {    
            int provIdx = _activeSampProviders[j];
            double pos = _positions[provIdx];
                
            int channels = _sampProviders[provIdx]->GetNumChannels();
            
            double posIdx = floor(pos);
            double t = pos - posIdx;

            int idx = posIdx - initBufPos[provIdx];
            
            if (idx*2 > buffers[provIdx].size())
                idx = buffers[provIdx].size()/2 - 2;
            
            if (idx - 1 >= 0)
            {
                _prevSamps[0][provIdx] = buffers[provIdx][(idx - 1)*2];
                _prevSamps[1][provIdx] = buffers[provIdx][(idx - 1)*2 + 1];
            }
            
            left += ((1.0 - t)*_prevSamps[0][provIdx] +
                     t*buffers[provIdx][idx*2])*envCoeff;
            
            right += ((1.0 - t)*_prevSamps[1][provIdx] +
                      t*buffers[provIdx][idx*2 + 1])*envCoeff;
            
            if (i == numFrames - 1)
                // Last loop step: keep value for using with next buffer
            {
                _prevSamps[0][provIdx] =
                    buffers[provIdx][buffers[provIdx].size() - 2];
                _prevSamps[1][provIdx] =
                    buffers[provIdx][buffers[provIdx].size() - 1];
            }
            
            if (!_loop)
                sampProvidersToKeep.push_back(provIdx);
        
            if (_loop)
                sampProvidersToKeep.push_back(provIdx);

            if (!_sampProviders[provIdx]->InternalPitchShift())
                // Manage pitch shift here
                _positions[provIdx] += step;
            else
                // Just advance by one
                _positions[provIdx] += step/_pitch;
        }
        
        _activeSampProviders = sampProvidersToKeep;

        FillResultBuf(left, right, outBuf, i, bufOffset);    
    }

    // Adjust position depending on time stretch
    for (int j = 0; j < _activeSampProviders.size(); j++)
    {
        int provIdx = _activeSampProviders[j];
        _positions[provIdx] = initBufPos[provIdx] + numFrames*_step*_timeStretch;
    }
    
    return newLoopCycle;
}

bool
AudioSource::GetSamplesFromProvider(SamplesProvider *provider, double paramY,
                                    vector<SAMPLE> *buffer, int numFrames,
                                    double pos, double startPos, double endPos)
{
    provider->SetParamY(paramY);
    provider->SetRange(_range);
        
    int blockMinSamples = provider->GetBlockMinSamples();
        
    if (_step > 0.0)
    {
        provider->GetSamples(buffer, pos, pos + numFrames,
                             _timeStretch, _direction, _pitch);
        
        if (pos + numFrames + blockMinSamples >= endPos)
            return true;
    }
    else
    {
        bool loop = false;

        int startPos0 = pos - numFrames;
        if (startPos0 < startPos)
        {
            startPos0 = startPos;

            loop = true;
        }
        
        provider->GetSamples(buffer, startPos0, pos,
                             _timeStretch, _direction, _pitch);
        
        if (loop)
            return true;
    }

    return false;
}

double
AudioSource::GetEnvelopeCoeff(bool *newLoopCycle)
{
    double envCoeff = 0.0;
    *newLoopCycle = false;
    
    if (_loop)
    {
        if (_envPos > ENVELOPE_LEN - 1)
            _envPos -= ENVELOPE_LEN - 1;
    }
    
    if (_envPos > (ENVELOPE_LEN - 1))
    {
        if (!_loop)
        {
            _envPos = 0.0;
            _isPlaying = false;
            
            *newLoopCycle = true;
        }
    }
    else
    {
        double posIdx = floor(_envPos);
        double t = _envPos - posIdx;
        envCoeff = (1.0 - t)*_envelopeSamples[(unsigned long)posIdx] +
            t*_envelopeSamples[(unsigned long)posIdx + 1];
        
        _envPos += _envStep;
    }

    return envCoeff;
}

void
AudioSource::FillResultBuf(SAMPLE left, SAMPLE right, vector<SAMPLE> *outBuf,
                           int bufPos, unsigned int bufOffset)
{
    if (outBuf == NULL)
        return;
        
    for (int k = 0; k < CHANNELS; k++)
    {
        if ((k % 2) == 0)
        {
            (*outBuf)[(bufOffset + bufPos)*CHANNELS + k] +=
                left*_chanMults[k]*_localAtten;
        }
        else
        {
            (*outBuf)[(bufOffset + bufPos)*CHANNELS + k] +=
                right*_chanMults[k]*_localAtten;
        }
        
        DSP::Clip(&(*outBuf)[(bufOffset + bufPos)*CHANNELS + k], -1.0, 1.0);
    }
}
