/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef AUDIO_SOURCE_H
#define AUDIO_SOURCE_H

#include <vector>
#include <math.h>
#include <time.h>
#include <ctime>

#include "Globals.h"
#include "AudioFileSet.h"
#include "Envelope.h"

class SamplesProvider;
class Envelope;
class AudioSource
{ 
public:    
    AudioSource(const vector<SamplesProvider *> &sampProviders,
                Envelope *envelope, double durationMs,
                double pitch, double timeStretch);

    virtual ~AudioSource();

    AudioSource *Clone() const;

    const vector<SamplesProvider *> &GetSampleProviders() const;
    
    void SetLoop(bool loop);
    
    // Return true is a loop cycle has just ended
    bool NextBuffer(vector<SAMPLE> *outBuf,
                    unsigned int numFrames,
                    unsigned int bufOffset);

    // Return true is play has just been triggered
    bool Play(double *startPositions,
              double *startParamY);

    bool IsPlaying();

    void Update(double *startPositions,
                double *startParamY);
        
    void SetDurationMs(double dur);
   
    void SetPitch(double pitch);
    double GetPitch();

    void SetTimeStretch(double timeStretch);
    double GetTimeStretch();
    
    void SetVolume(double normVolume);
    double GetVolume();
    
    void SetChannelMultipliers(const vector<SAMPLE> &multipliers);
    
    void SetDirection(double dir);
    
    void SetEnvelopeType(unsigned int envelopeType);

    void SetRange(double range);
    
    void UpdateParams();
    
protected:
    bool GetSamplesFromProvider(SamplesProvider *provider, double paramY,
                                vector<SAMPLE> *buffer, int numFrames,
                                double pos, double startPos, double endPos);

    double GetEnvelopeCoeff(bool *newLoopCycle);

    void FillResultBuf(SAMPLE left, SAMPLE right, vector<SAMPLE> *outBuf,
                       int bufPos, unsigned int bufOffset);
        
    bool _loop;
    
    vector<SamplesProvider *> _sampProviders;
    
    bool _isPlaying;
    
    bool _newParam;
    
    double _durationMs;
    double _queuedDurationMs;
    double _durationSamps;
    double _pitch;
    double _queuedPitch;
    double _timeStretch;
    double _queuedTimeStretch;
    double _direction;
    double _queuedDirection; 
    double _step;
    
    double _localAtten;
    double _queuedLocalAtten;
    
    vector<SAMPLE> _chanMults;
    vector<SAMPLE> _queuedChanMults;
    
    vector<int> _activeSampProviders;
    
    unsigned int _envelopeType;
    unsigned int _queuedEnvelopeType;
    
    double _envStep;
    double _envPos;

    Envelope *_envelope;
    vector<SAMPLE> _envelopeSamples;
    
    vector<double> _paramY;

    double _range;
    
    vector<double> _positions;
    vector<double> _startPositions;
    vector<double> _endPositions;

    vector<double> _prevSamps[2];

    bool _isClone;
};

#endif
