/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <string.h>

#include "DSP.h"

void
DSP::Normalize(vector<SAMPLE> *buffer)
{
    SAMPLE maxVal = 0.0;
    for(int k = 0; k < buffer->size(); k++)
    {
        if (fabs((*buffer)[k]) > maxVal)
            maxVal = fabs((*buffer)[k]);
    }
        
    SAMPLE normFactor = 1.0;
    if (maxVal > 1e-37)
        normFactor = 1.0/maxVal;
    
    normFactor *= DIM_FACTOR;
    
    for(int k = 0; k < buffer->size(); k++)
    {
        (*buffer)[k] *= normFactor;
    }
}

void
DSP::AutomaticClipping(vector<SAMPLE> *buffer)
{
#define THRESHOLD 2.0

    // Compute average
    SAMPLE avg = 0.0;
    for(int k = 0; k < buffer->size(); k++)
        avg += fabs((*buffer)[k]);
    if (buffer->size() > 0)
        avg /= buffer->size();

    // Compute standard deviation
    SAMPLE stdev = 0.0;
    for(int k = 0; k < buffer->size(); k++)
        stdev += fabs(fabs((*buffer)[k]) - avg);
    if (buffer->size() > 0)
        stdev /= buffer->size();

    // Clip
    double thrs = THRESHOLD*(avg + stdev);
    for(int k = 0; k < buffer->size(); k++)
    {
        if (fabs((*buffer)[k]) > thrs)
            // Must clip
        {
            if ((*buffer)[k] > 0.0)
                (*buffer)[k] = thrs;
            else
                (*buffer)[k] = -thrs;
        }
    }
}

void
DSP::Interleave(vector<SAMPLE> *buffer, const vector<SAMPLE> &lBuf,
                const vector<SAMPLE> &rBuf)
{
    for (int i = 0; i < lBuf.size(); i++)
    {
        (*buffer)[i*2] = lBuf[i];
        (*buffer)[i*2 + 1] = rBuf[i];
    }
}

void
DSP::DeInterleave(const vector<SAMPLE> &buffer, vector<SAMPLE> *lBuf, vector<SAMPLE> *rBuf)
{
    for (int i = 0; i < lBuf->size(); i++)
    {
        (*lBuf)[i] = buffer[i*2];
        (*rBuf)[i] = buffer[i*2 + 1];
    }
}

SAMPLE
DSP::CmpChannels(const vector<SAMPLE> &buffer)
{
    SAMPLE sum = 0.0;
    for (int i = 0; i < buffer.size()/2; i++)
        sum += fabs(buffer[i*2] - buffer[i*2 + 1]);

    sum /= buffer.size()/2;

    return sum;
}

void
DSP::FadeIn(vector<SAMPLE> *buffer, int nChan, int numSamplesFade)
{
    if (buffer->size()/nChan <= numSamplesFade)
        return;
    
    for (int i = 0; i < numSamplesFade; i++)
    {
        double t = ((double)i)/(numSamplesFade - 1);
        
        for (int k = 0; k < nChan; k++)
            (*buffer)[i*nChan + k] *= t;
    }
}

void
DSP::FadeOut(vector<SAMPLE> *buffer, int nChan, int numSamplesFade)
{
    int nFrames = buffer->size()/nChan;
    if (nFrames <= numSamplesFade)
        return;
    
    for (int i = nFrames - 1; i > nFrames - 1 - numSamplesFade; i--)
    {
        double t = ((double)nFrames - 1 - i)/(numSamplesFade - 1);

        for (int k = 0; k < nChan; k++)
            (*buffer)[i*nChan + k] *= t;
    }
}

void
DSP::Balance(vector<SAMPLE> *buffer, int nChan, double balance)
{
    if (nChan < 2)
        return;

    SAMPLE p = M_PI*(balance + 1.0)/4.0;
    SAMPLE gl = cos(p);
    SAMPLE gr = sin(p);
    
    // Coefficients to have no gain when center, and +3dB for extreme pan pos
    gl *= sqrt(2.0);
    gr *= sqrt(2.0);
    
    for (int i = 0; i < buffer->size()/nChan; i++)
    {
        SAMPLE l = (*buffer)[i*2];
        SAMPLE r = (*buffer)[i*2 + 1];
        
        l *= gl;
        r *= gr;
        
        (*buffer)[i*2] = l;
        (*buffer)[i*2 + 1] = r;
    }
}

void
DSP::BitCrush(vector<SAMPLE> *buffer, double ratio)
{
    // Must divide by 2 to have the right number of steps in the waveform
#define COEFF 0.5
    
    double depth = 1.0 + (1.0 - ratio)*15.0;
    long long int depthI = std::pow((double)2.0, depth);
    
    for (int i = 0; i < buffer->size(); i++)
    {
        if (ratio <= 0.0)
            continue;
        
        double s = (*buffer)[i];
             
        long long int si = ((s + 1.0)*0.5*COEFF)*depthI;
        s = ((SAMPLE)si)/(COEFF*depthI);
        s = s*2.0 - 1.0;
              
        (*buffer)[i] = s;
    }
}

void
DSP::Saturation(vector<SAMPLE> *buffer, double ratio)
{
    double cut = 1.0 - ratio;
    double factor = 1.0/(1.0 - ratio);
    
    for (int i = 0; i < buffer->size(); i++)
    {
        if (ratio <= 0.0)
            continue;
        
        double s0 = (*buffer)[i];
        double s1 = s0;
        
        if (s0 > cut)
            s1 = cut;
    
        if (s0 < -cut)
            s1 = -cut;
    
        s1 *= factor;
              
        (*buffer)[i] = s1;
    }
}

void
DSP::Silent(vector<SAMPLE> *buffer)
{
    memset(buffer->data(), 0, buffer->size()*sizeof(SAMPLE));
}

bool
DSP::IsSilence(const vector<SAMPLE> &buffer)
{
    for (int i = 0; i < buffer.size(); i++)
    {
        if (fabs(buffer[i]) > 1e-15)
            return false;
    }

    return true;
}

void
DSP::Reverse(vector<SAMPLE> *buffer, int nChan)
{
    vector<SAMPLE> copy = *buffer;
    
    for (int i = 0; i < buffer->size()/nChan; i++)
    {
        int j = buffer->size()/nChan - i - 1;
        
        for (int k = 0; k < nChan; k++)
            (*buffer)[i*nChan + k] = copy[j*nChan + k];
    }
}

void
DSP::Clip(SAMPLE *s, SAMPLE minS, SAMPLE maxS)
{
    if (*s < minS)
        *s = minS;
    if (*s > maxS)
        *s = maxS;
}
