/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Parameter.h"
#include "DelayLine.h"
#include "DelayEffect.h"

#define DEFAULT_DELAY_MS 46.0

DelayEffect::DelayEffect()
{
    _timeSamples = DEFAULT_DELAY_MS*0.001*SRATE;
        
    _feedback = 0.0;
    _wetMix = 0.5;

    _params[ON_PARAM] = new ParameterEffectOn(NULL, ON_PARAM);
    _params[TIME_PARAM] = new ParameterDelayEffectTime(NULL, TIME_PARAM);
    _params[FEEDBACK_PARAM] = new ParameterDelayEffectFeedback(NULL, FEEDBACK_PARAM);
    _params[WET_MIX_PARAM] = new ParameterDelayEffectWetMix(NULL, WET_MIX_PARAM);

    for (int i = 0; i < CHANNELS; i++)
        _delayLines[i] = new DelayLine(_timeSamples);
}

DelayEffect::~DelayEffect()
{
    for (int i = 0; i < NUM_PARAMS; i++)
        delete _params[i];

    for (int i = 0; i < CHANNELS; i++)
        delete _delayLines[i];
}

void
DelayEffect::Process(vector<SAMPLE> *buffer, int nFrames, int nChan)
{
    if (nChan < 2)
        return;

    UpdateParameters();

    for (int i = 0; i < CHANNELS; i++)
        _delayLines[i]->SetDelay(_timeSamples);
    
    double on = _params[ON_PARAM]->GetValue();
    if (on < 0.5)
        return;
    
    for (int i = 0; i < nFrames; i++)
    {
        for (int k = 0; k < nChan; k++)
        {
            double s = (*buffer)[i*nChan + k];
            double s0 = _delayLines[k]->Peek();
            double d = _delayLines[k]->Process(s + _feedback*s0);

            double r = s + _wetMix*d;
            (*buffer)[i*nChan + k] = r;
        }
    }
}

void
DelayEffect::SetTime(double timeMs)
{
    _timeSamples = timeMs*0.001*SRATE;
}

void
DelayEffect::SetFeedback(double feedback)
{
    _feedback = feedback*0.01;
}

void
DelayEffect::SetWetMix(double wetMix)
{
    _wetMix = wetMix*0.01;
}

Parameter *
DelayEffect::GetParameter(ParamIdx paramIdx)
{
    return _params[paramIdx];
}

void
DelayEffect::Serialize(Serializer *serializer)
{
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
DelayEffect::UnSerialize(Serializer *serializer)
{
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);

    UpdateParameters();
}

void
DelayEffect::UpdateParameters()
{
    double timeMs = _params[TIME_PARAM]->GetValue();
    _timeSamples = timeMs*0.001*SRATE;
    double feedback = _params[FEEDBACK_PARAM]->GetValue();
    _feedback = feedback*0.01;
    double wetMix = _params[WET_MIX_PARAM]->GetValue();
    _wetMix = wetMix*0.01;
}
