/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>

#include "DSP.h"
#include "DelayLine.h"

DelayLine::DelayLine(double delaySamples)
{
    _readPos = 0;
    
    int delayI = ceil(delaySamples);
    _delayLine.resize(delayI);
    
    DSP::Silent(&_delayLine);
   
    _writePos = _delayLine.size() - 1;
    if (_writePos < 0)
        _writePos = 0;

    _delay = 0.0;
    
    SetDelay(delaySamples);
}

DelayLine::~DelayLine() {}
    
void
DelayLine::Reset()
{
    DSP::Silent(&_delayLine);
    
    _readPos = 0;
    
    _writePos = _delayLine.size() - 1;
    if (_writePos < 0)
        _writePos = 0;
}
    
void
DelayLine::SetDelay(double delaySamples)
{
    if (fabs(delaySamples - _delay) < 1e-15)
        return;
    
    _delay = delaySamples;
 
    int delayI = ceil(_delay);
    if (delayI == 0)
        delayI = 1;

    if (_delayLine.size() < delayI)
        // We must grow the line size
    {
        int numToInsert = delayI - _delayLine.size();
        double s = _delayLine[_writePos];
        _delayLine.insert(_delayLine.begin() + _writePos,
                          numToInsert, s);

        if (_readPos > _writePos)
            _readPos = (_readPos + numToInsert) % _delayLine.size();
        _writePos = (_writePos + numToInsert) % _delayLine.size();
    }
    else if (_delayLine.size() > delayI)
    {
        while(_delayLine.size() > delayI)
        {
            _delayLine.erase(_delayLine.begin() + _writePos);
        
            if (_readPos > _writePos)
            {
                _readPos--;
                if (_readPos < 0)
                    _readPos += _delayLine.size();
            }
            
            _writePos--;
            if (_writePos < 0)
                _writePos += _delayLine.size();
        }
    }
}

double
DelayLine::Peek()
{
    if (_delayLine.empty())
        return 0.0;
    
    double s = _delayLine[_readPos];

    return s;
}

SAMPLE
DelayLine::Process(SAMPLE sample)
{
    if (_delayLine.empty())
        return sample;
    
    SAMPLE s = _delayLine[_readPos];
    _readPos++;
    if (_readPos >= _delayLine.size())
        _readPos -= _delayLine.size();
    
    _writePos++;
    if (_writePos >= _delayLine.size())
        _writePos -= _delayLine.size();
    _delayLine[_writePos] = sample;
    
    return s;
}

void
DelayLine::Process(vector<SAMPLE> *samples)
{
    for (int i = 0; i < samples->size(); i++)
    {
        double s = (*samples)[i];
        s = Process(s);
        (*samples)[i] = s;
    }
}
