/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "Serializer.h"
#include "Envelope.h"

Envelope::~Envelope() {}

Envelope::Envelope(unsigned long length)
{
    _envSize = length;
    
    _hanningEnv.resize(length);
    _triEnv.resize(length);
    _expDecEnv.resize(length);
    _rexpDecEnv.resize(length);
    _sincEnv.resize(length);
    _handDrawnEnv.resize(length);
    
    GenerateEnvelopes();
}

int
Envelope::GetNumEnvelopes()
{
    return NUM_ENVELOPES;
}

void
Envelope::Serialize(Serializer *serializer)
{
    int envSize = GetEnvelopeSize();
    serializer->PutInt(envSize);

    for (int i = 0; i < envSize; i++)
    {
        double s = _handDrawnEnv[i];
        serializer->PutDouble(s);
    }
}

void
Envelope::UnSerialize(Serializer *serializer)
{
    // UnSerialize hand draw envelope
    int envSize  = GetEnvelopeSize();
    serializer->GetInt(&envSize);

    for (int i = 0; i < envSize; i++)
    {
        double s;
        serializer->GetDouble(&s);
        _handDrawnEnv[i] = s;
    }
}

void
Envelope::GenerateEnvelopes()
{
    // Clear hand drawn envelopes
    for (int i = 0; i < _handDrawnEnv.size(); i++)
    {
        _handDrawnEnv[i] = (SAMPLE)1.0;
    }

    Hanning(&_hanningEnv);
    Triangle(&_triEnv);
    Sinc(&_sincEnv, 8);
    Expdec(&_expDecEnv, &_rexpDecEnv);
}

void
Envelope::Hanning(vector<SAMPLE> *envelope)
{
    unsigned long i;
    SAMPLE phase = 0.0;
    SAMPLE delta = 2 * M_PI / (SAMPLE)envelope->size();
    
    for(i = 0; i < envelope->size(); i++)
    {
        (*envelope)[i] = (SAMPLE)(0.5 * (1.0 - cos(phase)));
        phase += delta;
    }
}

void
Envelope::Triangle(vector<SAMPLE> *envelope)
{
    SAMPLE norm = (2.0 / ((SAMPLE)envelope->size() - 1));
    SAMPLE normInv = 1.0/norm;
    
    for (int i = 0; i < envelope->size(); i++)
        (*envelope)[i] = norm * (normInv - fabs(i - normInv));
}

void
Envelope::Expdec(vector<SAMPLE> *forEnv, vector<SAMPLE> *revEnv)
{
    SAMPLE tau = 512.0;
    SAMPLE tauInv = 1.0/tau;
    SAMPLE eps = 0.0005;
    unsigned long transition = forEnv->size() - 48;
    
    for (int i = 0; i < forEnv->size(); i++)
    {
        (*forEnv)[i] = exp(-((SAMPLE) i)*tauInv);
        if (revEnv)
            (*revEnv)[forEnv->size() - i - 1] = (*forEnv)[i]; 
    }
}

void
Envelope::Sinc(vector<SAMPLE> *envelope, int numZeroCross)
{
    SAMPLE step = numZeroCross/(SAMPLE)envelope->size();
    SAMPLE x = -(numZeroCross)/2.0;
    
    for (int i = 0; i < envelope->size(); i++)
    {
        if (x != 0)
            _sincEnv[i] = fabs(sin(M_PI*x)/(M_PI * x));
        else
            _sincEnv[i] = (SAMPLE) 1.0;

        x += step;
    }
}

const vector<SAMPLE> &
Envelope::GetEnvelope(int envelopeType)
{
    switch (envelopeType)
    {
        case HANNING:
            return _hanningEnv;
            break;
            
        case TRIANGLE:
            return _triEnv;
            break;
            
        case EXPDEC:
            return _expDecEnv;
            break;
            
        case REXPDEC:
            return _rexpDecEnv;
            break;
            
        case SINC:
            return _sincEnv;
            break;

        case HAND_DRAWN:
            return _handDrawnEnv;
            break;
            
        default:
            return _hanningEnv;
            break;
    }
}

void
Envelope::SetEnvelope(int envelopeType, const vector<SAMPLE> &data)
{
    switch (envelopeType)
    {
        case HANNING:
            _hanningEnv = data;
            break;
            
        case TRIANGLE:
            _triEnv = data;
            break;
            
        case EXPDEC:
            _expDecEnv = data;
            break;
            
        case REXPDEC:
            _rexpDecEnv = data;
            break;
            
        case SINC:
            _sincEnv = data;
            break;

        case HAND_DRAWN:
            _handDrawnEnv = data;
            break;
            
        default:
            _hanningEnv = data;
            break;
    }
}

int
Envelope::GetEnvelopeSize()
{
    return _envSize;
}
