/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>
#include <string.h>

#include <algorithm>
using namespace std;

#include "Globals.h"
#include "Scale.h"
#include "FilterBank.h"

#define FIX_ALIASING_LOW_FREQS 1
#define FIX_ALIASING_MIN_TRIANGLE_WIDTH 2.0

FilterBank::FilterBank(Scale::Type targetScaleType)
{
    _targetScaleType = targetScaleType;
}

FilterBank::~FilterBank() {}

void
FilterBank::HzToTargetScale(vector<SAMPLE> *data,
                            double sampleRate,
                            SAMPLE freqLo, SAMPLE freqHi,
                            int numFilters)
{
    if ((data->size() != _hzToTargetFilterBank._dataSize) ||
        (sampleRate != _hzToTargetFilterBank._sampleRate) ||
        (numFilters != _hzToTargetFilterBank._numFilters) ||
        (freqLo != _hzToTargetFilterBank._freqLo) ||
        (freqHi != _hzToTargetFilterBank._freqHi))
    {
        CreateFilterBankHzToTarget(&_hzToTargetFilterBank,
                                   data->size(),
                                   sampleRate, freqLo, freqHi,
                                   numFilters);
    }

    _tmpBuf2 = *data;
    ApplyFilterBank(data, _tmpBuf2, _hzToTargetFilterBank);
}

void
FilterBank::TargetScaleToHz(vector<SAMPLE> *data,
                            double sampleRate,
                            SAMPLE freqLo, SAMPLE freqHi,
                            int numFilters)
{
    if ((data->size() != _targetToHzFilterBank._dataSize) ||
        (sampleRate != _targetToHzFilterBank._sampleRate) ||
        (numFilters != _targetToHzFilterBank._numFilters) ||
        (freqLo != _hzToTargetFilterBank._freqLo) ||
        (freqHi != _hzToTargetFilterBank._freqHi))
    {
        CreateFilterBankTargetToHz(&_targetToHzFilterBank,
                                   data->size(),
                                   sampleRate, freqLo, freqHi,
                                   numFilters);
    }

    _tmpBuf2 = *data;
    ApplyFilterBank(data, _tmpBuf2, _targetToHzFilterBank);
}

SAMPLE
FilterBank::ComputeTriangleAreaBetween(SAMPLE txmin, SAMPLE txmid,
                                       SAMPLE txmax,
                                       SAMPLE x0, SAMPLE x1)
{
    if ((x0 > txmax) || (x1 < txmin))
        return 0.0;
    
    vector<SAMPLE> &x = _tmpBuf;
    x.resize(5);
    x[0] = txmin;
    x[1] = txmid;
    x[2] = txmax;
    x[3] = x0;
    x[4] = x1;
    
    sort(x.begin(), x.end());
    
    SAMPLE points[5][2];
    for (int i = 0; i < 5; i++)
    {
        points[i][0] = x[i];
        points[i][1] = ComputeTriangleY(txmin, txmid, txmax, x[i]);
    }
    
    SAMPLE area = 0.0;
    for (int i = 0; i < 4; i++)
    {
        // Suppress the cases which are out of [x0, x1] bounds
        if ((points[i][0] >= x1) ||
            (points[i + 1][0] <= x0))
            continue;
      
        SAMPLE y0 = points[i][1];
        SAMPLE y1 = points[i + 1][1];
        if (y0 > y1)
        {
            SAMPLE tmp = y0;
            y0 = y1;
            y1 = tmp;
        }
      
        SAMPLE a =
            (points[i + 1][0] - points[i][0])*(y0 + (y1 - y0)*0.5);
      
        area += a;
    }
    
    return area;
}
  
SAMPLE
FilterBank::ComputeTriangleY(SAMPLE txmin, SAMPLE txmid,
                             SAMPLE txmax, SAMPLE x)
{
    if (x <= txmin)
        return 0.0;
    if (x >= txmax)
        return 0.0;
    
    if (x <= txmid)
    {
        SAMPLE t = (x - txmin)/(txmid - txmin);
        
        return t;
    }
    else // x >= txmid
    {
        SAMPLE t = 1.0 - (x - txmid)/(txmax - txmid);
        
        return t;
    }
}

SAMPLE
FilterBank::ScaleValue(SAMPLE val, SAMPLE minFreq, SAMPLE maxFreq)
{
    switch (_targetScaleType)
    {
        case Scale::LINEAR:
        {
            SAMPLE minTargetValue = minFreq;
            SAMPLE maxTargetValue = maxFreq;
            val = (val - minTargetValue)/
                (maxTargetValue - minTargetValue);
        }
        break;
      
        case Scale::LOG:
        {
            SAMPLE minTargetValue = log(minFreq);
            SAMPLE maxTargetValue = log(maxFreq); 
            val = log(val);

            val = (val - minTargetValue)/(maxTargetValue - minTargetValue);
        }
        break;
      
        case Scale::MEL:
        {
            SAMPLE minTargetValue = Scale::HzToMel(minFreq);
            SAMPLE maxTargetValue = Scale::HzToMel(maxFreq);
            val = Scale::HzToMel(val);
        
            val = (val - minTargetValue)/(maxTargetValue - minTargetValue);
        }
        break;
      
        default:
            break;
    }
    
    return val;
}
  
SAMPLE
FilterBank::ScaleValueInv(SAMPLE val, SAMPLE minFreq, SAMPLE maxFreq)
{
    switch (_targetScaleType)
    {
        case Scale::LINEAR:
        {
            SAMPLE minTargetValue = minFreq;
            SAMPLE maxTargetValue = maxFreq;
            val = val*(maxTargetValue - minTargetValue) +
                minTargetValue;
        }
        break;
        
        case Scale::LOG:
        {
            SAMPLE minTargetValue = log(minFreq);
            SAMPLE maxTargetValue = log(maxFreq);
            val = val*(maxTargetValue - minTargetValue) +
                minTargetValue;
            
            val = exp(val);
        }
        break;
        
        case Scale::MEL:
        {
            SAMPLE minTargetValue = Scale::HzToMel(minFreq);
            SAMPLE maxTargetValue = Scale::HzToMel(maxFreq);
            val = val*(maxTargetValue - minTargetValue) +
                minTargetValue;
            
            val = Scale::MelToHz(val);
        }
        break;
        
        default:
            break;
    }
    
    return val;
}

// FilterBankObj
FilterBank::FilterBankObj::FilterBankObj()
{
    _sampleRate = 44100.0;
    _freqLo = 20.0;
    _freqHi = 22050.0;
    _dataSize = 0;
    _numFilters = 0;
}

FilterBank::FilterBankObj::FilterBankObj(int dataSize,
                                         double sampleRate,
                                         SAMPLE freqHi,
                                         SAMPLE freqLo,
                                         int numFilters)
: _dataSize(dataSize),
  _sampleRate(sampleRate),
  _freqLo(freqLo),
  _freqHi(freqHi),
  _numFilters(numFilters)
{  
    _filters.resize(_numFilters);
    for (int i = 0; i < _filters.size(); i++)
    {
        _filters[i]._data.resize(dataSize);
        memset(_filters[i]._data.data(), 0,
               _filters[i]._data.size()*sizeof(SAMPLE));
        
        _filters[i]._bounds[0] = -1;
        _filters[i]._bounds[1] = -1;
    }
}

FilterBank::FilterBankObj::~FilterBankObj() {}
  
void
FilterBank::CreateFilterBankHzToTarget(FilterBankObj* filterBank,
                                       int dataSize, double sampleRate,
                                       SAMPLE freqLo, SAMPLE freqHi,
                                       int numFilters)
{
    // Clear previous
    filterBank->_filters.resize(0);

    // Init
    filterBank->_dataSize = dataSize;
    filterBank->_sampleRate = sampleRate;
    filterBank->_numFilters = numFilters;
    filterBank->_freqLo = freqLo;
    filterBank->_freqHi = freqHi;
    
    filterBank->_filters.resize(filterBank->_numFilters);
    
    for (int i = 0; i < filterBank->_filters.size(); i++)
    {
        filterBank->_filters[i]._data.resize(dataSize);
        memset(filterBank->_filters[i]._data.data(),
               0,
               filterBank->_filters[i]._data.size()*sizeof(SAMPLE));
      
        filterBank->_filters[i]._bounds[0] = -1;
        filterBank->_filters[i]._bounds[1] = -1;
    }
      
    // Create filters
    SAMPLE lowFreqTarget = ScaleValue(freqLo, freqLo, freqHi);
    SAMPLE highFreqTarget = ScaleValue(freqHi, freqLo, freqHi);
    
    // Compute equally spaced target values
    vector<SAMPLE> targetPoints;
    targetPoints.resize(numFilters + 2);
    for (int i = 0; i < targetPoints.size(); i++)
    {
        // Compute target value
        SAMPLE t = ((SAMPLE)i)/(targetPoints.size() - 1);
        SAMPLE val = lowFreqTarget + t*(highFreqTarget - lowFreqTarget);
      
        targetPoints[i] = val;
    }
    
    // Compute target points
    vector<SAMPLE> hzPoints;
    hzPoints.resize(targetPoints.size());
    for (int i = 0; i < hzPoints.size(); i++)
    {
        // Compute hz value
        SAMPLE val = targetPoints[i];
        val = ScaleValueInv(val, freqLo, freqHi);
      
        hzPoints[i] = val;
    }
    
    // Compute bin points
    vector<SAMPLE> bin;
    bin.resize(hzPoints.size());
    
    SAMPLE hzPerBinInv = (dataSize + 1)/(freqHi - freqLo);
    for (int i = 0; i < bin.size(); i++)
    {
        // Compute hz value
        SAMPLE val = hzPoints[i];
      
        // For the new solution that fills holes, do not round or trunk
        val = (val + freqLo)*hzPerBinInv;
        
        bin[i] = val;
    }
    
    // For each filter
    for (int m = 1; m < numFilters + 1; m++)
    {
        SAMPLE fmin = bin[m - 1]; // left
        SAMPLE fmid = bin[m];     // center
        SAMPLE fmax = bin[m + 1]; // right
      
#if FIX_ALIASING_LOW_FREQS
        FixSmallTriangles(&fmin, &fmax, dataSize);
#endif
      
        filterBank->_filters[m - 1]._bounds[0] = floor(fmin);
        filterBank->_filters[m - 1]._bounds[1] = ceil(fmax);
      
        // Check upper bound
        if (filterBank->_filters[m - 1]._bounds[1] > dataSize - 1)
            filterBank->_filters[m - 1]._bounds[1] = dataSize - 1;
      
        for (int i = filterBank->_filters[m - 1]._bounds[0];
             i <= filterBank->_filters[m - 1]._bounds[1]; i++)
        {
            // Trapezoid
            SAMPLE x0 = i;
            if (fmin > x0)
                x0 = fmin;
        
            SAMPLE x1 = i + 1;
            if (fmax < x1)
                x1 = fmax;
        
            SAMPLE tarea =
                ComputeTriangleAreaBetween(fmin, fmid, fmax, x0, x1);
        
            // Normalize
            tarea /= (fmid - fmin)*0.5 + (fmax - fmid)*0.5;
        
            filterBank->_filters[m - 1]._data[i] += tarea;
        }
    }
}
  
void
FilterBank::CreateFilterBankTargetToHz(FilterBankObj *filterBank,
                                       int dataSize,
                                       double sampleRate,
                                       SAMPLE freqLo, SAMPLE freqHi,
                                       int numFilters)
{
    filterBank->_dataSize = dataSize;
    filterBank->_sampleRate = sampleRate;
    filterBank->_numFilters = numFilters;
    filterBank->_freqLo = freqLo;
    filterBank->_freqHi = freqHi;
      
    filterBank->_filters.resize(filterBank->_numFilters);
  
    for (int i = 0; i < filterBank->_filters.size(); i++)
    {
        filterBank->_filters[i]._data.resize(dataSize);
        memset(filterBank->_filters[i]._data.data(),
               0, filterBank->_filters[i]._data.size()*sizeof(SAMPLE));
      
        filterBank->_filters[i]._bounds[0] = -1;
        filterBank->_filters[i]._bounds[1] = -1;
    }
    
    // Create filters
    SAMPLE lowFreqHz = freqLo;
    SAMPLE highFreqHz = freqHi;
    
    vector<SAMPLE> hzPoints;
    hzPoints.resize(numFilters + 2);
    for (int i = 0; i < hzPoints.size(); i++)
    {
        // Compute hz value
        SAMPLE t = ((SAMPLE)i)/(hzPoints.size() - 1);
        SAMPLE val = lowFreqHz + t*(highFreqHz - lowFreqHz);
      
        hzPoints[i] = val;
    }
    
    // Compute hz points
    vector<SAMPLE> targetPoints;
    targetPoints.resize(hzPoints.size());
    for (int i = 0; i < targetPoints.size(); i++)
    {
        // Compute hz value
        SAMPLE val = hzPoints[i];
        val = ScaleValue(val, freqLo, freqHi);
      
        targetPoints[i] = val;
    }
    
    // Compute bin points
    vector<SAMPLE> bin;
    bin.resize(targetPoints.size());
    
    SAMPLE maxTarget = ScaleValue(freqHi, freqLo, freqHi);
    SAMPLE targetPerBinInv = (dataSize + 1)/maxTarget;
    for (int i = 0; i < bin.size(); i++)
    {
        // Compute target value
        SAMPLE val = targetPoints[i];
      
        // For the new solution that fills holes, do not round or trunk
        val = val*targetPerBinInv;
      
        bin[i] = val;
    }
    
    // For each filter
    for (int m = 1; m < numFilters; m++)
    {
        SAMPLE fmin = bin[m - 1]; // left
        SAMPLE fmid = bin[m];     // center
        SAMPLE fmax = bin[m + 1]; // right
      
#if FIX_ALIASING_LOW_FREQS
        FixSmallTriangles(&fmin, &fmax, dataSize);
#endif
      
        filterBank->_filters[m]._bounds[0] = floor(fmin);
        filterBank->_filters[m]._bounds[1] = ceil(fmax);
      
        // Check upper bound
        if (filterBank->_filters[m ]._bounds[1] > dataSize - 1)
            filterBank->_filters[m]._bounds[1] = dataSize - 1;
      
        for (int i = filterBank->_filters[m]._bounds[0];
             i <= filterBank->_filters[m]._bounds[1]; i++)
        {
            // Trapezoid
            SAMPLE x0 = i;
            if (fmin > x0)
                x0 = fmin;
        
            SAMPLE x1 = i + 1;
            if (fmax < x1)
                x1 = fmax;
        
            SAMPLE tarea =
                ComputeTriangleAreaBetween(fmin, fmid, fmax, x0, x1);
        
            // Normalize
            tarea /= (fmid - fmin)*0.5 + (fmax - fmid)*0.5;
        
            filterBank->_filters[m]._data[i] += tarea;
        }
    }
}

void
FilterBank::ApplyFilterBank(vector<SAMPLE> *output,
                            const vector<SAMPLE> &input,
                            const FilterBankObj &filterBank)
{
    output->resize(filterBank._numFilters);
    memset(output->data(), 0, output->size()*sizeof(SAMPLE));
    
    // For each filter
    for (int m = 0; m < filterBank._numFilters; m++)
    {
        const FilterBankObj::Filter &filter = filterBank._filters[m];
      
        const SAMPLE *filterData = filter._data.data();
        SAMPLE *outputData = output->data();
        const SAMPLE *inputData = input.data();
      
        // For each destination value
        for (int i = filter._bounds[0]; i <= filter._bounds[1]; i++)
        {
            if (i < 0)
                continue;
            if (i >= input.size())
                continue;
            
            // Apply the filter value
            outputData[m] += filterData[i]*inputData[i];
        }
    }
}

void
FilterBank::FixSmallTriangles(SAMPLE *fmin, SAMPLE *fmax,
                              int dataSize)
{
    if (dataSize < FIX_ALIASING_MIN_TRIANGLE_WIDTH)
        return;
  
    if (*fmax - *fmin < FIX_ALIASING_MIN_TRIANGLE_WIDTH)
    {
        SAMPLE diff =
            FIX_ALIASING_MIN_TRIANGLE_WIDTH - (*fmax - *fmin);
        *fmin -= diff*0.5;
        *fmax += diff*0.5;
      
        if (*fmin < 0.0)
        {
            *fmax += -*fmin;
            *fmin = 0.0;
        }
      
        if (*fmax > dataSize - 1)
        {
            *fmin -= *fmax - (dataSize - 1);
            *fmax = dataSize - 1;
        }
    }
}
