/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef FILTER_BANK_H
#define FILTER_BANK_H

#include <vector>
using namespace std;

#include "Scale.h"

// Use the following idea, originally used for mel scale:
// http://practicalcryptography.com/miscellaneous/machine-learning/guide-mel-frequency-cepstral-coefficients-mfccs/
// ... but generalize it to any scale type
//
// It can also be used to smoothly decimate or scale up data 
class FilterBank
{
public:
    FilterBank(Scale::Type targetScaleType);
    virtual ~FilterBank();
  
    // Convert from Hz to target scale using filters
    void HzToTargetScale(vector<SAMPLE> *data,
                         double sampleRate,
                         SAMPLE freqLo, SAMPLE freqHi,
                         int numFilters);
    
    // Convert from target scale to Hz filters */
    void TargetScaleToHz(vector<SAMPLE> *data,
                         double sampleRate,
                         SAMPLE freqLo, SAMPLE freqHi,
                         int numFilters);
    
protected:
    SAMPLE ComputeTriangleAreaBetween(SAMPLE txmin, SAMPLE txmid,
                                      SAMPLE txmax,
                                      SAMPLE x0, SAMPLE x1);
    
    SAMPLE ComputeTriangleY(SAMPLE txmin, SAMPLE txmid,
                            SAMPLE txmax, SAMPLE x);
  
    SAMPLE ScaleValue(SAMPLE val, SAMPLE minFreq, SAMPLE maxFreq);
  
    SAMPLE ScaleValueInv(SAMPLE val, SAMPLE minFreq, SAMPLE maxFreq);
    
    class FilterBankObj
    {
    public:
        FilterBankObj();
        FilterBankObj(int dataSize, double sampleRate,
                      SAMPLE freqHi, SAMPLE freqLo, int numFilters);
        virtual ~FilterBankObj();
      
    protected:
        friend class FilterBank;
    
        double _sampleRate;
        SAMPLE _freqLo;
        SAMPLE _freqHi;
        int _dataSize;
        int _numFilters;
      
        struct Filter
        {
            vector<SAMPLE> _data;
            int _bounds[2];
        };
        
        vector<Filter> _filters;
    };
  
    void CreateFilterBankHzToTarget(FilterBankObj *filterBank,
                                    int dataSize,
                                    double sampleRate,
                                    SAMPLE freqLo, SAMPLE freqHi,
                                    int numFilters);
  
    void CreateFilterBankTargetToHz(FilterBankObj *filterBank,
                                    int dataSize,
                                    double sampleRate,
                                    SAMPLE freqLo, SAMPLE freqHi,
                                    int numFilters);
  
    void ApplyFilterBank(vector<SAMPLE> *output,
                         const vector<SAMPLE> &input,
                         const FilterBankObj &filterBank);
    
    void FixSmallTriangles(SAMPLE *fmin, SAMPLE *fmax, int dataSize);
    
    FilterBankObj _hzToTargetFilterBank;
    FilterBankObj _targetToHzFilterBank;

    Scale::Type _targetScaleType;
    
private:
    // tmp buffers
    vector<SAMPLE> _tmpBuf;
    vector<SAMPLE> _tmpBuf2;
};

#endif
