/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "UIStyle.h"
#include "UIGrainCloud.h"
#include "Serializer.h"
#include "Parameter.h"
#include "App.h"
#include "GrainCloud.h"

GrainCloud::GrainCloud(App *app,
                       const vector<SamplesProvider *> &sampProviders,
                       double numVoices)
{
    _sourceType = Source::GRAIN_CLOUD;
    
    _app = app;
    
    _numVoices = numVoices;
    
    srand(time(NULL));
    
    _addFlag = false;
    _removeFlag = false;
    
    _nextGrain = 0;

    _duration = 500.0; // 500ms

    _pitch = 1.0f;
    
    _envelopeType = Envelope::HANNING;
    
    _pitchLFOFreq = 0.01f;
    _pitchLFOAmount = 0.0f;
    
    _channelMults.resize(CHANNELS);
    for (int i = 0; i < CHANNELS; i++)
        _channelMults[i] = 0.999f;
    
    _currentAroundChan = 1;
    _stereoSide = 0; 
    _side = 1;
    
    _spatialMode = UNITY;
    _channelLocation = -1;
    
    _dirMode = FORWARD;

    _sampProviders = sampProviders;
    _playPositions.resize(sampProviders.size());
    _playVols.resize(sampProviders.size());
    
    for (int i = 0; i < _numVoices; i++)
        _sources.push_back(new AudioSource(_sampProviders, _envelope,
                                           _duration, _pitch, _timeStretch));

    SetVolumeDb(0.0);

    _range = 10.0;
    
    SetOverlap(1.0f);

    _localTime = 0;
    _bangTime = _duration * SRATE * (double) 0.001 / _overlap;

    for (int i = 0; i < _sources.size(); i++)
        _sources[i]->SetDurationMs(_duration);

    InitParameters();
    _currentParamIdx = NO_PARAMETER;

    double paramsSensivity[NUM_PARAMS] =
    {
        3.0, // NUMGRAINS
        3.0, // DURATION
        6.0, // WINDOW
        2.0, // MOTIONX
        2.0, // MOTIONY
        1.0, // MOTIONXY
        6.0, // DIRECTION
        3.0, // OVERLAP
        1.0, // PITCH
        4.0, // P_LFO_FREQ
        3.0, // P_LFO_AMT
        6.0, // SPATIALIZE
        6.0, // VOLUME
        4.0, // RANGE
        6.0  // TIME_STRETCH
    };
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->SetSensivity(paramsSensivity[i]);
}

GrainCloud::~GrainCloud()
{
    for (int i = 0; i < _sources.size(); i++)
        delete _sources[i];

    if (_isClone)
        return;
    
    if (_uiGrainCloud)
        delete _uiGrainCloud;

    for (int i = 0; i < NUM_PARAMS; i++)
        delete _params[i];
}

Source *
GrainCloud::Clone() const
{
    GrainCloud *clone = new GrainCloud(_app, _sampProviders, _numVoices);
    Source::Clone(clone);

    clone->_uiGrainCloud = _uiGrainCloud;

    clone->_addFlag = _addFlag;
    clone->_removeFlag = _removeFlag;

    clone->_localTime = _localTime;
    clone->_bangTime = _bangTime;

    clone->_nextGrain = _nextGrain;
    
    clone->_currentAroundChan = _currentAroundChan;
    clone->_stereoSide = _stereoSide; 
    clone->_side = _side;
    
    clone->_channelMults = _channelMults;
    clone->_spatialMode = _spatialMode;
    clone->_channelLocation = _channelLocation;

    clone->_volumeDb = _volumeDb;
    clone->_normedVol = _normedVol;

    clone->_range = _range;

    // Prevent memory leak...
    for (int i = 0; i < clone->_sources.size(); i++)
        delete clone->_sources[i];
            
    clone->_sources.resize(_sources.size());
    for (int i = 0; i < _sources.size(); i++)
        clone->_sources[i] = _sources[i]->Clone();
    
    clone->_numVoices = _numVoices;

    clone->_overlap = _overlap;
    clone->_overlapNorm = _overlapNorm;
    clone->_pitch = _pitch;
    clone->_duration = _duration;
    clone->_pitchLFOFreq = _pitchLFOFreq;
    clone->_pitchLFOAmount = _pitchLFOAmount;
    
    clone->_dirMode = _dirMode;
    clone->_envelopeType = _envelopeType;

    for (int i = 0; i < NUM_PARAMS; i++)
        clone->_params[i] = _params[i];

    clone->_currentParamIdx = _currentParamIdx;

    return clone;
}

void
GrainCloud::RegisterUI(UISource *uiGrainCloud)
{
    _uiGrainCloud = (UIGrainCloud *)uiGrainCloud;
    _uiGrainCloud->SetDuration(_duration);
}

UIGrainCloud *
GrainCloud::GetUI()
{
    return _uiGrainCloud;
}

void
GrainCloud::SetEnvelopeType(int envType)
{
    int numEnvs = _envelope->GetNumEnvelopes();
    _envelopeType = envType % numEnvs;
    
    if (_envelopeType < 0)
        _envelopeType = _envelope->GetNumEnvelopes() - 1;

    for (int i = 0; i < _sources.size();i++)
        _sources[i]->SetEnvelopeType(_envelopeType);
}

int
GrainCloud::GetEnvelopeType()
{
    return _envelopeType;
}

void
GrainCloud::AddGrain()
{
    _addFlag = true;
    _uiGrainCloud->AddGrain();
}

void
GrainCloud::RemoveGrain()
{
    _removeFlag = true;
    _uiGrainCloud->RemoveGrain();
}

int
GrainCloud::GetNumGrains() const
{
    int numGrains = _uiGrainCloud->GetNumGrains();

    return numGrains;
}

void
GrainCloud::SetOverlap(double target)
{
    if (target > 1.0f)
        target = 1.0f;
    else if (target < 0.0f)
        target = 0.0f;
    _overlapNorm = target;
 
    double num = (double)_sources.size();

    _overlap = exp(log(num)*target);

    UpdateBangTime();
}

double
GrainCloud::GetOverlap()
{
    return _overlapNorm;
}

void
GrainCloud::SetDurationMs(double dur)
{
    _duration = dur;
    
    for (int i = 0; i < _sources.size(); i++)
        _sources[i]->SetDurationMs(_duration);
    
    UpdateBangTime();
    
    if (_uiGrainCloud)
        _uiGrainCloud->SetDuration(_duration);
}

void
GrainCloud::Serialize(Serializer *serializer)
{
    Source::Serialize(serializer);
    
    serializer->PutBool(_addFlag);
    serializer->PutBool(_removeFlag);
    serializer->PutLong(_localTime);
    serializer->PutDouble(_bangTime);
    
    serializer->PutInt(_currentAroundChan);
    serializer->PutInt(_stereoSide);
    serializer->PutInt(_side);
    
    serializer->PutInt(_spatialMode);
    serializer->PutInt(_channelLocation);
    serializer->PutDouble(_volumeDb);
    serializer->PutDouble(_normedVol);

    serializer->PutDouble(_range);
    
    serializer->PutInt(_numVoices);

    serializer->PutDouble(_overlap);
    serializer->PutDouble(_overlapNorm);
    serializer->PutDouble(_pitch);
    serializer->PutDouble(_duration);
    serializer->PutDouble(_pitchLFOFreq);
    serializer->PutDouble(_pitchLFOAmount);

    serializer->PutInt(_dirMode);
    serializer->PutInt(_envelopeType);

    SerializeParameters(serializer);
    serializer->PutInt(_currentParamIdx);
}

void
GrainCloud::UnSerialize(Serializer *serializer)
{
    Source::UnSerialize(serializer);
    
    serializer->GetBool(&_addFlag);
    serializer->GetBool(&_removeFlag);
    serializer->GetLong((long *)&_localTime);
    serializer->GetDouble(&_bangTime);

    serializer->GetInt(&_currentAroundChan);
    serializer->GetInt(&_stereoSide);
    serializer->GetInt(&_side);
    
    serializer->GetInt(&_spatialMode);
    serializer->GetInt(&_channelLocation);
    serializer->GetDouble(&_volumeDb);
    serializer->GetDouble(&_normedVol);

    serializer->GetDouble(&_range);
    
    serializer->GetInt((int *)&_numVoices);

    serializer->GetDouble(&_overlap);
    serializer->GetDouble(&_overlapNorm);
    serializer->GetDouble(&_pitch);
    serializer->GetDouble(&_duration);
    serializer->GetDouble(&_pitchLFOFreq);
    serializer->GetDouble(&_pitchLFOAmount);

    serializer->GetInt(&_dirMode);
    serializer->GetInt(&_envelopeType);

    UnSerializeParameters(serializer);
    serializer->GetInt((int *)&_currentParamIdx);
}

Parameter *
GrainCloud::GetParameter(int paramIdx)
{
    return GetParameter((ParamIndex)paramIdx);
}

int
GrainCloud::GetNumParams() const
{
    return (int)NUM_PARAMS;
}

double
GrainCloud::GetUIX() const
{
    return _uiGrainCloud->GetX();
}

double
GrainCloud::GetUIY() const
{
    return _uiGrainCloud->GetY();
}

double
GrainCloud::GetUIRadius() const
{
    return _uiGrainCloud->GetRadius();
}

void
GrainCloud::UpdateBangTime()
{
    _bangTime = _duration*SRATE*(double)0.001/_overlap;
}

void
GrainCloud::SetPitch(double targetPitch)
{
    if (targetPitch < 0.0001)
        targetPitch = 0.0001;
    
    _pitch = targetPitch;

    double notePitch = GetMidiNotePitch();
    double wheelPitch = GetMidiPitchWheelValue();
    for (int i = 0; i < _sources.size(); i++)
        _sources[i]->SetPitch(targetPitch*notePitch*wheelPitch);
}

double
GrainCloud::GetPitch()
{
    return _pitch;
}

void
GrainCloud::SetVolumeDb(double volDb)
{
    if (volDb > 6.0)
        volDb = 6.0;
    
    if (volDb < -60.0)
        volDb = -60.0;
    
    _volumeDb = volDb;
    
    _normedVol = pow(10.0 , volDb*0.05);

    for (int i = 0; i < _sources.size(); i++)
        _sources[i]->SetVolume(_normedVol);
}

double
GrainCloud::GetVolumeDb()
{
    return _volumeDb;
}

void
GrainCloud::SetRange(double range)
{
    _range = range;

    for (int i = 0; i < _sources.size(); i++)
        _sources[i]->SetRange(range);
}

double
GrainCloud::GetRange()
{
    return _range;
}

void
GrainCloud::SetDirection(int dirMode)
{
    _dirMode = dirMode % 3;
    if (_dirMode < 0)
        _dirMode = 2;
    
    switch (_dirMode)
    {
        case FORWARD:
                for (int i = 0; i < _sources.size(); i++)
                    _sources[i]->SetDirection(1.0);
            break;
            
        case BACKWARD:
            for (int i = 0; i < _sources.size(); i++)
                _sources[i]->SetDirection(-1.0);
            break;
            
        case RANDOM_DIR:
            for (int i = 0; i < _sources.size(); i++)
            {
                if (randf() > 0.5)
                    _sources[i]->SetDirection(1.0);
                else
                    _sources[i]->SetDirection(-1.0);
            }
            break;
            
        default:
            break;
    }
}

int
GrainCloud::GetDirection()
{
    return _dirMode;
}

double
GrainCloud::GetDurationMs()
{
    return _duration;
}

void
GrainCloud::NextBuffer(vector<SAMPLE> *accumBuff, unsigned int numFrames)
{
    if (_addFlag)
    {
        _addFlag = false;

        _sources.push_back(new AudioSource(_sampProviders, _envelope,
                                           _duration, _pitch, _timeStretch));
        int idx = _sources.size() - 1;
        _sources[idx]->SetEnvelopeType(_envelopeType);
        _sources[idx]->SetTimeStretch(_timeStretch);
        
        switch (_dirMode)
        {
            case FORWARD:
                _sources[idx]->SetDirection(1.0);
                break;
                
            case BACKWARD:
                _sources[idx]->SetDirection(-1.0);
                break;
                
            case RANDOM_DIR:
                if (randf() > 0.5)
                    _sources[idx]->SetDirection(1.0);
                else
                    _sources[idx]->SetDirection(-1.0);
                break;
                    
            default:
                break;
        }
        
        _sources[idx]->SetVolume(_normedVol);
        _numVoices += 1;
        SetOverlap(_overlapNorm);
    }

    
    if (_removeFlag)
    {   
        if (_sources.size() > 1)
        {
            if (_nextGrain >= _sources.size()-1)
                _nextGrain = 0;
            
            _sources.pop_back();
        }
        
        SetOverlap(_overlapNorm);
        
        _removeFlag = false;
    }
    
    if (!_isMuted)
    {    
        for (int i = 0; i < _playPositions.size(); i++)
            _playPositions[i] = -1.0;
        
        for (int i = 0; i < _playVols.size(); i++)
            _playVols[i] = 0.0;

        for (int k = 0; k < _sources.size(); k++)
            _sources[k]->SetTimeStretch(_timeStretch);
            
        unsigned int nextFrame = 0;
        int frameSkip = numFrames/2;
        for (int j = 0; j < numFrames/frameSkip; j++)
        {    
            if (_localTime > _bangTime)
            {    
                _localTime = 0;
                
                for (int k = 0; k < _playPositions.size(); k++)
                {
                    _playPositions[k] = (double)(-1.0);
                    _playVols[k] = (double) 0.0;
                }
               
                if (_uiGrainCloud)
                    _uiGrainCloud->GetTriggerPos(_nextGrain,
                                                 _playPositions.data(),
                                                 _playVols.data(),
                                                 _duration);
                
                _sources[_nextGrain]->UpdateParams();
                
                if ((_pitchLFOAmount > 0.0f) && (_pitchLFOFreq > 0.0f))
                {
                    double nextPitch =
                        fabs(_pitch + _pitchLFOAmount *
                             sin(2*M_PI*_pitchLFOFreq*Time::Instance().
                                 GetSeconds()));

                    double notePitch = GetMidiNotePitch();
                    double wheelPitch = GetMidiPitchWheelValue();
                    _sources[_nextGrain]->SetPitch(nextPitch*notePitch*wheelPitch);
                }
                
                UpdateSpatialization();

                _sources[_nextGrain]->
                    SetChannelMultipliers(_channelMults);

                bool newGrainTriggered = false;
                newGrainTriggered = _sources[_nextGrain]->Play(_playPositions.data(),
                                                               _playVols.data());
                
                if (newGrainTriggered)
                {
                    _nextGrain++;
                    
                    if (_nextGrain >= _sources.size())
                        _nextGrain = 0;
                }
            }
            
            _localTime += frameSkip;
            
            nextFrame = j*frameSkip;
            
            for (int k = 0; k < _sources.size(); k++)
            {
                _sources[k]->UpdateParams();

                double notePitch = GetMidiNotePitch();
                double wheelPitch = GetMidiPitchWheelValue();
                _sources[k]->SetPitch(_pitch*notePitch*wheelPitch);
        
                _sources[k]->NextBuffer(accumBuff, frameSkip, nextFrame);
            }
        }
    }
}

void
GrainCloud::SetPitchLFOFreq(double pfreq)
{
    _pitchLFOFreq = fabs(pfreq);
}

void
GrainCloud::SetPitchLFOAmount(double lfoamt)
{
    if (lfoamt < 0.0)
        lfoamt = 0.0f;
    
    _pitchLFOAmount = lfoamt;
}

double
GrainCloud::GetPitchLFOFreq()
{
    return _pitchLFOFreq;
}

double
GrainCloud::GetPitchLFOAmount()
{
    return _pitchLFOAmount;
}

void
GrainCloud::SetSpatialMode(int mode,
                           int channelNumber = -1)
{
    _spatialMode = mode % 3;
    if (_spatialMode < 0)
        _spatialMode = 2;
    
    if (channelNumber >= 0)
        _channelLocation = channelNumber;
}

int
GrainCloud::GetSpatialMode()
{
    return _spatialMode;
}

int
GrainCloud::GetSpatialChannel()
{
    return _channelLocation;
}

void
GrainCloud::UpdateSpatialization()
{
    switch (_spatialMode)
    {
        case UNITY:
            for (int i = 0; i < CHANNELS; i++)
                _channelMults[i] = 0.999f;
            break;
            
        case STEREO:
            if (_stereoSide == 0)
            {
                for (int i = 0; i < CHANNELS; i++)
                {
                    _channelMults[i] = 0.0f;
                    if ((i % 2) == 0)
                        _channelMults[i] = 0.999f;
                    else
                        _channelMults[i] = 0.0f;
                }
                _stereoSide = 1;
            }
            else
            {
                for (int i = 0; i < CHANNELS; i++)
                {
                    _channelMults[i] = 0.0f;
                    if ((i % 2) == 0)
                        _channelMults[i] = 0.0f;
                    else
                        _channelMults[i] = 0.999f;
                }
                _stereoSide = 0;
            }
            break;
            
        case AROUND:
            for (int i = 0; i < CHANNELS; i++)
                _channelMults[i] = 0;
           
            _channelMults[_currentAroundChan] = 0.999;
            _currentAroundChan += _side*2;
            if ((_currentAroundChan > CHANNELS) || (_currentAroundChan < 0))
            {
                _side = -1*_side;
                _currentAroundChan += _side*3;
            }
            break;
            
        default:
            break;
    }    
}

void
GrainCloud::SetCurrentParamIdx(GrainCloud::ParamIndex idx)
{
    _currentParamIdx = idx;
}

void
GrainCloud::ResetCurrentParameterIndex()
{
    _currentParamIdx = NO_PARAMETER;
}

Parameter *
GrainCloud::GetParameter(ParamIndex paramIdx)
{
    return _params[paramIdx];
}

Parameter *
GrainCloud::GetCurrentParameter()
{
    if (_currentParamIdx == NO_PARAMETER)
        return NULL;

    return _params[_currentParamIdx];
}

void
GrainCloud::InitParameters()
{
    _params[NUMGRAINS] = new ParameterNumGrains(_app, NUMGRAINS);
    _params[DURATION] = new ParameterDuration(_app, DURATION);
    _params[ENVELOPE] = new ParameterEnvelope(_app, ENVELOPE);
    _params[MOTIONX] = new ParameterMotionX(_app, MOTIONX);
    _params[MOTIONY] = new ParameterMotionY(_app, MOTIONY);
    _params[MOTIONXY] = new ParameterMotionXY(_app, MOTIONXY);
    _params[DIRECTION] = new ParameterDirection(_app, DIRECTION);
    _params[OVERLAP] = new ParameterOverlap(_app, OVERLAP);
    _params[PITCH] = new ParameterPitch(_app, PITCH);
    _params[P_LFO_FREQ] = new ParameterLFOFreq(_app, P_LFO_FREQ);
    _params[P_LFO_AMP] = new ParameterLFOAmp(_app, P_LFO_AMP);
    _params[SPATIALIZE] = new ParameterSpatialize(_app, SPATIALIZE);
    _params[VOLUME] = new ParameterVolume(_app, VOLUME);
    _params[RANGE] = new ParameterRange(_app, RANGE);
    _params[TIME_STRETCH] = new ParameterTimeStretch(_app, TIME_STRETCH);
}

void
GrainCloud::SerializeParameters(Serializer *serializer)
{
    int size = NUM_PARAMS;
    serializer->PutInt(size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
GrainCloud::UnSerializeParameters(Serializer *serializer)
{
    int size;
    serializer->GetInt(&size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);
}
