/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef GRAIN_CLOUD_H
#define GRAIN_CLOUD_H

#include <map>
#include <vector>
#include <iostream>
#include <string>
#include <cstdlib>
#include <time.h>
#include <ctime>

#include "Stk.h"

#include "Source.h"
#include "AudioSource.h"
#include "Globals.h"
#include "Envelope.h"
#include "Thread.h"
#include "UISoundRect.h"
#include "UIGrainCloud.h"

using namespace std;

class UIStyle;
class Serializer;
class Parameter;
class App;
class SamplesProvider;
class GrainCloud : public Source
{    
public:
    enum ParamIndex
    {
        NO_PARAMETER = -1,
        NUMGRAINS, // = 0,
        DURATION,
        ENVELOPE,
        MOTIONX,
        MOTIONY,
        MOTIONXY,
        DIRECTION,
        OVERLAP,
        PITCH,
        P_LFO_FREQ,
        P_LFO_AMP,
        SPATIALIZE,
        VOLUME,
        RANGE,
        TIME_STRETCH,
        NUM_PARAMS
    };
    
    enum
    {
        FORWARD,
        BACKWARD,
        RANDOM_DIR
    };

    enum
    {
        UNITY,
        STEREO,
        AROUND
    };

    GrainCloud(App *app,
               const vector<SamplesProvider *> &sampProviders, double numVoices);
  
    virtual ~GrainCloud();

    Source *Clone() const override;
    
    void NextBuffer(vector<SAMPLE> *accumBuff, unsigned int numFrames) override;
    
    void SetDurationMs(double dur);
    double GetDurationMs();
    
    void SetOverlap(double targetOverlap);
    double GetOverlap();
    
    void SetPitch(double targetPitch);
    double GetPitch();
   
    void SetPitchLFOFreq(double pfreq);
    double GetPitchLFOFreq();
    void SetPitchLFOAmount(double lfoamt);
    double GetPitchLFOAmount();
    
    void SetDirection(int dirMode);
    int GetDirection();
    
    void AddGrain();
    void RemoveGrain();
    int GetNumGrains() const;
    
    void SetEnvelopeType(int envelopeType);
    int GetEnvelopeType();
    
    void SetSpatialMode(int mode, int channelNumber);
    int GetSpatialMode();
    int GetSpatialChannel();
    
    void SetVolumeDb(double volDb);
    double GetVolumeDb();

    void SetRange(double range);
    double GetRange();
    
    void RegisterUI(UISource *uiGrainCloud) override;
    UIGrainCloud *GetUI() override;

    Parameter *GetParameter(ParamIndex paramIdx);
    void SetCurrentParamIdx(ParamIndex idx);
    void ResetCurrentParameterIndex() override;
    
    void Serialize(Serializer *serializer) override;
    void UnSerialize(Serializer *serializer) override;

    // ParameterProvider
    Parameter *GetParameter(int paramIdx) override;
    Parameter *GetCurrentParameter() override;
    int GetNumParams() const override;
    double GetUIX() const override;
    double GetUIY() const override;
    double GetUIRadius() const override;
    
protected:
    void SerializeParameters(Serializer *serializer);
    void UnSerializeParameters(Serializer *serializer);
    
    void InitParameters();
    
    void UpdateBangTime();
    
    void UpdateSpatialization();
    
    App *_app;
    UIGrainCloud *_uiGrainCloud;
    
    bool _addFlag;
    bool _removeFlag;

    unsigned long _localTime;
    double _bangTime;

    unsigned int _nextGrain;
    
    int _currentAroundChan;
    int _stereoSide; 
    int _side;
    
    vector<SAMPLE> _channelMults;
    int _spatialMode;
    int _channelLocation;

    double _volumeDb;
    double _normedVol;

    double _range;
    
    vector<AudioSource *> _sources;
    
    unsigned int _numVoices;

    double _overlap;
    double _overlapNorm;
    double _pitch;
    double _duration;
    double _pitchLFOFreq;
    double _pitchLFOAmount;
    
    int _dirMode;
    int _envelopeType;
    
    vector<SamplesProvider *> _sampProviders;
    vector<double> _playPositions;
    vector<double> _playVols;
    
    Parameter *_params[NUM_PARAMS];
    ParamIndex _currentParamIdx;
};

#endif
