/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "UIStyle.h"
#include "UILoopSource.h"
#include "AudioSource.h"
#include "SamplesProvider.h"
#include "Serializer.h"
#include "Parameter.h"
#include "App.h"
#include "LoopSource.h"

LoopSource::LoopSource(App *app, const vector<SamplesProvider*> &sampProviders)
{
    _sourceType = Source::LOOP_SOURCE;
    
    _app = app;

    _uiLoopSource = NULL;
    
    srand(time(NULL));
    
    _duration = 500.0;

    _pitch = 1.0f;
    
    _envelopeType = Envelope::HANNING;
    
    _pitchLFOFreq = 0.01f;
    _pitchLFOAmount = 0.0f;
    
    _channelMults.resize(CHANNELS);
    for (int i = 0; i < CHANNELS; i++)
        _channelMults[i] = 0.999f;
    
    _currentAroundChan = 1;
    _stereoSide = 0; 
    _side = 1;
    
    _spatialMode = UNITY;
    _channelLocation = -1;
    
    _dirMode = FORWARD;

    _playPositions.resize(sampProviders.size());
    _playVols.resize(sampProviders.size());
    
    _source = new AudioSource(sampProviders, _envelope, _duration, _pitch, _timeStretch);
    _source->SetLoop(true);
    
    SetVolumeDb(0.0);

    _range = 10.0;
    
    _bufferPos = 0;
        
    _source->SetDurationMs(_duration);

    InitParameters();
    _currentParamIdx = NO_PARAMETER;

    double paramsSensivity[NUM_PARAMS] =
    {
        3.0, // DURATION
        6.0, // WINDOW
        6.0, // DIRECTION
        1.0, // PITCH
        4.0, // P_LFO_FREQ
        3.0, // P_LFO_AMT
        6.0, // SPATIALIZE
        6.0, // VOLUME
        4.0, // RANGE
        6.0  // TIME_STRETCH
    };
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->SetSensivity(paramsSensivity[i]);
}

LoopSource::~LoopSource()
{
    delete _source;
    
    if (_isClone)
        return;
    
    if (_uiLoopSource)
        delete _uiLoopSource;

    for (int i = 0; i < NUM_PARAMS; i++)
        delete _params[i];
}

Source *
LoopSource::Clone() const
{
    const vector<SamplesProvider*> &sampProviders = _source->GetSampleProviders();
    LoopSource *clone = new LoopSource(_app, sampProviders);
    Source::Clone(clone);

    clone->_app = _app;
    clone->_uiLoopSource = _uiLoopSource;
    
    clone->_currentAroundChan = _currentAroundChan;
    clone->_stereoSide = _stereoSide; 
    clone->_side = _side;
    
    clone->_channelMults = _channelMults;
    clone->_spatialMode = _spatialMode;
    clone->_channelLocation = _channelLocation;

    clone->_volumeDb = _volumeDb;
    clone->_normedVol = _normedVol;

    clone->_range = _range;

    // Prevent memory leak...
    delete clone->_source;
    
    clone->_source = _source->Clone();

    clone->_bufferPos = _bufferPos;
    
    clone->_pitch = _pitch;
    clone->_duration = _duration;
    clone->_pitchLFOFreq = _pitchLFOFreq;
    clone->_pitchLFOAmount = _pitchLFOAmount;
    
    clone->_dirMode = _dirMode;
    clone->_envelopeType = _envelopeType;

    for (int i = 0; i < NUM_PARAMS; i++)
        clone->_params[i] = _params[i];
    clone->_currentParamIdx = _currentParamIdx;

    return clone;
}

void
LoopSource::RegisterUI(UISource *uiLoopSource)
{
    _uiLoopSource = (UILoopSource *)uiLoopSource;
    _uiLoopSource->SetDuration(_duration);
}

UILoopSource *
LoopSource::GetUI()
{
    return _uiLoopSource;
}

void
LoopSource::SetEnvelopeType(int envType)
{
    int numEnvs = _envelope->GetNumEnvelopes();
    _envelopeType = envType % numEnvs;
    
    if (_envelopeType < 0)
        _envelopeType = _envelope->GetNumEnvelopes() - 1;

    _source->SetEnvelopeType(_envelopeType);
}

int
LoopSource::GetEnvelopeType()
{
    return _envelopeType;
}

void
LoopSource::SetDurationMs(double dur)
{
    _duration = dur;
    
    _source->SetDurationMs(_duration);   
    
    if (_uiLoopSource)
        _uiLoopSource->SetDuration(_duration);
}

void
LoopSource::Serialize(Serializer *serializer)
{
    Source::Serialize(serializer);
    
    serializer->PutInt(_currentAroundChan);
    serializer->PutInt(_stereoSide);
    serializer->PutInt(_side);
    
    serializer->PutInt(_spatialMode);
    serializer->PutInt(_channelLocation);
    serializer->PutDouble(_volumeDb);
    serializer->PutDouble(_normedVol);

    serializer->PutDouble(_range);
    
    serializer->PutDouble(_pitch);
    serializer->PutDouble(_duration);
    serializer->PutDouble(_pitchLFOFreq);
    serializer->PutDouble(_pitchLFOAmount);

    serializer->PutInt(_dirMode);
    serializer->PutInt(_envelopeType);

    SerializeParameters(serializer);
    serializer->PutInt(_currentParamIdx);
}

void
LoopSource::UnSerialize(Serializer *serializer)
{
    Source::UnSerialize(serializer);
    
    serializer->GetInt(&_currentAroundChan);
    serializer->GetInt(&_stereoSide);
    serializer->GetInt(&_side);
    
    serializer->GetInt(&_spatialMode);
    serializer->GetInt(&_channelLocation);
    serializer->GetDouble(&_volumeDb);
    serializer->GetDouble(&_normedVol);

    serializer->GetDouble(&_range);
    
    serializer->GetDouble(&_pitch);
    serializer->GetDouble(&_duration);
    serializer->GetDouble(&_pitchLFOFreq);
    serializer->GetDouble(&_pitchLFOAmount);

    serializer->GetInt(&_dirMode);
    serializer->GetInt(&_envelopeType);

    UnSerializeParameters(serializer);
    serializer->GetInt((int *)&_currentParamIdx);
}

Parameter *
LoopSource::GetParameter(int paramIdx)
{
    return GetParameter((ParamIndex)paramIdx);
}

int
LoopSource::GetNumParams() const
{
    return (int)NUM_PARAMS;
}

double
LoopSource::GetUIX() const
{
    return _uiLoopSource->GetX();
}

double
LoopSource::GetUIY() const
{
    return _uiLoopSource->GetY();
}

double
LoopSource::GetUIRadius() const
{
    return _uiLoopSource->GetRadius();
}

void
LoopSource::SetPitch(double targetPitch)
{
    if (targetPitch < 0.0001)
        targetPitch = 0.0001;
    
    _pitch = targetPitch;

    double notePitch = GetMidiNotePitch();
    double wheelPitch = GetMidiPitchWheelValue();
    _source->SetPitch(targetPitch*notePitch*wheelPitch);
}

double
LoopSource::GetPitch()
{
    return _pitch;
}

void
LoopSource::SetVolumeDb(double volDb)
{
    if (volDb > 6.0)
        volDb = 6.0;
    
    if (volDb < -60.0)
        volDb = -60.0;
    
    _volumeDb = volDb;
    
    _normedVol = pow(10.0, volDb*0.05);

    _source->SetVolume(_normedVol);
}

double
LoopSource::GetVolumeDb()
{
    return _volumeDb;
}

void
LoopSource::SetRange(double range)
{
    _range = range;

    _source->SetRange(range);
}

double
LoopSource::GetRange()
{
    return _range;
}

void
LoopSource::SetDirection(int dirMode)
{
    _dirMode = dirMode % 3;
    if (_dirMode < 0)
        _dirMode = 2;
    
    switch (_dirMode)
    {
        case FORWARD:
            _source->SetDirection(1.0);
            break;
            
        case BACKWARD:
            _source->SetDirection(-1.0);
            break;
            
        case RANDOM_DIR:
            if (randf() > 0.5)
                _source->SetDirection(1.0);
            else
                _source->SetDirection(-1.0);
            break;
            
        default:
            break;
    }
}

int
LoopSource::GetDirection()
{
    return _dirMode;
}

double
LoopSource::GetDurationMs()
{
    return _duration;
}

void
LoopSource::UpdateSource()
{
    _source->SetDurationMs(_duration);

    double notePitch = GetMidiNotePitch();
    double wheelPitch = GetMidiPitchWheelValue();
    _source->SetPitch(_pitch*notePitch*wheelPitch);
    _source->SetTimeStretch(_timeStretch);
    _source->SetEnvelopeType(_envelopeType);
        
    _source->SetVolume(_normedVol);

    _source->SetRange(_range);
}

void
LoopSource::NextBuffer(vector<SAMPLE> *accumBuff, unsigned int numFrames)
{
    if (_isMuted)
        return;
    
    for (int i = 0; i < _playPositions.size(); i++)
        _playPositions[i] = -1.0;
        
    for (int i = 0; i < _playVols.size(); i++)
        _playVols[i] = 0.0;
    
    if (_uiLoopSource)
        _uiLoopSource->GetTriggerPos(_playPositions.data(),
                                     _playVols.data(),
                                     _duration);
    
    _source->Play(_playPositions.data(), _playVols.data());
    _source->Update(_playPositions.data(), _playVols.data());

    UpdateSource();
        
    if ((_pitchLFOAmount > 0.0f) && (_pitchLFOFreq > 0.0f))
    {
        double nextPitch =
            fabs(_pitch + _pitchLFOAmount *
                 sin(2*PI*_pitchLFOFreq*Time::Instance().
                     GetSeconds()));

        double notePitch = GetMidiNotePitch();
        double wheelPitch = GetMidiPitchWheelValue();
        _source->SetPitch(nextPitch*notePitch*wheelPitch);
    }
        
    _source->SetChannelMultipliers(_channelMults);
    
    bool newLoop = _source->NextBuffer(accumBuff, numFrames, _bufferPos);
    if (newLoop)
    {
        UpdateSpatialization();
        // Possibly update random direction
        SetDirection(_dirMode);
    }
}

void
LoopSource::SetPitchLFOFreq(double pfreq)
{
    _pitchLFOFreq = fabs(pfreq);
}

void
LoopSource::SetPitchLFOAmount(double lfoamt)
{
    if (lfoamt < 0.0)
        lfoamt = 0.0f;
    
    _pitchLFOAmount = lfoamt;
}

double
LoopSource::GetPitchLFOFreq()
{
    return _pitchLFOFreq;
}

double
LoopSource::GetPitchLFOAmount()
{
    return _pitchLFOAmount;
}

void
LoopSource::SetSpatialMode(int mode,
                           int channelNumber = -1)
{
    _spatialMode = mode % 3;
    if (_spatialMode < 0)
        _spatialMode = 2;
    
    if (channelNumber >= 0)
        _channelLocation = channelNumber;
}

int
LoopSource::GetSpatialMode()
{
    return _spatialMode;
}

int
LoopSource::GetSpatialChannel()
{
    return _channelLocation;
}

void
LoopSource::UpdateSpatialization()
{
    switch (_spatialMode)
    {
        case UNITY:
            for (int i = 0; i < CHANNELS; i++)
                _channelMults[i] = 0.999f;
            break;
            
        case STEREO:
            if (_stereoSide == 0)
            {
                for (int i = 0; i < CHANNELS; i++)
                {
                    _channelMults[i] = 0.0f;
                    if ((i % 2) == 0)
                        _channelMults[i] = 0.999f;
                    else
                        _channelMults[i] = 0.0f;
                }
                _stereoSide = 1;
            }
            else
            {
                for (int i = 0; i < CHANNELS; i++)
                {
                    _channelMults[i] = 0.0f;
                    if ((i % 2) == 0)
                        _channelMults[i] = 0.0f;
                    else
                        _channelMults[i] = 0.999f;
                }
                _stereoSide = 0;
            }
            break;
            
        case AROUND:
            for (int i = 0; i < CHANNELS; i++)
                _channelMults[i] = 0;
            
            _channelMults[_currentAroundChan] = 0.999;
            _currentAroundChan += _side*2;
            if ((_currentAroundChan > CHANNELS) ||
                (_currentAroundChan < 0))
            {
                _side = -_side;
                _currentAroundChan += _side*3;
            }
            break;
            
        default:
            break;
    }    
}

void
LoopSource::SetCurrentParamIdx(LoopSource::ParamIndex idx)
{
    _currentParamIdx = idx;
}

void
LoopSource::ResetCurrentParameterIndex()
{
    _currentParamIdx = NO_PARAMETER;
}

Parameter *
LoopSource::GetParameter(ParamIndex paramIdx)
{
    return _params[paramIdx];
}

Parameter *
LoopSource::GetCurrentParameter()
{
    if (_currentParamIdx == NO_PARAMETER)
        return NULL;

    return _params[_currentParamIdx];
}

void
LoopSource::InitParameters()
{
    _params[DURATION] = new ParameterDuration(_app, DURATION);
    _params[ENVELOPE] = new ParameterEnvelope(_app, ENVELOPE);
    _params[DIRECTION] = new ParameterDirection(_app, DIRECTION);
    _params[PITCH] = new ParameterPitch(_app, PITCH);
    _params[P_LFO_FREQ] = new ParameterLFOFreq(_app, P_LFO_FREQ);
    _params[P_LFO_AMP] = new ParameterLFOAmp(_app, P_LFO_AMP);
    _params[SPATIALIZE] = new ParameterSpatialize(_app, SPATIALIZE);
    _params[VOLUME] = new ParameterVolume(_app, VOLUME);
    _params[RANGE] = new ParameterRange(_app, RANGE);
    _params[TIME_STRETCH] = new ParameterTimeStretch(_app, TIME_STRETCH);
}

void
LoopSource::SerializeParameters(Serializer *serializer)
{
    int size = NUM_PARAMS;
    serializer->PutInt(size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
LoopSource::UnSerializeParameters(Serializer *serializer)
{
    int size;
    serializer->GetInt(&size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);
}
