/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "RtMidi.h"

#include "MidiQueue.h"
#include "Globals.h"
#include "Thread.h"
#include "MyRtMidiIn.h"

static void
_callback(double deltatime, vector<unsigned char> *message, void *userData)
{
    MyRtMidiIn *rtMidi = (MyRtMidiIn *)userData;
    rtMidi->Callback(deltatime, message);
}

MyRtMidiIn::MyRtMidiIn(Mutex *mutex)
{
    _mutex = mutex;
    
    _midiIn = NULL;

    _midiQueue = new MidiQueue();
  
    try
    {
        _midiIn = new RtMidiIn();
    }
    catch (RtMidiError &error)
    {
        error.printMessage();
    }
}

MyRtMidiIn::~MyRtMidiIn()
{
    if (_midiIn != NULL)
        delete _midiIn;

    delete _midiQueue;
}

int
MyRtMidiIn::Open(int midiInPort)
{
    if (_midiIn == NULL)
        return -1;

    int port = -1;
    try
    {
        port = ChooseMidiPort(midiInPort);
        if (port == -1)
            return -1;
        
        _midiIn->setCallback(&_callback, (void *)this);
        
        // Don't ignore sysex, timing, or active sensing messages.
        _midiIn->ignoreTypes(false, false, false);
    }
    catch (RtMidiError &error)
    {
        error.printMessage();
        return -1;
    }

    return port;
}

void
MyRtMidiIn::Callback(double deltatime, vector<unsigned char> *message)
{
    _mutex->lock();
    
#if 0
    unsigned int nBytes = message->size();
    for (unsigned int i = 0; i < nBytes; i++)
        fprintf(stdout, "Byte %d %d stamp=%d\n", i, (int)message->at(i), deltatime);
#endif

    _midiQueue->PushMessage(deltatime, *message);

    _mutex->unlock();
}

MidiQueue *
MyRtMidiIn::GetMidiQueue()
{
    return _midiQueue;
}

int
MyRtMidiIn::ChooseMidiPort(int midiInPort)
{
#if 0 // Virtual input port
    _midiIn->openVirtualPort();
    return 0;
#endif

    string portName;
    unsigned int nPorts = _midiIn->getPortCount();

    // Display possible ports
    fprintf(stdout, "Midi Input Ports:\n");
    for (int i = 0; i < nPorts; i++)
    {
        portName = _midiIn->getPortName(i);
        fprintf(stdout, "  Input port #%d | %s\n", i, portName.c_str());
    }
    
    if (nPorts == 0)
    {
        fprintf(stdout, "No Midi input ports available!\n");

        return -1;
    }
    
    if (nPorts == 1)
    {
        fprintf(stdout, "Opening:  Input port #%d | %s\n",
                0, _midiIn->getPortName().c_str());
        _midiIn->openPort(0);
        
        return 0;
    }
    else
    {
        fprintf(stdout, "Opening:  Input port #%d | %s\n",
                midiInPort, _midiIn->getPortName(midiInPort).c_str());
        
        _midiIn->openPort(midiInPort);
    }

    return midiInPort;
}
