/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "App.h"
#include "SingleGrain.h"
#include "GrainCloud.h"
#include "LoopSource.h"

#include "UIStyle.h"
#include "UIParameter.h"
#include "UIGrainCloud.h"
#include "UISingleGrain.h"
#include "Serializer.h"
#include "Parameter.h"

Parameter::Parameter(App *app, int paramIdx)
{
    _app = app;
    _idx = paramIdx;

    _value = -1.0;
    _minValue = -1.0;
    _maxValue = -1.0;

    _sensivity = 1.0;

    _powFactor = 1.0;
    
    _uiParam = NULL;

    memset(_valueStr, 0, 255);
    
    if (_app != NULL)
    {
        UIScreen *screen = app->GetScreen();
        const UIStyle *style = app->GetStyle();
        _uiParam = new UIParameter(this, screen, (UIStyle *)style);
    }
}

Parameter::~Parameter()
{
    if (_uiParam != NULL)
        delete _uiParam;
}

double
Parameter::GetValue() const
{
    return _value;
}

double
Parameter::GetDefaultValue() const
{
    return _defaultValue;
}

const char *
Parameter::GetValueStr() const
{
    if (_uiParam != NULL)
        return _uiParam->GetValueStr();

    return _valueStr;
}

void
Parameter::SetNormalizedValue(double value)
{
    if (value < 0.0)
        value = 0.0;
    if (value > 1.0)
        value = 1.0;
    
    value = pow(value, _powFactor);
    
    _value = _minValue + value*(_maxValue - _minValue);
    
    SetValue(_value);
}

double
Parameter::GetNormalizedValue() const
{
    double value = (_value - _minValue)/(_maxValue - _minValue);

    value = pow(value, 1.0/_powFactor);
    
    return value;
}

void
Parameter::SetSensivity(double sensivity)
{
    _sensivity = sensivity;
}

double
Parameter::GetSensivity() const
{
    return _sensivity;
}

int
Parameter::GetIndex() const
{
    return _idx;
}

UIParameter *
Parameter::GetUI()
{
    return _uiParam;
}

void
Parameter::Serialize(Serializer *serializer)
{
    serializer->PutInt(_idx);
    
    serializer->PutDouble(_value);
    //serializer->PutDouble(_minValue);
    //serializer->PutDouble(_maxValue);

    serializer->PutDouble(_sensivity);

    if (_uiParam != NULL)
        _uiParam->Serialize(serializer);
}

void
Parameter::UnSerialize(Serializer *serializer)
{
    serializer->GetInt(&_idx);
    
    serializer->GetDouble(&_value);
    //serializer->GetDouble(&_minValue);
    //serializer->GetDouble(&_maxValue);
    
    SetValue(_value);

    serializer->GetDouble(&_sensivity);

    if (_uiParam != NULL)
        _uiParam->UnSerialize(serializer);
}

void
Parameter::SetName(const char *name)
{
    if (_uiParam != NULL)
        _uiParam->SetName(name);
}

void
Parameter::SetValueStr(const char *valueStr)
{
    strcpy(_valueStr, valueStr);
    
    if (_uiParam != NULL)
        _uiParam->SetValueStr(valueStr);
}

void
Parameter::CheckValueBounds(double *value)
{
    if (*value < _minValue)
        *value = _minValue;
    else if (*value > _maxValue)
        *value = _maxValue;
}

void
Parameter::SetValueCheckBounds(double value)
{
    CheckValueBounds(&value);
   
    _value = value;
}

ParameterNumGrains::ParameterNumGrains(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Voices");

    _minValue = 1.0;
    _maxValue = 100.0;
    _defaultValue = 8.0;
    _value = _defaultValue;
}

ParameterNumGrains::~ParameterNumGrains() {}

void
ParameterNumGrains::UpdateValue()
{
    char valueStr[255];
    
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
        {
            _value = ((GrainCloud *)source)->GetNumGrains();
            CheckValueBounds(&_value);
            sprintf(valueStr, "%d", (int)_value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

void
ParameterNumGrains::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);
    
    if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
    {
        int numGrainsToAdd = (int)_value - ((GrainCloud *)source)->GetNumGrains();
            
        if (numGrainsToAdd > 0)
        {    
            for (int i = 0; i < numGrainsToAdd; i++)
                ((GrainCloud *)source)->AddGrain();
        }
        else if (numGrainsToAdd < 0)
        {
            for (int i = 0; i < -numGrainsToAdd; i++)
                ((GrainCloud *)source)->RemoveGrain();
        }
    }
}

void
ParameterNumGrains::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

ParameterDuration::ParameterDuration(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Duration");

    _minValue = 0.5;
    //_maxValue = 1000.0;
    _maxValue = 5000.0; // using long samples
    _defaultValue = 500.0;
    _value = _defaultValue;

    _powFactor = 3.0;
}

ParameterDuration::~ParameterDuration() {}

void
ParameterDuration::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;

    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetDurationMs(_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetDurationMs(_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetDurationMs(_value);
    }
}

void
ParameterDuration::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterDuration::UpdateValue()
{
    char valueStr[255];
                
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetDurationMs();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetDurationMs();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetDurationMs();
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g ms", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterEnvelope::ParameterEnvelope(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Envelope");

    _minValue = Envelope::HANNING;
    _maxValue = Envelope::HAND_DRAWN;
    _defaultValue = Envelope::HANNING;
    _value = _defaultValue;
}

ParameterEnvelope::~ParameterEnvelope() {}

void
ParameterEnvelope::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetEnvelopeType((int)_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetEnvelopeType((int)_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetEnvelopeType((int)_value);
    }
}

void
ParameterEnvelope::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterEnvelope::UpdateValue()
{
    Source *source = _app->GetCurrentSource();
    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            _value = ((GrainCloud *)source)->GetEnvelopeType();
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            _value = ((SingleGrain *)source)->GetEnvelopeType();
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            _value = ((LoopSource *)source)->GetEnvelopeType();
        
        CheckValueBounds(&_value);
        
        char valueStr[255];
        switch ((int)_value)
        {
            case Envelope::HANNING:
                sprintf(valueStr, "HANNING");
                break;
                
            case Envelope::TRIANGLE:
                sprintf(valueStr, "TRIANGLE");
                break;
                
            case Envelope::REXPDEC:
                sprintf(valueStr, "REXPDEC");
            break;
            
            case Envelope::EXPDEC:
                sprintf(valueStr, "EXPDEC");
            break;
            
            case Envelope::SINC:
                sprintf(valueStr, "SINC");
                break;

            case Envelope::HAND_DRAWN:
                sprintf(valueStr, "HAND_DRAWN");
                break;
            
            default:
                sprintf(valueStr, "");
                break;
        }

        SetValueStr(valueStr);
    }
}
    
ParameterMotionX::ParameterMotionX(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("X");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterMotionX::~ParameterMotionX() {}

void
ParameterMotionX::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
        ((GrainCloud *)source)->GetUI()->SetXRandExtent(_value);
}

void
ParameterMotionX::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterMotionX::UpdateValue()
{
    // If the window was resized...
    _minValue = 0.0;
    _maxValue = 1.0;

    char valueStr[255];
                
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
        {
            _value = ((GrainCloud *)source)->GetUI()->GetXRandExtent();
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterMotionY::ParameterMotionY(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Y");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterMotionY::~ParameterMotionY() {}

void
ParameterMotionY::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
        ((GrainCloud *)source)->GetUI()->SetYRandExtent(_value);
}

void
ParameterMotionY::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterMotionY::UpdateValue()
{
    _minValue = 0.0;
    _maxValue = 1.0;

    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            _value = ((GrainCloud *)source)->GetUI()->GetYRandExtent();
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterMotionXY::ParameterMotionXY(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("XY");

    // This is a coeff for both x and y extents
    
    _minValue = 0.0;
    _maxValue = 4.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterMotionXY::~ParameterMotionXY() {}

void
ParameterMotionXY::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);
    
    // In theory here, we should manage x and y separately
    // Instead, we use a coeff on existing x and y extents

    if (source != NULL)
        ((GrainCloud *)source)->GetUI()->SetRandExtentCoeff(_value);
}

void
ParameterMotionXY::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterMotionXY::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            double xr = ((GrainCloud *)source)->GetUI()->GetXRandExtent();
            double yr = ((GrainCloud *)source)->GetUI()->GetYRandExtent();
            
            _value = ((GrainCloud *)source)->GetUI()->GetRandExtentCoeff();
            CheckValueBounds(&_value);

            xr *= _value;
            yr *= _value;
            
            sprintf(valueStr, "%g, %g", xr, yr);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterDirection::ParameterDirection(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Direction");

    // ...
    _minValue = GrainCloud::FORWARD;
    _maxValue = GrainCloud::RANDOM_DIR;
    _defaultValue = GrainCloud::FORWARD;
    _value = _defaultValue;
}

ParameterDirection::~ParameterDirection() {}

void
ParameterDirection::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetDirection(_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetDirection(_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetDirection(_value);
    }
}

void
ParameterDirection::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterDirection::UpdateValue()
{
    Source *source = _app->GetCurrentSource();
    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            _value = ((GrainCloud *)source)->GetDirection();
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            _value = ((SingleGrain *)source)->GetDirection();
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            _value = ((LoopSource *)source)->GetDirection();
        
        CheckValueBounds(&_value);
        
        char valueStr[255];
        switch((int)_value)
        {
            case GrainCloud::FORWARD:
                sprintf(valueStr, "FORWARD");
                break;
                
            case GrainCloud::BACKWARD:
                sprintf(valueStr, "BACKWARD");
                break;
                
            case GrainCloud::RANDOM_DIR:
                sprintf(valueStr, "RANDOM");
                break;
                
            default:
                sprintf(valueStr, "");
                break;                   
        }
        SetValueStr(valueStr);
    }
}
    
ParameterOverlap::ParameterOverlap(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Overlap");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 1.0;
    _value = _defaultValue;
}

ParameterOverlap::~ParameterOverlap() {}

void
ParameterOverlap::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
        ((GrainCloud *)source)->SetOverlap(_value);
}

void
ParameterOverlap::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterOverlap::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if ((source != NULL) && (source->GetSourceType() == Source::GRAIN_CLOUD))
        {
            _value = ((GrainCloud *)source)->GetOverlap();
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g ms", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterPitch::ParameterPitch(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Pitch");

    _minValue = 0.1;
    _maxValue = 5.0;
    _defaultValue = 1.0;
    _value = _defaultValue;
}

ParameterPitch::~ParameterPitch() {}

void
ParameterPitch::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetPitch(_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetPitch(_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetPitch(_value);
    }
}

void
ParameterPitch::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterPitch::UpdateValue()
{
    char valueStr[255];            
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetPitch();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetPitch();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetPitch();
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

ParameterLFOFreq::ParameterLFOFreq(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("LFO Freq");

    _minValue = 0.1;
    _maxValue = 1.0;
    _defaultValue = 1.0;
    _value = _defaultValue;
}

ParameterLFOFreq::~ParameterLFOFreq() {}

void
ParameterLFOFreq::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetPitchLFOFreq(_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetPitchLFOFreq(_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetPitchLFOFreq(_value);
    }
}

void
ParameterLFOFreq::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterLFOFreq::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetPitchLFOFreq();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetPitchLFOFreq();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetPitchLFOFreq();
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterLFOAmp::ParameterLFOAmp(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("LFO Amount");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterLFOAmp::~ParameterLFOAmp() {}

void
ParameterLFOAmp::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetPitchLFOAmount(_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetPitchLFOAmount(_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetPitchLFOAmount(_value);
    }
}

void
ParameterLFOAmp::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterLFOAmp::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetPitchLFOAmount();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetPitchLFOAmount();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetPitchLFOAmount();
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}
    
ParameterSpatialize::ParameterSpatialize(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Spatial Mode");

    // ...
    _minValue = GrainCloud::UNITY;
    _maxValue = GrainCloud::AROUND;
    _defaultValue = GrainCloud::UNITY;
    _value = _defaultValue;
}

ParameterSpatialize::~ParameterSpatialize() {}

void
ParameterSpatialize::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetSpatialMode((int)_value, -1);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetSpatialMode((int)_value, -1);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetSpatialMode((int)_value, -1);
    }
}

void
ParameterSpatialize::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterSpatialize::UpdateValue()
{
    char valueStr[255];
    Source *source = _app->GetCurrentSource();
    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            _value = ((GrainCloud *)source)->GetSpatialMode();
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            _value = ((SingleGrain *)source)->GetSpatialMode();
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            _value = ((LoopSource *)source)->GetSpatialMode();
        
        CheckValueBounds(&_value);
        switch((int)_value)
        {
            case GrainCloud::UNITY:
                sprintf(valueStr, "UNITY");
                break;
                
            case GrainCloud::STEREO:
                sprintf(valueStr, "STEREO");
                break;
                
            case GrainCloud::AROUND:
                sprintf(valueStr, "AROUND");
                break;
                
            default:
                sprintf(valueStr, "");
                break;
        }

        SetValueStr(valueStr);
    }
}
    
ParameterVolume::ParameterVolume(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Volume");

    _minValue = -6.0;
    _maxValue = 6.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterVolume::~ParameterVolume() {}

void
ParameterVolume::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;

    Parameter::SetValueCheckBounds(value);
    
    if (source->GetSourceType() == Source::GRAIN_CLOUD)
        ((GrainCloud *)source)->SetVolumeDb(_value);
    else if (source->GetSourceType() == Source::SINGLE_GRAIN)
        ((SingleGrain *)source)->SetVolumeDb(_value);
    else if (source->GetSourceType() == Source::LOOP_SOURCE)
        ((LoopSource *)source)->SetVolumeDb(_value);
}

void
ParameterVolume::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterVolume::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetVolumeDb();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetVolumeDb();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetVolumeDb();
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g dB", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

ParameterRange::ParameterRange(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Range");

    _minValue = 0.0;
    _maxValue = 100.0;
    _defaultValue = 50.0;
    _value = _defaultValue;
}

ParameterRange::~ParameterRange() {}

void
ParameterRange::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;

    Parameter::SetValueCheckBounds(value);
    
    if (source->GetSourceType() == Source::GRAIN_CLOUD)
        ((GrainCloud *)source)->SetRange(_value);
    else if (source->GetSourceType() == Source::SINGLE_GRAIN)
        ((SingleGrain *)source)->SetRange(_value);
    else if (source->GetSourceType() == Source::LOOP_SOURCE)
        ((LoopSource *)source)->SetRange(_value);
}

void
ParameterRange::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterRange::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetRange();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetRange();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetRange();
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g %%", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

ParameterSoundRectType::ParameterSoundRectType(App *app,
                                               int paramId)
: Parameter(app, paramId)
{
    SetName("Type");

    _minValue = UISoundRect::WAVEFORM;
    _maxValue = UISoundRect::CHROMAGRAM;
    _defaultValue = UISoundRect::WAVEFORM;
    _value = _defaultValue;
}

ParameterSoundRectType::~ParameterSoundRectType() {}

void
ParameterSoundRectType::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOUND_RECT)
        return;

    UISoundRect *rect = (UISoundRect *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (rect != NULL)
        rect->SetRectType((UISoundRect::RectType)_value);
}

void
ParameterSoundRectType::SetValue(double value)
{
    UISoundRect *rect = _app->GetCurrentSoundRect();
    SetValue(rect, value);
}

void
ParameterSoundRectType::UpdateValue()
{
    char valueStr[255];
    UISoundRect *rect = _app->GetCurrentSoundRect();
    if (rect != NULL)
    {
        _value = rect->GetRectType();
        CheckValueBounds(&_value);
        switch((int)_value)
        {
            case UISoundRect::WAVEFORM:
                sprintf(valueStr, "WAVEFORM");
                break;
                
            case UISoundRect::SPECTROGRAM:
                sprintf(valueStr, "SPECTROGRAM");
                break;
                
            case UISoundRect::CHROMAGRAM:
                sprintf(valueStr, "CHROMAGRAM");
                break;
                
            default:
                sprintf(valueStr, "");
                break;
        }

        SetValueStr(valueStr);
    }
}

ParameterColorRange::ParameterColorRange(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("Range");

    _minValue = 0.0;
    _maxValue = 2.0;
    _defaultValue = 1.0;
    _value = _defaultValue;
}

ParameterColorRange::~ParameterColorRange() {}

void
ParameterColorRange::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOUND_RECT)
        return;

    UISoundRect *rect = (UISoundRect *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (rect != NULL)
        rect->SetColorRange(_value);
}

void
ParameterColorRange::SetValue(double value)
{
    UISoundRect *rect = _app->GetCurrentSoundRect();
    SetValue(rect, value);
}

void
ParameterColorRange::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        UISoundRect *rect = _app->GetCurrentSoundRect();
        if (rect != NULL)
        { 
            _value = rect->GetColorRange();
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentRectParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

ParameterColorContrast::ParameterColorContrast(App *app,
                                               int paramId)
: Parameter(app, paramId)
{
    SetName("Contrast");

    _minValue = 0.01;
    _maxValue = 0.99;
    _defaultValue = 0.5;
    _value = _defaultValue;
}

ParameterColorContrast::~ParameterColorContrast() {}

void
ParameterColorContrast::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOUND_RECT)
        return;

    UISoundRect *rect = (UISoundRect *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (rect != NULL)
        rect->SetColorContrast(_value);
}

void
ParameterColorContrast::SetValue(double value)
{
    UISoundRect *rect = _app->GetCurrentSoundRect();
    SetValue(rect, value);
}

void
ParameterColorContrast::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        UISoundRect *rect = _app->GetCurrentSoundRect();
        if (rect != NULL)
        { 
            _value = rect->GetColorContrast();
            CheckValueBounds(&_value);
            sprintf(valueStr, "%g", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentRectParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

ParameterSoundRectOrientation::
ParameterSoundRectOrientation(App *app,
                              int paramId)
: Parameter(app, paramId)
{
    SetName("Orientation");

    _minValue = UISoundRect::HORIZONTAL;
    _maxValue = UISoundRect::VERTICAL;
    _defaultValue = UISoundRect::HORIZONTAL;
    _value = _defaultValue;
}

ParameterSoundRectOrientation::~ParameterSoundRectOrientation() {}

void
ParameterSoundRectOrientation::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOUND_RECT)
        return;

    UISoundRect *rect = (UISoundRect *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (rect != NULL)
        rect->SetOrientation((UISoundRect::Orientation)_value);
}

void
ParameterSoundRectOrientation::SetValue(double value)
{
    UISoundRect *rect = _app->GetCurrentSoundRect();
    SetValue(rect, value);
}

void
ParameterSoundRectOrientation::UpdateValue()
{
    char valueStr[255];
    UISoundRect *rect = _app->GetCurrentSoundRect();
    if (rect != NULL)
    {
        _value = rect->GetOrientation();
        CheckValueBounds(&_value);
        switch((int)_value)
        {
            case UISoundRect::HORIZONTAL:
                sprintf(valueStr, "HORIZONTAL");
                break;
                
            case UISoundRect::VERTICAL:
                sprintf(valueStr, "VERTICAL");
                break;
                
            default:
                sprintf(valueStr, "");
                break;
        }

        SetValueStr(valueStr);
    }
}

ParameterTimeStretch::ParameterTimeStretch(App *app,
                                           int paramId)
: Parameter(app, paramId)
{
    SetName("Time Stretch");

    _minValue = 0.0; //0.05;
    _maxValue = 2.0;
    _defaultValue = 1.0;
    _value = _defaultValue;

    _powFactor = 2.0;
}

ParameterTimeStretch::~ParameterTimeStretch() {}

void
ParameterTimeStretch::SetValue(ParameterProvider *provider, double value)
{
    if (provider == NULL)
        return;
    
    if (provider->GetParamProviderType() != ParameterProvider::SOURCE)
        return;

    Source *source = (Source *)provider;
    
    Parameter::SetValueCheckBounds(value);

    if (source != NULL)
    {
        if (source->GetSourceType() == Source::GRAIN_CLOUD)
            ((GrainCloud *)source)->SetTimeStretch(_value);
        else if (source->GetSourceType() == Source::SINGLE_GRAIN)
            ((SingleGrain *)source)->SetTimeStretch(_value);
        else if (source->GetSourceType() == Source::LOOP_SOURCE)
            ((LoopSource *)source)->SetTimeStretch(_value);
    }
}

void
ParameterTimeStretch::SetValue(double value)
{
    Source *source = _app->GetCurrentSource();
    SetValue(source, value);
}

void
ParameterTimeStretch::UpdateValue()
{
    char valueStr[255];
    const char *paramValueString = _app->GetParamValueString();
    if (paramValueString == NULL)
    {
        Source *source = _app->GetCurrentSource();
        if (source != NULL)
        {
            if (source->GetSourceType() == Source::GRAIN_CLOUD)
                _value = ((GrainCloud *)source)->GetTimeStretch();
            else if (source->GetSourceType() == Source::SINGLE_GRAIN)
                _value = ((SingleGrain *)source)->GetTimeStretch();
            else if (source->GetSourceType() == Source::LOOP_SOURCE)
                _value = ((LoopSource *)source)->GetTimeStretch(); 
            
            CheckValueBounds(&_value);
            sprintf(valueStr, "%.2f", _value);
            SetValueStr(valueStr);
        }
    }
    else
    {
        if (_app->GetCurrentSourceParamIdx() == _idx)
        {
            sprintf(valueStr, "%s", paramValueString);
            SetValueStr(valueStr);
        }
    }
}

ParameterStereoEffectPan::ParameterStereoEffectPan(App *app,
                                                   int paramId)
: Parameter(app, paramId)
{
    SetName("Pan");

    _minValue = -100.0;
    _maxValue = 100.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterStereoEffectPan::~ParameterStereoEffectPan() {}

void
ParameterStereoEffectPan::SetValue(ParameterProvider *provider, double value) {}

void
ParameterStereoEffectPan::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterStereoEffectPan::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f %%", _value);
    SetValueStr(valueStr);
}

ParameterStereoEffectWidth::ParameterStereoEffectWidth(App *app,
                                                       int paramId)
: Parameter(app, paramId)
{
    SetName("Width");

    _minValue = -100.0;
    _maxValue = 100.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterStereoEffectWidth::~ParameterStereoEffectWidth() {}

void
ParameterStereoEffectWidth::SetValue(ParameterProvider *provider, double value) {}

void
ParameterStereoEffectWidth::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterStereoEffectWidth::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f %%", _value);
    SetValueStr(valueStr);
}

ParameterReverbEffectOrder::ParameterReverbEffectOrder(App *app,
                                                       int paramId)
: Parameter(app, paramId)
{
    SetName("Order");

    _minValue = 1.0;
    _maxValue = 10.0;
    _defaultValue = 8.0;
    _value = _defaultValue;
}

ParameterReverbEffectOrder::~ParameterReverbEffectOrder() {}

void
ParameterReverbEffectOrder::SetValue(ParameterProvider *provider, double value) {}

void
ParameterReverbEffectOrder::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterReverbEffectOrder::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%d", (int)_value);
    SetValueStr(valueStr);
}

ParameterReverbEffectFeedDryIn::ParameterReverbEffectFeedDryIn(App *app,
                                                               int paramId)
: Parameter(app, paramId)
{
    SetName("Feed Dry In");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.2;
    _value = _defaultValue;
}

ParameterReverbEffectFeedDryIn::~ParameterReverbEffectFeedDryIn() {}

void
ParameterReverbEffectFeedDryIn::SetValue(ParameterProvider *provider, double value) {}

void
ParameterReverbEffectFeedDryIn::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterReverbEffectFeedDryIn::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.3f", _value);
    SetValueStr(valueStr);
}

ParameterReverbEffectFeedWetOut::ParameterReverbEffectFeedWetOut(App *app,
                                                                 int paramId)
: Parameter(app, paramId)
{
    SetName("Feed Wet Out");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.2;
    _value = _defaultValue;
}

ParameterReverbEffectFeedWetOut::~ParameterReverbEffectFeedWetOut() {}

void
ParameterReverbEffectFeedWetOut::SetValue(ParameterProvider *provider, double value) {}

void
ParameterReverbEffectFeedWetOut::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterReverbEffectFeedWetOut::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.3f", _value);
    SetValueStr(valueStr);
}

ParameterReverbEffectDryMix::ParameterReverbEffectDryMix(App *app,
                                                         int paramId)
: Parameter(app, paramId)
{
    SetName("Dry Mix");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.7;
    _value = _defaultValue;
}

ParameterReverbEffectDryMix::~ParameterReverbEffectDryMix() {}

void
ParameterReverbEffectDryMix::SetValue(ParameterProvider *provider, double value) {}

void
ParameterReverbEffectDryMix::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterReverbEffectDryMix::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.3f", _value);
    SetValueStr(valueStr);
}

ParameterReverbEffectDelayTime::ParameterReverbEffectDelayTime(App *app,
                                                               int paramId)
: Parameter(app, paramId)
{
    SetName("Delay Time");

    _minValue = 1.0;
    _maxValue = 10000.0;
    _defaultValue = 7000.0;
    _value = _defaultValue;
}

ParameterReverbEffectDelayTime::~ParameterReverbEffectDelayTime() {}

void
ParameterReverbEffectDelayTime::SetValue(ParameterProvider *provider, double value) {}

void
ParameterReverbEffectDelayTime::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterReverbEffectDelayTime::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%d", (int)_value);
    SetValueStr(valueStr);
}

ParameterReverbEffectDecay::ParameterReverbEffectDecay(App *app,
                                                       int paramId)
: Parameter(app, paramId)
{
    SetName("Decay");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.95;
    _value = _defaultValue;
}

ParameterReverbEffectDecay::~ParameterReverbEffectDecay() {}

void
ParameterReverbEffectDecay::SetValue(ParameterProvider *provider, double value) {}

void
ParameterReverbEffectDecay::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterReverbEffectDecay::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.3f", _value);
    SetValueStr(valueStr);
}

ParameterBitCrushEffectRatio::ParameterBitCrushEffectRatio(App *app,
                                                           int paramId)
: Parameter(app, paramId)
{
    SetName("Ratio");

    _minValue = 0.0;
    _maxValue = 100.0;
    _defaultValue = 0.0;
    _value = _defaultValue;

    _powFactor = 1.0/3.0;
}

ParameterBitCrushEffectRatio::~ParameterBitCrushEffectRatio() {}

void
ParameterBitCrushEffectRatio::SetValue(ParameterProvider *provider, double value) {}

void
ParameterBitCrushEffectRatio::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterBitCrushEffectRatio::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f %%", _value);
    SetValueStr(valueStr);
}

ParameterSaturationEffectRatio::ParameterSaturationEffectRatio(App *app,
                                                               int paramId)
: Parameter(app, paramId)
{
    SetName("Ratio");

    _minValue = 0.0;
    _maxValue = 100.0;
    _defaultValue = 0.0;
    _value = _defaultValue;

    _powFactor = 1.0/3.0;
}

ParameterSaturationEffectRatio::~ParameterSaturationEffectRatio() {}

void
ParameterSaturationEffectRatio::SetValue(ParameterProvider *provider, double value) {}

void
ParameterSaturationEffectRatio::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterSaturationEffectRatio::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f %%", _value);
    SetValueStr(valueStr);
}

ParameterDelayEffectTime::ParameterDelayEffectTime(App *app,
                                                   int paramId)
: Parameter(app, paramId)
{
    SetName("Time");

    _minValue = 1.0;
    _maxValue = 5000.0;
    _defaultValue = 1000.0;
    _value = _defaultValue;
}

ParameterDelayEffectTime::~ParameterDelayEffectTime() {}

void
ParameterDelayEffectTime::SetValue(ParameterProvider *provider, double value) {}

void
ParameterDelayEffectTime::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterDelayEffectTime::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%d ms", (int)_value);
    SetValueStr(valueStr);
}

ParameterDelayEffectFeedback::ParameterDelayEffectFeedback(App *app,
                                                           int paramId)
: Parameter(app, paramId)
{
    SetName("Feedback");

    _minValue = 0.0;
    _maxValue = 100.0;
    _defaultValue = 50.0;
    _value = _defaultValue;
}

ParameterDelayEffectFeedback::~ParameterDelayEffectFeedback() {}

void
ParameterDelayEffectFeedback::SetValue(ParameterProvider *provider, double value) {}

void
ParameterDelayEffectFeedback::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterDelayEffectFeedback::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f %%", _value);
    SetValueStr(valueStr);
}

ParameterDelayEffectWetMix::ParameterDelayEffectWetMix(App *app,
                                                       int paramId)
: Parameter(app, paramId)
{
    SetName("Wet");

    _minValue = 0.0;
    _maxValue = 100.0;
    _defaultValue = 50.0;
    _value = _defaultValue;
}

ParameterDelayEffectWetMix::~ParameterDelayEffectWetMix() {}

void
ParameterDelayEffectWetMix::SetValue(ParameterProvider *provider, double value) {}

void
ParameterDelayEffectWetMix::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterDelayEffectWetMix::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f %%", _value);
    SetValueStr(valueStr);
}

ParameterVolumeEffectVolume::ParameterVolumeEffectVolume(App *app,
                                                         int paramId)
: Parameter(app, paramId)
{
    SetName("Volume");

    _minValue = -12.0;
    _maxValue = 12.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterVolumeEffectVolume::~ParameterVolumeEffectVolume() {}

void
ParameterVolumeEffectVolume::SetValue(ParameterProvider *provider, double value) {}

void
ParameterVolumeEffectVolume::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterVolumeEffectVolume::UpdateValue()
{
    char valueStr[255];
    sprintf(valueStr, "%.1f dB", _value);
    SetValueStr(valueStr);
}

ParameterVolumeEffectMute::ParameterVolumeEffectMute(App *app,
                                                     int paramId)
: Parameter(app, paramId)
{
    SetName("Mute");
    
    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterVolumeEffectMute::~ParameterVolumeEffectMute() {}

void
ParameterVolumeEffectMute::SetValue(ParameterProvider *provider, double value) {}

void
ParameterVolumeEffectMute::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterVolumeEffectMute::UpdateValue()
{
    char valueStr[255];
    if (_value < 0.5)
        sprintf(valueStr, "");
    else
        sprintf(valueStr, "muted");
    
    SetValueStr(valueStr);
}

ParameterEffectOn::ParameterEffectOn(App *app, int paramId)
: Parameter(app, paramId)
{
    SetName("On");

    _minValue = 0.0;
    _maxValue = 1.0;
    _defaultValue = 0.0;
    _value = _defaultValue;
}

ParameterEffectOn::~ParameterEffectOn() {}

void
ParameterEffectOn::SetValue(ParameterProvider *provider, double value) {}

void
ParameterEffectOn::SetValue(double value)
{
    _value = value;

    UpdateValue();
}

void
ParameterEffectOn::UpdateValue()
{
    char valueStr[255];
    if (_value < 0.5)
        sprintf(valueStr, "off");
    else
        sprintf(valueStr, "on");
    
    SetValueStr(valueStr);
}
