/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

// Author: Micha Hanselmann
// https://github.com/deermichel/reverb
// Many thanks to him to have made possible to re-use and modify his original reverb code.

#include <math.h>

#include "Reverb.h"

Reverb::Reverb(int order, int maxNumSamples)
{
    _maxNumSamples = maxNumSamples;

    _drySignalMix = 1.0;
    _delayTime = -1;

    _order = order;
        
    Init();
}

Reverb::~Reverb() {}

void
Reverb::Process(vector<SAMPLE> *samples)
{
    _inputBuffer.Push(samples->data(), samples->size());
    _inputBuffer.Pop(samples->size());
    
    for (int i = 0; i < _delayBuffer.size(); i++)
        _delayBuffer[i].Pop(samples->size());
    
    for (int i = 0; i < samples->size(); i++)
        (*samples)[i] = ProcessMainLine(i);
}

void
Reverb::SetOrder(int order)
{
    if (order != _order)
    {
        _order = order;
        
        Init();
    }
}

void
Reverb::SetFeedDryIn(double feedDryIn)
{
    for (int i = 0; i < _feedDryIn.size(); i++)
        _feedDryIn[i] = feedDryIn;
}

void
Reverb::SetFeedWetOut(double feedWetOut)
{
    for (int i = 0; i < _feedWetOut.size(); i++)
        _feedWetOut[i] = feedWetOut;
}

void
Reverb::SetDrySignalMix(double drySignalMix)
{
    _drySignalMix = drySignalMix;
}

void
Reverb::SetDelayTapTime(int delayTime)
{
    if (delayTime == _delayTime)
        return;

    _delayTime = delayTime;
    
    double distribution = log(delayTime)/_order;
    
    for (int i = 0; i < _delayTaps.size(); i++)
        _delayTaps[i] = exp(distribution * (i + 1));
    
    _inputBuffer.SetCapacity(_maxNumSamples + delayTime);
    
    for (int i = 0; i < _delayBuffer.size(); i++)    
        _delayBuffer[i].SetCapacity(_maxNumSamples + delayTime);
}

void
Reverb::SetDecay(double decay)
{
    for (int i = 0; i < _decay.size(); i++)
        _decay[i] = decay;
}

SAMPLE
Reverb::ProcessMainLine(int z)
{
    SAMPLE result = _drySignalMix * _inputBuffer.Peek(z);
    
    for (int i = 0; i < _feedWetOut.size(); i++)
        result += _feedWetOut[i] * ProcessDelayLine(z, i);
    
    return result;
}

SAMPLE
Reverb::ProcessDelayLine(int z, int n)
{
    SAMPLE result = _feedDryIn[n] * _inputBuffer.Peek(z - _delayTaps[n]);
    
    for (int i = 0; i < _feedbackMatrix[n].size(); i++)
        result += _feedbackMatrix[n][i] *
            _delayBuffer[i].Peek(z - _delayTaps[n])*_decay[i];
    
    _delayBuffer[n].Poke(z, result);
    
    return result;
}

void
Reverb::Init()
{
    _inputBuffer.SetCapacity(_maxNumSamples);
    
    _delayBuffer.resize(_order);
    for (int i = 0; i < _order; i++)
        _delayBuffer[i].SetCapacity(_maxNumSamples);
    
    _feedDryIn.resize(_order);
    _feedWetOut.resize(_order);
    _delayTaps.resize(_order);
    _decay.resize(_order);

    _feedbackMatrix.resize(_order);
    for (int i = 0; i < _order; i++)
        _feedbackMatrix[i].resize(_order);
    
    GenerateHadamardMatrix(&_feedbackMatrix, 1.0/sqrt(_order));
}

void
Reverb::GenerateHadamardMatrix(vector<vector<SAMPLE> > *result, double value)
{
    (*result)[0][0] = value;
    
    for (int n = 2; n <= result->size(); n *= 2)
    {
        for (int i = 0; i < n / 2; i++)
        {
            for (int j = 0; j < n / 2; j++)
            {
                (*result)[i][j + n / 2] = (*result)[i][j];
                (*result)[i + n / 2][j] = (*result)[i][j];
                (*result)[i + n / 2][j + n / 2] = -(*result)[i][j];
            }
        }
    }
}
