/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Parameter.h"
#include "DSP.h"
#include "Reverb.h"
#include "ReverbEffect.h"

#define DEFAULT_ORDER 4
#define NUM_SAMPLES 2048
//#define NUM_SAMPLES 20000

ReverbEffect::ReverbEffect()
{
    for (int i = 0; i < 2; i++)
        _reverbs[i] = new Reverb(DEFAULT_ORDER, NUM_SAMPLES);

    _params[ON_PARAM] =
        new ParameterEffectOn(NULL, ON_PARAM);
    _params[ORDER_PARAM] =
        new ParameterReverbEffectOrder(NULL, ORDER_PARAM);
    _params[FEED_DRY_IN_PARAM] =
        new ParameterReverbEffectFeedDryIn(NULL, FEED_DRY_IN_PARAM);
    _params[FEED_WET_OUT_PARAM] =
        new ParameterReverbEffectFeedWetOut(NULL, FEED_WET_OUT_PARAM);
    _params[DRY_MIX_PARAM] =
        new ParameterReverbEffectDryMix(NULL, DRY_MIX_PARAM);
    _params[DELAY_TIME_PARAM] =
        new ParameterReverbEffectDelayTime(NULL, DELAY_TIME_PARAM);
    _params[DECAY_PARAM] =
        new ParameterReverbEffectDecay(NULL, DECAY_PARAM);
}

ReverbEffect::~ReverbEffect()
{
    for (int i = 0; i < 2; i++)
        delete _reverbs[i];

    for (int i = 0; i < NUM_PARAMS; i++)
        delete _params[i];
}

void
ReverbEffect::Process(vector<SAMPLE> *buffer, int nFrames, int nChan)
{
    UpdateParameters();

    double on = _params[ON_PARAM]->GetValue();
    if (on < 0.5)
        return;
    
    if (nChan == 1)
        _reverbs[0]->Process(buffer);
    else
        // nChan == 2
    {
        vector<SAMPLE> lBuf;
        lBuf.resize(nFrames);

        vector<SAMPLE> rBuf;
        rBuf.resize(nFrames);

        DSP::DeInterleave(*buffer, &lBuf, &rBuf);

        _reverbs[0]->Process(&lBuf);
        _reverbs[1]->Process(&rBuf);

        DSP::Interleave(buffer, lBuf, rBuf);
    }
}

Parameter *
ReverbEffect::GetParameter(ParamIdx paramIdx)
{
    return _params[paramIdx];
}

void
ReverbEffect::Serialize(Serializer *serializer)
{
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
ReverbEffect::UnSerialize(Serializer *serializer)
{
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);

    UpdateParameters();
}

void
ReverbEffect::UpdateParameters()
{
    double order = _params[ORDER_PARAM]->GetValue();
    double feedDryIn = _params[FEED_DRY_IN_PARAM]->GetValue();
    double feedWetOut = _params[FEED_WET_OUT_PARAM]->GetValue();
    double dryMix = _params[DRY_MIX_PARAM]->GetValue();
    double delayTime = _params[DELAY_TIME_PARAM]->GetValue();
    double decay = _params[DECAY_PARAM]->GetValue();

    for (int i = 0; i < 2; i++)
    {
        _reverbs[i]->SetOrder((int)order);
        _reverbs[i]->SetFeedDryIn(feedDryIn);
        _reverbs[i]->SetFeedWetOut(feedWetOut);
        _reverbs[i]->SetDrySignalMix(dryMix);
        _reverbs[i]->SetDelayTapTime(delayTime);
        _reverbs[i]->SetDecay(decay);
    }
}
