/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Spectrogram.h"
#include "SpectroOverlapAddProcessor.h"
#include "ChromaOverlapAddProcessor.h"
#include "SamplesToSpectro.h"

#define STS_BLOCK_SIZE 512

Spectrogram *
SamplesToSpectro::SamplesToSpectrogram(const vector<SAMPLE> &samples,
                                       int nFrames, int nChan, int chanNum)
{
    // TODO: manage the end of the data,
    // by padding zeros to avoid skipping the last samples
    int numCols = SPECTRO_OVERLAP*nFrames/SPECTRO_FFT_SIZE;
    // Reduce, to avoid 4 cols of silence at the end
    numCols -= SPECTRO_OVERLAP;

    Spectrogram *spectro = new Spectrogram(SPECTRO_FFT_SIZE, numCols);
    
    OverlapAdd overAdd(SPECTRO_FFT_SIZE, SPECTRO_OVERLAP, true, false);
    SpectroOverlapAddProcessor processor(spectro, chanNum);
    overAdd.AddProcessor(&processor);
        
    SamplesToSpectrogram(samples, nFrames, nChan, &overAdd);

    return spectro;
}

Spectrogram *
SamplesToSpectro::SamplesToChromagram(const vector<SAMPLE> &samples,
                                      int nFrames, int nChan, int chanNum)
{
    // TODO: manage the end of the data,
    // by padding zeros to avoid skipping the last samples
    int numCols = SPECTRO_OVERLAP*nFrames/SPECTRO_FFT_SIZE;
    // Reduce, to avoid 4 cols of silence at the end
    numCols -= SPECTRO_OVERLAP;
    
    Spectrogram *spectro = new Spectrogram(SPECTRO_FFT_SIZE, numCols);

    OverlapAdd overAdd(SPECTRO_FFT_SIZE, SPECTRO_OVERLAP, true, false);
    ChromaOverlapAddProcessor processor(SPECTRO_FFT_SIZE, 44100.0, spectro, chanNum);
    overAdd.AddProcessor(&processor);
    
    SamplesToSpectrogram(samples, nFrames, nChan, &overAdd);

    return spectro;
}

void
SamplesToSpectro::SamplesToSpectrogram(const vector<SAMPLE> &samples,
                                       int nFrames, int nChan, OverlapAdd *overAdd)
{    
    vector<SAMPLE> buf;
    buf.resize(STS_BLOCK_SIZE*nChan);
    
    int frameNum = 0;
    while(frameNum < nFrames - STS_BLOCK_SIZE)
    {
        // Fill the block with deinterleaved data
        if (nChan == 1)
            memcpy(buf.data(), samples.data(), STS_BLOCK_SIZE*sizeof(SAMPLE));
        else
        {
            // de-interleave
            for (int i = frameNum; i < frameNum + STS_BLOCK_SIZE; i++)
            {
                for (int k = 0; k < nChan; k++)
                    buf[i - frameNum + k*STS_BLOCK_SIZE] = samples[i*nChan + k];
            }
        }
        
        overAdd->Feed(buf, STS_BLOCK_SIZE, nChan);
        
        frameNum += STS_BLOCK_SIZE;
    }
}
