/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "UIStyle.h"
#include "UISingleGrain.h"
#include "AudioSource.h"
#include "Serializer.h"
#include "Parameter.h"
#include "App.h"
#include "SingleGrain.h"

SingleGrain::SingleGrain(App *app,
                         const vector<SamplesProvider *> &sampProviders)
{
    _sourceType = Source::SINGLE_GRAIN;
    
    _app = app;

    _uiSingleGrain = NULL;
    
    srand(time(NULL));
    
    _playPositions.resize(sampProviders.size());
    _playVols.resize(sampProviders.size());
    
    _duration = 500.0;

    _pitch = 1.0f;
    
    _envelopeType = Envelope::HANNING;
    
    _pitchLFOFreq = 0.01f;
    _pitchLFOAmount = 0.0f;
    
    _channelMults.resize(CHANNELS);
    for (int i = 0; i < CHANNELS; i++)
        _channelMults[i] = 0.999f;
    
    _currentAroundChan = 1;
    _stereoSide = 0; 
    _side = 1;
    
    _spatialMode = UNITY;
    _channelLocation = -1;
    
    _dirMode = FORWARD;
    
    _source = new AudioSource(sampProviders, _envelope,
                              _duration, _pitch, _timeStretch);

    SetVolumeDb(0.0);

    _range = 10.0;
    
    _localTime = 0;
    _bangTime = _duration * SRATE * (double)0.001;
    
    _source->SetDurationMs(_duration);

    InitParameters();
    _currentParamIdx = NO_PARAMETER;

    double paramsSensivity[NUM_PARAMS] =
    {
        3.0, // DURATION
        6.0, // WINDOW
        6.0, // DIRECTION
        1.0, // PITCH
        4.0, // P_LFO_FREQ
        3.0, // P_LFO_AMT
        6.0, // SPATIALIZE
        6.0, // VOLUME
        4.0, // RANGE
        6.0  // TIME_STRETCH
    };
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->SetSensivity(paramsSensivity[i]);
}

SingleGrain::~SingleGrain()
{
    if (_source != NULL)
        delete _source;

    if (_isClone)
        return;
    
    if (_uiSingleGrain)
        delete _uiSingleGrain;

    for (int i = 0; i < NUM_PARAMS; i++)
        delete _params[i];
}

Source *
SingleGrain::Clone() const
{
    const vector<SamplesProvider*> &sampProviders = _source->GetSampleProviders();
    SingleGrain *clone = new SingleGrain(_app, sampProviders);
    Source::Clone(clone);

    clone->_app = _app;
    clone->_uiSingleGrain = _uiSingleGrain;

    clone->_localTime = _localTime;
    clone->_bangTime = _bangTime;
    
    clone->_currentAroundChan = _currentAroundChan;
    clone->_stereoSide = _stereoSide; 
    clone->_side = _side;
    
    clone->_channelMults = _channelMults;
    clone->_spatialMode = _spatialMode;
    clone->_channelLocation = _channelLocation;

    clone->_volumeDb = _volumeDb;
    clone->_normedVol = _normedVol;

    clone->_range = _range;

    // Prevent memory leak...
    delete clone->_source;
    
    clone->_source = _source->Clone();

    clone->_pitch = _pitch;
    clone->_duration = _duration;
    clone->_pitchLFOFreq = _pitchLFOFreq;
    clone->_pitchLFOAmount = _pitchLFOAmount;
    
    clone->_dirMode = _dirMode;
    clone->_envelopeType = _envelopeType;

    for (int i = 0; i < NUM_PARAMS; i++)
        clone->_params[i] = _params[i];
    clone->_currentParamIdx = _currentParamIdx;

    return clone;
}
    
void
SingleGrain::RegisterUI(UISource *uiSingleGrain)
{
    _uiSingleGrain = (UISingleGrain *)uiSingleGrain;
    _uiSingleGrain->SetDuration(_duration);
}

UISingleGrain *
SingleGrain::GetUI()
{
    return _uiSingleGrain;
}

void
SingleGrain::SetEnvelopeType(int envType)
{
    int numEnvs = _envelope->GetNumEnvelopes();
    _envelopeType = envType % numEnvs;
    
    if (_envelopeType < 0)
        _envelopeType = _envelope->GetNumEnvelopes() - 1;

    _source->SetEnvelopeType(_envelopeType);
}

int
SingleGrain::GetEnvelopeType()
{
    return _envelopeType;
}

void
SingleGrain::SetDurationMs(double dur)
{
    _duration = dur;
    
    _source->SetDurationMs(_duration);
    
    UpdateBangTime();
    
    if (_uiSingleGrain)
        _uiSingleGrain->SetDuration(_duration);
}

void
SingleGrain::Serialize(Serializer *serializer)
{
    Source::Serialize(serializer);
    
    serializer->PutLong(_localTime);
    serializer->PutDouble(_bangTime);
    serializer->PutInt(_currentAroundChan);
    serializer->PutInt(_stereoSide);
    serializer->PutInt(_side);
    
    serializer->PutInt(_spatialMode);
    serializer->PutInt(_channelLocation);
    serializer->PutDouble(_volumeDb);
    serializer->PutDouble(_normedVol);

    serializer->PutDouble(_range);
    
    serializer->PutDouble(_pitch);
    serializer->PutDouble(_duration);
    serializer->PutDouble(_pitchLFOFreq);
    serializer->PutDouble(_pitchLFOAmount);

    serializer->PutInt(_dirMode);
    serializer->PutInt(_envelopeType);

    SerializeParameters(serializer);
    serializer->PutInt(_currentParamIdx);
}

void
SingleGrain::UnSerialize(Serializer *serializer)
{
    Source::UnSerialize(serializer);
    
    serializer->GetLong((long *)&_localTime);
    serializer->GetDouble(&_bangTime);
    serializer->GetInt(&_currentAroundChan);
    serializer->GetInt(&_stereoSide);
    serializer->GetInt(&_side);
    
    serializer->GetInt(&_spatialMode);
    serializer->GetInt(&_channelLocation);
    serializer->GetDouble(&_volumeDb);
    serializer->GetDouble(&_normedVol);

    serializer->GetDouble(&_range);
    
    serializer->GetDouble(&_pitch);
    serializer->GetDouble(&_duration);
    serializer->GetDouble(&_pitchLFOFreq);
    serializer->GetDouble(&_pitchLFOAmount);

    serializer->GetInt(&_dirMode);
    serializer->GetInt(&_envelopeType);

    UnSerializeParameters(serializer);
    serializer->GetInt((int *)&_currentParamIdx);
}

Parameter *
SingleGrain::GetParameter(int paramIdx)
{
    return GetParameter((ParamIndex)paramIdx);
}

int
SingleGrain::GetNumParams() const
{
    return (int)NUM_PARAMS;
}

double
SingleGrain::GetUIX() const
{
    return _uiSingleGrain->GetX();
}

double
SingleGrain::GetUIY() const
{
    return _uiSingleGrain->GetY();
}

double
SingleGrain::GetUIRadius() const
{
    return _uiSingleGrain->GetRadius();
}

void
SingleGrain::UpdateBangTime()
{
    _bangTime = _duration * SRATE * (double)0.001;
}

void
SingleGrain::SetPitch(double targetPitch)
{
    if (targetPitch < 0.0001)
        targetPitch = 0.0001;
    
    _pitch = targetPitch;

    double notePitch = GetMidiNotePitch();
    double wheelPitch = GetMidiPitchWheelValue();
    _source->SetPitch(targetPitch*notePitch*wheelPitch);
}

double
SingleGrain::GetPitch()
{
    return _pitch;
}

void
SingleGrain::SetVolumeDb(double volDb)
{
    if (volDb > 6.0)
        volDb = 6.0;
    
    if (volDb < -60.0)
        volDb = -60.0;
    
    _volumeDb = volDb;
    
    _normedVol = pow( 10.0 , volDb * 0.05 );

    _source->SetVolume(_normedVol);
}

double
SingleGrain::GetVolumeDb()
{
    return _volumeDb;
}

void
SingleGrain::SetRange(double range)
{
    _range = range;

    _source->SetRange(range);
}

double
SingleGrain::GetRange()
{
    return _range;
}

void
SingleGrain::SetDirection(int dirMode)
{
    _dirMode = dirMode % 3;
    if (_dirMode < 0)
        _dirMode = 2;
    
    switch (_dirMode)
    {
        case FORWARD:
            _source->SetDirection(1.0);
            break;
            
        case BACKWARD:
            _source->SetDirection(-1.0);
            break;
            
        case RANDOM_DIR:
            if (randf() > 0.5)
                _source->SetDirection(1.0);
            else
                _source->SetDirection(-1.0);
            break;
            
        default:
            break;
    }
}

int
SingleGrain::GetDirection()
{
    return _dirMode;
}

double
SingleGrain::GetDurationMs()
{
    return _duration;
}

void
SingleGrain::UpdateSource()
{
    _source->SetDurationMs(_duration);

    double notePitch = GetMidiNotePitch();
    double wheelPitch = GetMidiPitchWheelValue();
    _source->SetPitch(_pitch*notePitch*wheelPitch);
    
    _source->SetEnvelopeType(_envelopeType);
    _source->SetTimeStretch(_timeStretch);
    
    switch (_dirMode)
    {
        case FORWARD:
            _source->SetDirection(1.0);
            break;
                
        case BACKWARD:
            _source->SetDirection(-1.0);
            break;
                
        case RANDOM_DIR:
            if (randf()>0.5)
                _source->SetDirection(1.0);
            else
                _source->SetDirection(-1.0);
            break;
                    
        default:
            break;
    }
    
    _source->SetVolume(_normedVol);

    _source->SetRange(_range);
}

void
SingleGrain::NextBuffer(vector<SAMPLE> *accumBuff, unsigned int numFrames)
{
    if (_isMuted)
        return;
        
    for (int i = 0; i < _playPositions.size(); i++)
        _playPositions[i] = -1.0;
    
    for (int i = 0; i < _playVols.size(); i++)
        _playVols[i] = 0.0;
        
    unsigned int nextFrame = 0;
    int frameSkip = numFrames/2;
    for (int j = 0; j < numFrames/frameSkip; j++)
    {
        // Check for bang
        if ((_localTime > _bangTime))
        { 
            _localTime = 0;
                
            for (int k = 0; k < _playPositions.size(); k++)
            {
                _playPositions[k] = (double)-1.0;
                _playVols[k] = (double)0.0;
            }
                
            if (_uiSingleGrain)
                _uiSingleGrain->GetTriggerPos(_playPositions.data(),
                                              _playVols.data(),
                                              _duration);

            // NOTE: this will update the random direction,
            // each time we start playing a grain
            UpdateSource();
        
            if ((_pitchLFOAmount > 0.0f) && (_pitchLFOFreq > 0.0f))
            {
                double nextPitch =
                    fabs(_pitch + _pitchLFOAmount *
                         sin(2*PI*_pitchLFOFreq*Time::Instance().
                             GetSeconds()));
                
                double notePitch = GetMidiNotePitch();
                double wheelPitch = GetMidiPitchWheelValue();
                _source->SetPitch(nextPitch*notePitch*wheelPitch);
            }
            
            UpdateSpatialization();
            
            _source->SetChannelMultipliers(_channelMults);
            _source->Play(_playPositions.data(), _playVols.data());
        }
        
        // Advance time
        _localTime += frameSkip;
        
        // Sample offset (1 sample at a time for now)
        nextFrame = j*frameSkip;
        
        double notePitch = GetMidiNotePitch();
        double wheelPitch = GetMidiPitchWheelValue();
        _source->SetPitch(_pitch*notePitch*wheelPitch);
        _source->NextBuffer(accumBuff, frameSkip, nextFrame);
    }
}

void
SingleGrain::SetPitchLFOFreq(double pfreq)
{
    _pitchLFOFreq = fabs(pfreq);
}

void
SingleGrain::SetPitchLFOAmount(double lfoamt)
{
    if (lfoamt < 0.0)
        lfoamt = 0.0f;
    
    _pitchLFOAmount = lfoamt;
}

double
SingleGrain::GetPitchLFOFreq()
{
    return _pitchLFOFreq;
}

double
SingleGrain::GetPitchLFOAmount()
{
    return _pitchLFOAmount;
}

void
SingleGrain::SetSpatialMode(int mode,
                           int channelNumber = -1)
{
    _spatialMode = mode % 3;
    if (_spatialMode < 0)
        _spatialMode = 2;
    
    // For positioning in a single audio channel. - not used currently
    // eventually swap out for azimuth instead of single channel
    if (channelNumber >= 0)
        _channelLocation = channelNumber;
}

int
SingleGrain::GetSpatialMode()
{
    return _spatialMode;
}

int
SingleGrain::GetSpatialChannel()
{
    return _channelLocation;
}

void
SingleGrain::UpdateSpatialization()
{
    // Currently assumes orientation L: 0,2,4,...  R: 1,3,5, etc (interleaved) 
    switch (_spatialMode)
    {
        case UNITY:
            for (int i = 0; i < CHANNELS; i++)
                _channelMults[i] = 0.999f;
            break;
            
        case STEREO:
            if (_stereoSide == 0)
                // Left
            {
                for (int i = 0; i < CHANNELS; i++)
                {
                    _channelMults[i]= 0.0f;
                    if ((i % 2) == 0)
                        _channelMults[i] = 0.999f;
                    else
                        _channelMults[i] = 0.0f;
                }
                _stereoSide = 1;
            }
            else
                // Right
            {
                for (int i = 0; i < CHANNELS; i++)
                {
                    _channelMults[i]= 0.0f;
                    if ((i % 2) == 0)
                        _channelMults[i] = 0.0f;
                    else
                        _channelMults[i] = 0.999f;
                }
                _stereoSide = 0;
            }
            break;
            
        case AROUND:
            for (int i = 0; i < CHANNELS; i++)
                _channelMults[i] = 0;
           
            // 1 3 5 7 6 4 2 0
            
            _channelMults[_currentAroundChan] = 0.999;
            _currentAroundChan += _side*2;
            if ((_currentAroundChan > CHANNELS) ||
                (_currentAroundChan < 0))
            {
                _side = -1*_side;
                _currentAroundChan += _side*3;
            }
            break;
            
        default:
            break;
    }    
}

void
SingleGrain::SetCurrentParamIdx(SingleGrain::ParamIndex idx)
{
    _currentParamIdx = idx;
}

void
SingleGrain::ResetCurrentParameterIndex()
{
    _currentParamIdx = NO_PARAMETER;
}

Parameter *
SingleGrain::GetParameter(ParamIndex paramIdx)
{
    return _params[paramIdx];
}

Parameter *
SingleGrain::GetCurrentParameter()
{
    if (_currentParamIdx == NO_PARAMETER)
        return NULL;

    return _params[_currentParamIdx];
}

void
SingleGrain::InitParameters()
{
    _params[DURATION] = new ParameterDuration(_app, DURATION);
    _params[ENVELOPE] = new ParameterEnvelope(_app, ENVELOPE);
    _params[DIRECTION] = new ParameterDirection(_app, DIRECTION);
    _params[PITCH] = new ParameterPitch(_app, PITCH);
    _params[P_LFO_FREQ] = new ParameterLFOFreq(_app, P_LFO_FREQ);
    _params[P_LFO_AMP] = new ParameterLFOAmp(_app, P_LFO_AMP);
    _params[SPATIALIZE] = new ParameterSpatialize(_app, SPATIALIZE);
    _params[VOLUME] = new ParameterVolume(_app, VOLUME);
    _params[RANGE] = new ParameterRange(_app, RANGE);
    _params[TIME_STRETCH] = new ParameterTimeStretch(_app, TIME_STRETCH);
}

void
SingleGrain::SerializeParameters(Serializer *serializer)
{
    int size = NUM_PARAMS;
    serializer->PutInt(size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
SingleGrain::UnSerializeParameters(Serializer *serializer)
{
    int size;
    serializer->GetInt(&size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);
}
