/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "SpectroToSamples.h"
#include "SpectroToSamplesOverlapAddProcessor.h"
#include "ChromaToSamplesOverlapAddProcessor.h"
#include "Spectrogram.h"
#include "Serializer.h"
#include "DSP.h"
#include "Scale.h"
#include "UISoundRect.h"
#include "SoundRect.h"

unsigned int SoundRect::_currentId = 0;

SoundRect::SoundRect()
{
    _id = _currentId++;
    
    _uiRect = NULL;

    _nChans = 0;
    _nFrames = 0;
    
    _spectro[0] = NULL;
    _spectro[1] = NULL;
    
    _chroma[0] = NULL;
    _chroma[1] = NULL;

    for (int i = 0; i < 2; i++)
        for (int j = 0; j < 2; j++)
            _spectroToSamples[i][j] = NULL;
    
    _paramY = 1.0;
    _range = 10.0;

    _isClone = false;
}

SoundRect::~SoundRect()
{ 
    for (int i = 0; i < 2; i++)
        for (int j = 0; j < 2; j++)
        {
            if (_spectroToSamples[i][j] != NULL)
                delete _spectroToSamples[i][j];
        }
}

void
SoundRect::RegisterUI(UISoundRect *uiRect)
{
    _uiRect = uiRect;
}

unsigned int
SoundRect::GetId() const
{
    return _id;
}

UISoundRect *
SoundRect::GetUI()
{
    return _uiRect;
}

void
SoundRect::AssociateSamples(const vector<SAMPLE> &samples,
                            unsigned long nFrames,
                            unsigned int nChans)
{
    // Make a copy for displaying
    _samples.resize(nFrames*nChans);
    memcpy(_samples.data(), samples.data(), nFrames*nChans*sizeof(SAMPLE));
    
    _nFrames = nFrames;
    _nChans = nChans;
}

vector<SAMPLE> *
SoundRect::GetSamples(int *nFrames, int *nChans)
{
    *nFrames = _nFrames;
    *nChans = _nChans;

    return &_samples;
}

void
SoundRect::AssociateSpectrogram(int chanNum, Spectrogram *spectro)
{
    _spectro[chanNum] = spectro;
    _spectro[chanNum]->PreProcessPhases();

    if (_spectroToSamples[SPECTROGRAM][chanNum] != NULL)
        delete _spectroToSamples[SPECTROGRAM][chanNum];

    SpectroToSamplesOverlapAddProcessor *processor =
        new SpectroToSamplesOverlapAddProcessor(_spectro[chanNum]);
    _spectroToSamples[SPECTROGRAM][chanNum] = new SpectroToSamples(processor); 
}

Spectrogram *
SoundRect::GetSpectrogram(int chanNum)
{
    return _spectro[chanNum];
}

void
SoundRect::AssociateChromagram(int chanNum, Spectrogram *chroma)
{
    _chroma[chanNum] = chroma;
    _chroma[chanNum]->PreProcessPhases();
 
    if (_spectroToSamples[CHROMAGRAM][chanNum] != NULL)
        delete _spectroToSamples[CHROMAGRAM][chanNum];

    ChromaToSamplesOverlapAddProcessor *processor =
        new ChromaToSamplesOverlapAddProcessor(SPECTRO_FFT_SIZE, SRATE, _chroma[chanNum]);
    _spectroToSamples[CHROMAGRAM][chanNum] = new SpectroToSamples(processor); 
}

Spectrogram *
SoundRect::GetChromagram(int chanNum)
{
    return _chroma[chanNum];
}

SamplesProvider *
SoundRect::Clone() const
{
    SoundRect *clone = new SoundRect();
    clone->_isClone = true;
    
    clone->_uiRect = _uiRect;

    //_id;
    
    clone->_samples = _samples;
    clone->_nFrames = _nFrames;
    clone->_nChans = _nChans;

    for (int i = 0; i < 2; i++)
        clone->_spectro[i] = _spectro[i];

    for (int i = 0; i < 2; i++)
        clone->_chroma[i] = _chroma[i];

    for (int i = 0; i < 2; i++)
        for (int j = 0; j < 2; j++)
            clone->_spectroToSamples[i][j] = _spectroToSamples[i][j]->Clone();
    
    clone->_paramY = _paramY;
    clone->_range = _range;

    return clone;
}

void
SoundRect::SetParamY(double paramY)
{
    _paramY = paramY;
}

void
SoundRect::SetRange(double range)
{
    _range = range;
}
    
int
SoundRect::GetNumSamples() const
{
    return _nFrames*_nChans;
}

int
SoundRect::GetNumChannels() const
{
    return _nChans;
}

int
SoundRect::GetNumFrames() const
{
    return _nFrames;
}

int
SoundRect::GetBlockMinSamples() const
{
    switch(_uiRect->GetRectType())
    {
        case UISoundRect::WAVEFORM:
            return 0;
            break;

        case UISoundRect::SPECTROGRAM:
        case UISoundRect::CHROMAGRAM:
            return SPECTRO_FFT_SIZE;
            break;

        default:
            break;
    }

    return 0;
}

bool
SoundRect::InternalPitchShift() const
{
    return (_uiRect->GetRectType() != UISoundRect::WAVEFORM);
}

void
SoundRect::GetSamples(vector<SAMPLE> *samples,
                      double start, double end,
                      double timeStretch, double direction, double pitchShift) const
{
    if (start < 0)
        return;
    if (end - start > _nFrames - 1)
        end = start + _nFrames - 1;
    if (end - start <= 0)
        return;

    samples->resize(((int)(end - start))*_nChans);
    DSP::Silent(samples);
    
    if (_uiRect->GetRectType() == UISoundRect::WAVEFORM)
    {
        if (fabs(timeStretch - 1.0) < 1e-15)
            // No time stretch
        {
            int numToCopy = ((int)(end - start))*_nChans;
            if (((int)start)*_nChans + numToCopy > _nFrames*_nChans)
                numToCopy = _nFrames*_nChans - ((int)start)*_nChans;
            if (numToCopy < 0)
                numToCopy = 0;
            memcpy(samples->data(), &_samples.data()[((int)start)*_nChans],
                   numToCopy*sizeof(SAMPLE));
            
            // paramY is volume
            for (int i = 0; i < samples->size(); i++)
                (*samples)[i] *= _paramY;

            if (direction < 0.0)
                DSP::Reverse(samples, CHANNELS);
        }
        else
        {
            // Use spectrogram for doing timeStretch
            double selSamples[2] = { start, end };
            double selFreq[2] = { 0.0, 1.0 };
            
            vector<SAMPLE> samps[2];
            for (int k = 0; k < 2; k++)
            {
                if (_spectroToSamples[SPECTROGRAM][k] != NULL)
                    _spectroToSamples[SPECTROGRAM][k]->
                        Process(&samps[k], selSamples, selFreq,
                                timeStretch, direction, pitchShift);
            }

            if (_nChans == 1)
                *samples = samps[0];
            else
                DSP::Interleave(samples, samps[0], samps[1]);
        }
    }
    else if (_uiRect->GetRectType() == UISoundRect::SPECTROGRAM)
    {
        double selSamples[2] = { start, end };
        double selFreq[2] = { _paramY - _range/100.0f, _paramY + _range/100.0f };
        
        for (int k = 0; k < 2; k++)
        {
            if (selFreq[k] < 0.0)
                selFreq[k] = 0.0;
            if (selFreq[k] > 1.0)
                selFreq[k] = 1.0;
        }

        // Convert to normalized Mel
        double hzPerBin = Scale::HzPerBin(SPECTRO_FFT_SIZE, SRATE);
        double minMel = Scale::HzToMel(1.0*hzPerBin);
        double maxMel = Scale::HzToMel(0.5*SPECTRO_FFT_SIZE*hzPerBin);
        for (int k = 0; k < 2; k++)
        {
            double mel = minMel + selFreq[k]*(maxMel - minMel);
            double freq = Scale::MelToHz(mel);
            double t = freq/(0.5*SRATE);
            
            selFreq[k] = t;
        }
        
        vector<SAMPLE> samps[2];
        for (int k = 0; k < 2; k++)
        {
            if (_spectroToSamples[SPECTROGRAM][k] != NULL)
                _spectroToSamples[SPECTROGRAM][k]->
                    Process(&samps[k], selSamples, selFreq,
                            timeStretch, direction, pitchShift);
        }

        if (_nChans == 1)
            *samples = samps[0];
        else
            DSP::Interleave(samples, samps[0], samps[1]);
    }
    else if (_uiRect->GetRectType() == UISoundRect::CHROMAGRAM)
    {
        double selSamples[2] = { start, end };
        double selChroma[2] = { _paramY - _range/100.0f, _paramY + _range/100.0f };
        
        for (int k = 0; k < 2; k++)
        {
            if (selChroma[k] < 0.0)
                selChroma[k] = 0.0;
            if (selChroma[k] > 1.0)
                selChroma[k] = 1.0;
        }
                                      
        vector<SAMPLE> samps[2];
        for (int k = 0; k < 2; k++)
        {
            if (_spectroToSamples[CHROMAGRAM][k] != NULL)
                _spectroToSamples[CHROMAGRAM][k]->
                    Process(&samps[k], selSamples, selChroma,
                            timeStretch, direction, pitchShift);
        }

        if (_nChans == 1)
            *samples = samps[0];
        else
            DSP::Interleave(samples, samps[0], samps[1]);
    }
}

void
SoundRect::Serialize(Serializer *serializer)
{
    serializer->PutInt(_id);

    serializer->PutLong(_nFrames);
    serializer->PutInt(_nChans);
    
    _uiRect->Serialize(serializer);
}

void
SoundRect::UnSerialize(Serializer *serializer)
{
    serializer->GetInt((int *)&_id);
    
    serializer->GetLong((long *)&_nFrames);
    serializer->GetInt((int *)&_nChans);
    
    _uiRect->UnSerialize(serializer);
}
