/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>

#include "Serializer.h"
#include "EffectsBox.h"
#include "Envelope.h"
#include "Source.h"

unsigned int Source::_currentId = 0;

Source::Source()
{
    _paramProviderType = ParameterProvider::SOURCE;
    _sourceType = UNDEFINED_TYPE;

    _midiNote = -1;
    _midiPitchWheelValue = 0.0;

    _timeStretch = 1.0;
    
    _id = _currentId++;

    _isMuted = false;
    
    _effectsBox = new EffectsBox();

    _envelope = new Envelope();
    
    _isClone = false;
}

Source::~Source()
{
    delete _effectsBox;
    delete _envelope;
}

Source::SourceType
Source::GetSourceType() const
{
    return _sourceType;
}

unsigned int
Source::GetId() const
{
    return _id;
}

void
Source::SetMidiNote(int midiNote)
{
    _midiNote = midiNote;
}

int
Source::GetMidiNote() const
{
    return _midiNote;
}

void
Source::SetMidiPitchWheelValue(double pitch)
{
    _midiPitchWheelValue = pitch;
}

double
Source::GetMidiPitchWheelValue() const
{
    double pitch = (_midiPitchWheelValue > 0.0) ? 1.0 + _midiPitchWheelValue :
        1.0 + _midiPitchWheelValue*0.5;

    return pitch;
}

double
Source::GetTimeStretch() const
{
    return _timeStretch;
}

void
Source::SetTimeStretch(double stretch)
{
    _timeStretch = stretch;
}

EffectsBox *
Source::GetEffectsBox()
{
    return _effectsBox;
}

void
Source::SetMuted(bool flag)
{
    _isMuted = flag;
}

bool
Source::IsMuted() const
{
    return _isMuted;
}

Envelope *
Source::GetEnvelope()
{
    return _envelope;
}

void
Source::Serialize(Serializer *serializer)
{
    serializer->PutBool(_isMuted);
    serializer->PutDouble(_timeStretch);
    _effectsBox->Serialize(serializer);
    _envelope->Serialize(serializer);
}

void
Source::UnSerialize(Serializer *serializer)
{
    serializer->GetBool(&_isMuted);
    serializer->GetDouble(&_timeStretch);
    _effectsBox->UnSerialize(serializer);
    _envelope->UnSerialize(serializer);
}

void
Source::Clone(Source *clone) const
{
    clone->_isClone = true;
    clone->_isMuted = _isMuted;
    clone->_sourceType = _sourceType;
    clone->_id = _id;
    clone->_midiNote = _midiNote;
    //clone->_effectsBox = _effectsBox;
    clone->_timeStretch = _timeStretch;
}

double
Source::GetMidiNotePitch() const
{
    if (_midiNote == -1)
        return 1.0;
    
    // C5 is 72
    // Take it as reference
    
    const double halfToneCoeff = pow(2.0, 1.0/12.0);

    double pitch = pow(halfToneCoeff, _midiNote - 72.0);

    return pitch;
}
