/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Stk.h"

#include "Spectrogram.h"
#include "SpectroToSamplesOverlapAddProcessor.h"
#include "ChromaToSamplesOverlapAddProcessor.h"
#include "PitchShiftOverlapAddProcessor.h"
#include "SpectroToSamples.h"

#define STS_BLOCK_SIZE 512

SpectroToSamples::SpectroToSamples(OverlapAddProcessor *processor)
{
    _processor = processor;
    _overlapAdd = new OverlapAdd(SPECTRO_FFT_SIZE, SPECTRO_OVERLAP, false, true);
    _overlapAdd->AddProcessor(_processor);

    _pitchProcessor = new PitchShiftOverlapAddProcessor(SPECTRO_FFT_SIZE, SRATE);
    _overlapAdd->AddProcessor(_pitchProcessor);
}

SpectroToSamples::~SpectroToSamples()
{
    delete _overlapAdd;
    delete _processor;
    delete _pitchProcessor;
}

SpectroToSamples *
SpectroToSamples::Clone() const
{
    SpectroToSamples *clone = new SpectroToSamples(_processor->Clone());
    return clone;
}

void
SpectroToSamples::Process(vector<SAMPLE> *samples, double selSamples[2],
                          double selFreq[2],
                          double timeStretch, double direction, double pitchShift)
{
    _processor->SetSelFreq(selFreq);
    _processor->SetTimeStretch(timeStretch*direction);

    _overlapAdd->SetSynthTimeStretch(pitchShift);
    _pitchProcessor->SetPitchShift(pitchShift);
        
    // Target size
    samples->resize(selSamples[1] - selSamples[0]);
    
    // Spectrogram column number
    double pos = (direction > 0.0) ? selSamples[0] : selSamples[1] - 1; 
    double colNum = (pos/(SPECTRO_FFT_SIZE/SPECTRO_OVERLAP));
    _processor->SetSpectroColNum(colNum);
    
    // Buffer for feeding
    vector<SAMPLE> blockBuf;
    blockBuf.resize(STS_BLOCK_SIZE);
    memset(blockBuf.data(), 0, blockBuf.size()*sizeof(SAMPLE));
    
    // Process until we have enough samples
    vector<SAMPLE> outBlockBuf;
    while(true)
    {
        _overlapAdd->Feed(blockBuf, blockBuf.size(), 1);        
        _overlapAdd->GetOutSamples(&outBlockBuf);
        
        if (outBlockBuf.size() >= samples->size())
            break;
    }
    
    _overlapAdd->FlushOutSamples(samples->size());
        
    memcpy(samples->data(), outBlockBuf.data(), samples->size()*sizeof(SAMPLE));
}
