/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <math.h>

#include "Complex.h"
#include "Spectrogram.h"

Spectrogram::Spectrogram(int fftSize, int numCols)
{
    _fftSize = fftSize;
    _numCols = numCols;

    _compBuf.resize(_fftSize*2*_numCols);
    _magnBuf.resize(_fftSize/2*_numCols);
    _phaseBuf.resize(_fftSize/2*_numCols);

    _magnBufDisp.resize(_fftSize/2*_numCols);
}

Spectrogram::~Spectrogram() {}

void
Spectrogram::SetCol(int colNum, const vector<SAMPLE> &compBuf)
{
    memcpy(&_compBuf.data()[colNum*_fftSize*2], compBuf.data(),
           _fftSize*2*sizeof(SAMPLE));

    _tmpMagnBuf.resize(_fftSize/2);
    _tmpPhaseBuf.resize(_fftSize/2);
    
    Complex c;
    for (int i = 0; i < _fftSize/2; i ++)
    {
        c._re = compBuf[i*2];
        c._im = compBuf[i*2 + 1];
        
        SAMPLE magn = COMP_MAGN(c);
        _tmpMagnBuf[i] = magn;

        SAMPLE phase = COMP_PHASE(c);
        _tmpPhaseBuf[i] = phase;
    }

    SetMagnCol(colNum, _tmpMagnBuf);
    SetPhaseCol(colNum, _tmpPhaseBuf);
}

void
Spectrogram::GetCol(int colNum, vector<SAMPLE> *compBuf)
{
    compBuf->resize(_fftSize*2);
    
    memcpy(compBuf->data(), &_compBuf.data()[colNum*_fftSize*2],
           _fftSize*2*sizeof(SAMPLE));
}

void
Spectrogram::GetCol(double colNum, vector<SAMPLE> *compBuf)
{
    compBuf->resize(_fftSize*2);

    if (colNum + 1 >= GetNumCols())
        // Last column, simply copy
    {
        vector<SAMPLE> compBuf0;
        GetCol((int)colNum, &compBuf0);
    
        memcpy(compBuf->data(), compBuf0.data(), compBuf0.size()*sizeof(SAMPLE));

        return;
    }

    vector<SAMPLE> magnBuf0;
    GetMagnCol((int)colNum, &magnBuf0);

    vector<SAMPLE> phaseBuf0;
    GetPhaseCol((int)colNum, &phaseBuf0);

    vector<SAMPLE> magnBuf1;
    GetMagnCol((int)colNum + 1, &magnBuf1);

    vector<SAMPLE> phaseBuf1;
    GetPhaseCol((int)colNum + 1, &phaseBuf1);
    
    double t = colNum - (int)colNum;
    for (int i = 0; i < compBuf->size()/2; i++)
    {
        if (i < compBuf->size()*0.25)
            // First half
        {
            double magn0 = magnBuf0[i];
            double magn1 = magnBuf1[i];
            double magn = (1.0 - t)*magn0 + t*magn1;

            double phase0 = phaseBuf0[i];
            double phase1 = phaseBuf1[i];
            
            //FindClosestPhase(&phase1, phase0);
            double phase = (1.0 - t)*phase0 + t*phase1;

            MAGN_PHASE_COMP(magn, phase, (*(Complex *)&compBuf->data()[i*2]));
        }
        else
            // Second half
        {
            // Just copy the first half
            *((Complex *)&compBuf->data()[i*2]) =
                *((Complex *)&compBuf->data()[compBuf->size() - i*2 - 2]);
        }
    }

    // And finally fill the second half of the comp buffer
    for (int i = 0; i < compBuf->size()/2; i++)
        (*compBuf)[compBuf->size() - i - 1] = (*compBuf)[i];
}

int
Spectrogram::GetFFTSize() const
{
    return _fftSize;
}

int
Spectrogram::GetNumCols() const
{
    return _numCols;
}

void
Spectrogram::SetMagnColDisp(int colNum, const vector<SAMPLE> &magnBuf)
{
    memcpy(&_magnBufDisp.data()[colNum*_fftSize/2], magnBuf.data(),
           (_fftSize/2)*sizeof(SAMPLE));
}

const vector<SAMPLE> *
Spectrogram::GetMagnDataDisp()
{
    return &_magnBufDisp;
}

void
Spectrogram::NormalizeMagnsDisp()
{
    SAMPLE min = 1e15;
    SAMPLE max = -1e15;

    for (int i = 0; i < _magnBufDisp.size(); i++)
    {
        SAMPLE m = _magnBufDisp[i];
        if (m < min)
            min = m;
        if (m > max)
            max = m;
    }

    SAMPLE coeff = 1.0;
    if (fabs(max - min) > 1e15)
        coeff = 1.0/(max - min);
    
    for (int i = 0; i < _magnBufDisp.size(); i++)
    {
        SAMPLE m = _magnBufDisp[i];
        m = (m - min)*coeff;
        _magnBufDisp[i] = m;
    }
}

void
Spectrogram::PreProcessPhases()
{
    vector<SAMPLE> phases;
    vector<SAMPLE> prevPhases;
    for (int i = 0; i < GetNumCols(); i++)
    {
        GetPhaseCol(i, &phases);
        
        if (i == 0)
        {
            prevPhases = phases;
            
            continue;
        }

        for (int j = 0; j < phases.size(); j++)
        {
            double phase0 = prevPhases[j];
            double phase1 = phases[j];
            
            FindClosestPhase(&phase1, phase0);

            // Real unwrap ?
            //if (phase1 < phase0)
            //    phase1 += TWO_PI;
            
            phases[j] = phase1;
        }

        SetPhaseCol(i, phases);
        
        prevPhases = phases;
    }
}

void
Spectrogram::SetMagnCol(int colNum, const vector<SAMPLE> &magnBuf)
{
    memcpy(&_magnBuf.data()[colNum*_fftSize/2], magnBuf.data(),
           (_fftSize/2)*sizeof(SAMPLE));
}

void
Spectrogram::GetMagnCol(int colNum, vector<SAMPLE> *magnBuf)
{
    magnBuf->resize(_fftSize/2);
    memcpy(magnBuf->data(), &_magnBuf.data()[colNum*_fftSize/2],
           (_fftSize/2)*sizeof(SAMPLE));
}

void
Spectrogram::SetPhaseCol(int colNum, const vector<SAMPLE> &phaseBuf)
{
    memcpy(&_phaseBuf.data()[colNum*_fftSize/2], phaseBuf.data(),
           (_fftSize/2)*sizeof(SAMPLE));
}

void
Spectrogram::GetPhaseCol(int colNum, vector<SAMPLE> *phaseBuf)
{
    phaseBuf->resize(_fftSize/2);
    memcpy(phaseBuf->data(), &_phaseBuf.data()[colNum*_fftSize/2],
           (_fftSize/2)*sizeof(SAMPLE));
}

double
Spectrogram::FmodNegative(double x, double y)
{
    // Move input to range [0 , 2*pi]
    if (x < 0.0)
    {
        double modulus = -x;
        modulus = fmod(modulus, y);
        modulus = -modulus + y;
        
        return modulus;
    }
    
    return fmod(x, y);
}

void
Spectrogram::
FindClosestPhase(double *phase, double refPhase)
{
    double refMod = FmodNegative(refPhase, TWO_PI);
    double pMod = FmodNegative(*phase, TWO_PI);
    
    // Find closest
    if (fabs((pMod - TWO_PI) - refMod) < fabs(pMod - refMod))
        pMod -= TWO_PI;
    else if (fabs((pMod + TWO_PI) - refMod) < fabs(pMod - refMod))
        pMod += TWO_PI;
    
    *phase = (refPhase - refMod) + pMod;

    // Fix the limit case
    if (*phase - refPhase > M_PI)
        *phase -= TWO_PI;
    if (*phase - refPhase < -M_PI)
        *phase += TWO_PI;
}
