/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Globals.h"

#define kiss_fft_scalar SAMPLE
#include "kiss_fft.h"

#include "Ops.h"
#include "DSP.h"
#include "Complex.h"
#include "Scale.h"
#include "Convolver.h"
#include "Stereoizer.h"

// For 44100Hz
#define IR_SIZE 2048

// Fix the gain loss after converting mono to stereo (pan law principle)
#define GAIN_ADJUST_DB 6.0 

Stereoizer::Stereoizer(double width)
{
    _width = width;
        
    vector<SAMPLE> irl;
    vector<SAMPLE> irr;
    GenerateIRs(&irl, &irr);
    
    _convolverL = new Convolver(irl);
    _convolverR = new Convolver(irr);
}

Stereoizer::~Stereoizer()
{
    delete _convolverL;
    delete _convolverR;
}

void
Stereoizer::Reset()
{
    _convolverL->Reset();
    _convolverR->Reset();
}

void
Stereoizer::SetWidth(double width)
{
    _width = width;
}

void
Stereoizer::Process(const vector<SAMPLE> &mono,
                    vector<SAMPLE> *leftChan,
                    vector<SAMPLE> *rightChan)
{       
    leftChan->resize(mono.size());
    rightChan->resize(mono.size());

    // NOTE: processing block by block gives a better sound than convolving in a single step
    
    // Process block by block
    vector<SAMPLE> monoBlock;
    monoBlock.resize(IR_SIZE);

    vector<SAMPLE> leftBlock;
    vector<SAMPLE> rightBlock;

    int samplePos = 0;
    while(samplePos < mono.size())
    {
        int numToCopy = IR_SIZE;
        if (samplePos + numToCopy > mono.size())
            numToCopy = mono.size() - samplePos;
        memcpy(monoBlock.data(), &mono.data()[samplePos], numToCopy*sizeof(SAMPLE));
            
        _convolverL->Process(monoBlock, &leftBlock);
        _convolverR->Process(monoBlock, &rightBlock);

        memcpy(&leftChan->data()[samplePos], leftBlock.data(), numToCopy*sizeof(SAMPLE));
        memcpy(&rightChan->data()[samplePos], rightBlock.data(), numToCopy*sizeof(SAMPLE));
            
        samplePos += IR_SIZE;
    }    
    
    // Adjust the gain so the processing is transparent
    AdjustGain(leftChan);
    AdjustGain(rightChan);
}

void
Stereoizer::GenerateIRs(vector<SAMPLE> *irl, vector<SAMPLE> *irr)
{
    irl->resize(IR_SIZE);
    
    // Generate the random number vector
    vector<SAMPLE> R;
    R.resize(irl->size());
    for (int i = 0; i < R.size(); i++)
    {
        // NOTE: don't normalize here, values will be bounded to [0, 1] by atan() later
        SAMPLE rd = ((SAMPLE)rand()) - RAND_MAX/2;
        
        R[i] = rd;
    }
    
    // Compute Hl
    vector<SAMPLE> Hl;
    Hl.resize(irl->size());
    for (int i = 0; i < Hl.size(); i++)
    {
        SAMPLE wr = _width*R[i];
        SAMPLE at = atan(wr);
        
        SAMPLE h = 0.5 + (1.0/M_PI)*at;
        
        Hl[i] = h;
    }
                
    // Compute Fft-1(Hl)
    vector<SAMPLE> hl;
    ComputeIFFT(Hl, &hl);
    *irl = hl;

    vector<SAMPLE> Hr;
    Hr.resize(Hl.size());
    for (int i = 0; i < Hl.size(); i++)
        Hr[i] = 1.0 - Hl[i];

    vector<SAMPLE> hr;
    ComputeIFFT(Hr, &hr);
    *irr = hr;
}

void
Stereoizer::AdjustGain(vector<SAMPLE> *samples)
{
    SAMPLE gain = Scale::DbToAmp(GAIN_ADJUST_DB);
    
    Ops::Mult(samples, gain);
}

void
Stereoizer::ComputeIFFT(const vector<SAMPLE> &inReal, vector<SAMPLE> *outReal)
{
    int fftSize = inReal.size();
    
    kiss_fft_cfg fft = kiss_fft_alloc(fftSize, 1, 0, 0);

    vector<Complex> inComp;
    inComp.resize(inReal.size());
    for (int i = 0; i < inComp.size(); i++)
    {
        inComp[i]._re = inReal[i];
        inComp[i]._im = 0.0;
    }

    vector<Complex> outComp;
    outComp.resize(inReal.size());
    
    // Apply inverse FFT
    kiss_fft(fft,
             (const kiss_fft_cpx *)inComp.data(),
             (kiss_fft_cpx *)outComp.data());

    outReal->resize(inReal.size());

    for (int i = 0; i < outComp.size(); i++)
        (*outReal)[i] = outComp[i]._re;

    // NOTE: no need to normalize here
    
    kiss_fft_free(fft);
}
