/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "Globals.h"
#include "Time.h"
#include "UIScreen.h"
#include "UIStyle.h"
#include "UIWidget.h"
#include "UIWindow.h"
#include "UICanvas.h"

#define TITLE_HEIGHT 0.0434
#define TITLE_X_MARGIN 0.0086

#define CONTENTS_Y_MARGIN 0.015

UICanvas::UICanvas(UIScreen *screen,
                   UIStyle *style,
                   const char *title,
                   double x, double y,
                   double width, double height)
{
    _screen = screen;
    _style = style;

    sprintf(_title, title);
    
    _x = x;
    _y = y;
    
    _width = width;
    _height = height;

    _mouseDown = false;
    _prevDragX = 0.0;
    _prevDragY = 0.0;

    _visible = false;

    _drawBorder = false;
    _centerTitle = false;
}

UICanvas::~UICanvas() {}

void
UICanvas::Draw()
{
    if (!_visible)
        return;

    DrawTitle();
    DrawContents();
    if (_drawBorder)
        DrawBorder();
}

void
UICanvas::SetVisible(bool flag)
{
    _visible = flag;
}

void
UICanvas::UpdateStyle(UIStyle *style)
{
    _style = style;
}

void
UICanvas::GetContentsRect(double *x, double *y,
                          double *width, double *height)
{
    double titleHeight = TITLE_HEIGHT;
    
    if (strlen(_title) == 0)
        titleHeight = 0.0;
    
    *x = _x;
    *y = _y + titleHeight + CONTENTS_Y_MARGIN;
    *width = _width;
    *height = _height - titleHeight - CONTENTS_Y_MARGIN*2.0;
}

void
UICanvas::OnMouse(int button, int state, double x, double y)
{
    if (!_visible)
        return;
    
    if ((button == GLUT_LEFT_BUTTON) &&
        (state == GLUT_UP))
        _mouseDown = false;
    
    if ((x < _x) || (y < _y) ||
        (x > _x + _width) || (y > _y + _height))
        // Outside
        return;

    bool inside = ((x >= _x) && (y >= _y) &
                   (x <= _x + _width) && (y < _y + _height));
    
    if ((button == GLUT_LEFT_BUTTON) &&
        (state == GLUT_DOWN) && inside)
    {
        _mouseDown = true;
        _prevDragX = x;
        _prevDragY = y;
    }
}

void
UICanvas::OnMouseDrag(double x, double y, double fineDragFactor)
{
    if (!_visible)
        return;
    
    if (!_mouseDown)
        return;
    
    double dx = x - _prevDragX;
    double dy = y - _prevDragY;

    dx *= fineDragFactor;
    dy *= fineDragFactor;
    
    _prevDragX = x;
    _prevDragY = y;
        
    bool inside = ((x >= _x) && (y >= _y) &
                   (x <= _x + _width) && (y < _y + _height));

    x = (x - _x)/_width;
    y = (y - _y)/_height;

    dx /= _width;
    dy /= _height;
    
    OnMouseDragNorm(x, y, dx, dy, inside);
}

bool
UICanvas::Pick(double x, double y) const
{
    if (!_visible)
        return false;
    
    bool inside = ((x >= _x) && (y >= _y) &&
                   (x <= _x + _width) && (y < _y + _height));

    return inside;
}

void
UICanvas::OnMouseDragNorm(double nx, double ny,
                          double dx, double dy, bool inside) {}

void
UICanvas::DrawBorder()
{
    double corners[4][2];
    UIWidget::ComputeCorners(corners, _screen, _x, _y, _width, _height);

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
    
    double col[4];
    _style->GetUIWindowBorderColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);
    
    double lineWidth = _style->GetUIWindowContentsLineWidth();
    glLineWidth(lineWidth);
    
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable
    
    glBegin(GL_LINE_LOOP);
    glVertex3f(corners[0][0], corners[0][1], 0.0);
    glVertex3f(corners[1][0], corners[1][1], 0.0);
    glVertex3f(corners[2][0], corners[2][1], 0.0);
    glVertex3f(corners[3][0], corners[3][1], 0.0);
    glEnd();
  
    glPopMatrix();
}

void
UICanvas::DrawContents()
{
    double lineWidth = _style->GetUIWindowContentsLineWidth();
    glLineWidth(lineWidth);
    
    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
    
    double col[4];
    _style->GetUIWindowTitleColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);

    // Children will draw contents
}

void
UICanvas::DrawTitle()
{
    if (strlen(_title) == 0)
        return;
    
    // Draw title label

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
    
    double col[4];
    _style->GetUIWindowTitleColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);
    
    GLint len = strlen(_title);

    // Compute string length in pixels
    double lenPix = glutStrokeLength(GLUT_STROKE_MONO_ROMAN,
                                    (const unsigned char *)_title);
    double offsetX = _centerTitle ?
        _width*0.5*_screen->GetWidth() - lenPix*0.5*0.1 :
        TITLE_X_MARGIN*_screen->GetHeight();
    
    double paramLineWidth = _style->GetParamLineWidth();
    glLineWidth(paramLineWidth);
    
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable

    glTranslatef(_x*_screen->GetWidth() + offsetX,
                 (1.0 - _y - TITLE_HEIGHT*0.75)*
                 _screen->GetHeight(),
                 0.0);
    glScalef(0.1, 0.1, 0.1);
    
    for(int i = 0; i < len; i++)
        glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, _title[i]);
    
    glPopMatrix();
}
