/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <math.h>

#include "Globals.h"
#include "UIScreen.h"
#include "UIStyle.h"
#include "UIVolumeEffect.h"
#include "UIStereoEffect.h"
#include "UIBitCrushEffect.h"
#include "UISaturationEffect.h"
#include "UIDelayEffect.h"
#include "UIReverbEffect.h"
#include "UIEffectsCanvas.h"

UIEffectsCanvas::UIEffectsCanvas(UIScreen *screen,
                                 UIStyle *style,
                                 const char *title,
                                 double x, double y,
                                 double width, double height)
: UICanvas(screen, style, title, x, y, width, height)
{    
    GenerateContents();
}

UIEffectsCanvas::~UIEffectsCanvas()
{
    for (int i = 0; i < EffectsBox::NUM_EFFECTS; i++)
        delete _effects[i];
}

void
UIEffectsCanvas::SetVisible(bool flag)
{
    UICanvas::SetVisible(flag);

    for (int i = 0; i < EffectsBox::NUM_EFFECTS; i++)
        _effects[i]->SetVisible(flag);
}

void
UIEffectsCanvas::OnMouse(int button, int state, double x, double y)
{
    UICanvas::OnMouse(button, state, x, y);

    for (int i = 0; i < EffectsBox::NUM_EFFECTS; i++)
        _effects[i]->OnMouse(button, state, x, y);
}

void
UIEffectsCanvas::OnMouseDrag(double x, double y, double fineDragFactor)
{
    UICanvas::OnMouseDrag(x, y, fineDragFactor);

    for (int i = 0; i < EffectsBox::NUM_EFFECTS; i++)
        _effects[i]->OnMouseDrag(x, y, fineDragFactor);
}

void
UIEffectsCanvas::DrawContents()
{
    for (int i = 0; i < EffectsBox::NUM_EFFECTS; i++)
        _effects[i]->Draw();
}

UIEffect *
UIEffectsCanvas::GetEffect(EffectsBox::EffectIdx effectIdx)
{
    return _effects[effectIdx];
}

void
UIEffectsCanvas::GenerateContents()
{
    double x;
    double y;
    double width;
    double height;
    GetContentsRect(&x, &y, &width, &height);
    
#define MARGIN 0.02

    double effectWidth = width/3.0 - 1.33*MARGIN;
    double effectHeight = height/2.0 - 1.33*MARGIN;
    
    UIVolumeEffect *volumeEffect =
        new UIVolumeEffect(_screen, _style, "Volume",
                           x + MARGIN,
                           y + MARGIN,
                           effectWidth,
                           effectHeight);
    _effects[EffectsBox::EFFECT_VOLUME] = volumeEffect;
    
    UIStereoEffect *stereoEffect =
        new UIStereoEffect(_screen, _style, "Stereo",
                           x + 2.0*MARGIN + effectWidth,
                           y + MARGIN,
                           effectWidth,
                           effectHeight);
    _effects[EffectsBox::EFFECT_STEREO] = stereoEffect;

    UISaturationEffect *saturationEffect =
        new UISaturationEffect(_screen, _style, "Sat",
                               x + 3.0*MARGIN + 2.0*effectWidth,
                               y + MARGIN,
                               //effectWidth,
                               effectWidth/2 - MARGIN/2,
                               effectHeight);
    _effects[EffectsBox::EFFECT_SATURATION] = saturationEffect;

    UIBitCrushEffect *bitCrushEffect =
        new UIBitCrushEffect(_screen, _style, "Crush",
                             x + 7.0*MARGIN/2.0 + 5.0*effectWidth/2.0,
                             y + MARGIN,
                             //effectWidth,
                             effectWidth/2 - MARGIN/2,
                             effectHeight);
    _effects[EffectsBox::EFFECT_BIT_CRUSH] = bitCrushEffect;
    
    UIDelayEffect *delayEffect =
        new UIDelayEffect(_screen, _style, "Delay",
                          x + MARGIN,
                          y + 2.0*MARGIN + effectHeight,
                          effectWidth,
                          effectHeight);
    _effects[EffectsBox::EFFECT_DELAY] = delayEffect;
    
    UIReverbEffect *reverbEffect =
        new UIReverbEffect(_screen, _style, "Reverb",
                           //x + 3.0*MARGIN + 2.0*effectWidth,
                           x + 2.0*MARGIN + effectWidth,
                           y + 2.0*MARGIN + effectHeight,
                           effectWidth*2.0 + MARGIN,
                           //effectWidth,
                           effectHeight);
    _effects[EffectsBox::EFFECT_REVERB] = reverbEffect;
}
