/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <GL/freeglut_ext.h> // for glutMouseWheelFunc()
#endif

#include "Globals.h"
#include "GrainCloud.h"
#include "Parameter.h"
#include "SoundRect.h"
#include "UIStyle.h"
#include "UIScreen.h"
#include "UIGrain.h"
#include "UISoundRect.h"
#include "UIParameter.h"
#include "Serializer.h"
#include "Time.h"
#include "UIGrainCloud.h"

#define MUTE_CROSS_SIZE 0.01

UIGrainCloud::UIGrainCloud(GrainCloud *grain,
                           UIStyle *style,
                           UIScreen *screen,
                           double x, double y,
                           unsigned int numVoices,
                           const vector<SoundRect *> &rects)
{
    _style = style;
    _screen = screen;

    _grain = grain;
    
    _startTime = Time::Instance().GetSeconds();

    _x = x;
    _y = y;

    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    _xRandExtent = 3.0/w;
    _yRandExtent = 3.0/h;
    _randExtentCoeff = 1.0;
    
    _isSelected = true;
    
    _freq = 1.0f;
    
    _rects = rects;
    
    for (int i = 0; i < numVoices; i++)
        _uiGrains.push_back(new UIGrain(_style, _screen, _x, _y));
    
    _minSelRad = 15.0f;
    _maxSelRad = 19.0f;
    _lambda = 0.997;
    _selRad = _minSelRad;
    _targetRad = _maxSelRad;
}

UIGrainCloud::~UIGrainCloud() {}

void
UIGrainCloud::SetDuration(double dur)
{
    _freq = 1000.0/dur;
}

double
UIGrainCloud::GetX()
{
    return _x;
}

double
UIGrainCloud::GetY()
{
    return _y;
}

double
UIGrainCloud::GetRadius()
{
    return _selRad;
}

void
UIGrainCloud::Draw()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
        
    double tsec = Time::Instance().GetSeconds()  - _startTime ;
    
    glPushMatrix();
    glTranslatef((GLdouble)_x*w, (GLdouble)_y*h,0.0);
    
    if (_isSelected)
    {
        double grainColorSelected[4];
        _style->GetGrainColorSelected(grainColorSelected);
        glColor4f(grainColorSelected[0], grainColorSelected[1],
                  grainColorSelected[2], grainColorSelected[3]);
    }
    else
    {
        double grainColor[4];
        _style->GetGrainColor(grainColor);
        glColor4f(grainColor[0], grainColor[1],
                  grainColor[2], grainColor[3]);
    }
    
    _selRad = _minSelRad +
        0.5*(_maxSelRad - _minSelRad)*sin(2*M_PI*(_freq*tsec + 0.125));
    gluDisk(gluNewQuadric(), _selRad, _selRad + 5.0, 128, 2);
    glPopMatrix();
    
    glPushMatrix();

    for (int i = 0; i < _uiGrains.size(); i++)
        _uiGrains[i]->Draw();
     
    glPopMatrix();

    if (_grain->IsMuted())
    {
        glLineWidth(3.0);

        if (_isSelected)
        {
            double grainColorSelected[4];
            _style->GetGrainColorSelected(grainColorSelected);
            glColor4f(grainColorSelected[0], grainColorSelected[1],
                      grainColorSelected[2], grainColorSelected[3]);
        }
        else
        {
            double grainColor[4];
            _style->GetGrainColor(grainColor);
            glColor4f(grainColor[0], grainColor[1],
                      grainColor[2], grainColor[3]);
        }
        
        glPushMatrix();

        glBegin(GL_LINES);
        glVertex2f(_x - MUTE_CROSS_SIZE*w,
                   _y - MUTE_CROSS_SIZE*h);
        glVertex2f(_x + MUTE_CROSS_SIZE*w,
                   _y + MUTE_CROSS_SIZE*h);
        glEnd();

        glBegin(GL_LINES);
        glVertex2f(_x + MUTE_CROSS_SIZE*w,
                   _y - MUTE_CROSS_SIZE*h);
        glVertex2f(_x - MUTE_CROSS_SIZE*w,
                   _y + MUTE_CROSS_SIZE*h);
        glEnd();
        
        glPopMatrix();
    }
}

void
UIGrainCloud::GetTriggerPos(unsigned int idx,
                            double *playPos, double *playVol,
                            double dur)
{
    bool trigger = false;
    SoundRect *rect = NULL;
    if (idx < _uiGrains.size())
    {
        UIGrain *grain = _uiGrains[idx];
        UpdateGrainPosition(idx,
                            _x +
                            (randf()*_xRandExtent*_randExtentCoeff -
                             randf()*_xRandExtent*_randExtentCoeff),
                            _y +
                            (randf()*_yRandExtent*_randExtentCoeff -
                             randf()*_yRandExtent*_randExtentCoeff));
        
        for (int i = 0; i < _rects.size(); i++)
        {
            rect = _rects[i];
            bool tempTrig = false;
            tempTrig =
                rect->GetUI()->GetNormedPosition(playPos,
                                                 playVol,
                                                 grain->GetX(),
                                                 grain->GetY(),
                                                 i);
            if (tempTrig)
                trigger = true;
        }
        
        // Visually trigger
        grain->Trigger(dur);
    }
}

void
UIGrainCloud::SetXRandExtentMouse(double mouseX)
{
    double w = _screen->GetWidth();
    
    _xRandExtent = fabs(mouseX - _x)/w;
    if (_xRandExtent < 2.0f/w)
        _xRandExtent = 0.0f;
}

void
UIGrainCloud::SetRandExtentMouse(double mouseX,
                                 double mouseY)
{
    SetXRandExtentMouse(mouseX);
    SetYRandExtentMouse(mouseY);
}

void
UIGrainCloud::SetYRandExtentMouse(double mouseY)
{
    double h = _screen->GetHeight();
        
    _yRandExtent = fabs(mouseY - _y)/h;
    if (_yRandExtent < 2.0f/h)
        _yRandExtent = 0.0f;
}

void
UIGrainCloud::SetXRandExtent(double value)
{
    _xRandExtent = value;
}

void
UIGrainCloud::SetYRandExtent(double value)
{
    _yRandExtent = value;
}

void
UIGrainCloud::SetRandExtent(double valueX, double valueY)
{
    SetXRandExtent(valueX);
    SetYRandExtent(valueY);
}

double
UIGrainCloud::GetRandExtentCoeff() const
{
    return _randExtentCoeff;
}

void
UIGrainCloud::SetRandExtentCoeff(double coeff)
{
    _randExtentCoeff = coeff;
}

double
UIGrainCloud::GetXRandExtent()
{
    return _xRandExtent;
}

double UIGrainCloud::GetYRandExtent()
{
    return _yRandExtent;
}

void
UIGrainCloud::UpdatePosition(double x, double y)
{
    double dx = x - _x;
    double dy = y - _y;
    _x = x;
    _y = y;
    
    for (int i = 0; i < _uiGrains.size(); i++)
    {
        double newGrainX = _uiGrains[i]->GetX() + dx;
        double newGrainY = _uiGrains[i]->GetY() + dy;
        _uiGrains[i]->MoveTo(newGrainX, newGrainY);
    }

    for (int i = 0; i < GrainCloud::NUM_PARAMS; i++)
    {
        _grain->GetParameter((GrainCloud::ParamIndex)i)->
            GetUI()->TranslateX(dx);
        _grain->GetParameter((GrainCloud::ParamIndex)i)->
            GetUI()->TranslateY(-dy);
    }
}

void
UIGrainCloud::UpdateGrainPosition(int idx, double x, double y)
{
    if (idx < _uiGrains.size())
        _uiGrains[idx]->MoveTo(x,y);
}

void
UIGrainCloud::SetMuted(bool flag)
{
    _grain->SetMuted(flag);
}

bool
UIGrainCloud::Pick(double x, double y)
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    double xdiff = (x - _x)*w;
    double ydiff = (y - _y)*h;
    
    if (sqrt(xdiff*xdiff + ydiff*ydiff) < _maxSelRad)
        return true;
    else
        return false;
}

void
UIGrainCloud::SetSelectState(bool selectState)
{
    _isSelected = selectState;
}

void
UIGrainCloud::AddGrain()
{
    _uiGrains.push_back(new UIGrain(_style, _screen, _x, _y));
}

void
UIGrainCloud::RemoveGrain()
{
    // NOTE: also delete here?
    if (_uiGrains.size() > 1)
        _uiGrains.pop_back();
}

int
UIGrainCloud::GetNumGrains() const
{
    return _uiGrains.size();
}

void
UIGrainCloud::UpdateStyle(UIStyle *style)
{
    _style = style;

    for (int i = 0; i < _uiGrains.size(); i++)
        _uiGrains[i]->UpdateStyle(style);

    for (int i = 0; i < GrainCloud::NUM_PARAMS; i++)
        _grain->GetParameter((GrainCloud::ParamIndex)i)->
            GetUI()->UpdateStyle(_style);
}

bool
UIGrainCloud::PickParameters(double x, double y)
{
    GrainCloud::ParamIndex idx = GrainCloud::NO_PARAMETER;
    for (int i = 0; i < GrainCloud::NUM_PARAMS; i++)
    {
        if (_grain->GetParameter((GrainCloud::ParamIndex)i)->
            GetUI()->Pick(x, y))
        {
            idx = (GrainCloud::ParamIndex)i;
            _grain->SetCurrentParamIdx(idx);
            
            return true;
        }
    }

    _grain->SetCurrentParamIdx(idx);

    return false;
}

void
UIGrainCloud::Serialize(Serializer *serializer)
{
    serializer->PutBool(_isOn);
    serializer->PutBool(_isSelected);
    serializer->PutDouble(_startTime);
    
    serializer->PutDouble(_xRandExtent);
    serializer->PutDouble(_yRandExtent);
    serializer->PutDouble(_randExtentCoeff);
    
    serializer->PutDouble(_freq);
    serializer->PutDouble(_x);
    serializer->PutDouble(_y);
    
    serializer->PutDouble(_selRad);
    serializer->PutDouble(_lambda);
    
    serializer->PutDouble(_maxSelRad);
    serializer->PutDouble(_minSelRad);
    serializer->PutDouble(_targetRad);
}

void
UIGrainCloud::UnSerialize(Serializer *serializer)
{
    serializer->GetBool(&_isOn);
    serializer->GetBool(&_isSelected);
    serializer->GetDouble(&_startTime);
    
    serializer->GetDouble(&_xRandExtent);
    serializer->GetDouble(&_yRandExtent);
    serializer->GetDouble(&_randExtentCoeff);
    
    serializer->GetDouble(&_freq);
    serializer->GetDouble(&_x);
    serializer->GetDouble(&_y);
    
    serializer->GetDouble(&_selRad);
    serializer->GetDouble(&_lambda);
    
    serializer->GetDouble(&_maxSelRad);
    serializer->GetDouble(&_minSelRad);
    serializer->GetDouble(&_targetRad);
}
