/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "Globals.h"
#include "Parameter.h"
#include "UIScreen.h"
#include "UIStyle.h"
#include "UIKnobWidget.h"

#define MIN_ANGLE -120.0
#define MAX_ANGLE 120.0
#define KNOB_MARK_SIZE 0.85

UIKnobWidget::UIKnobWidget(UIScreen *screen,
                           UIStyle *style,
                           const char *title,
                           double x, double y,
                           double width, double height)
: UIEffectWidget(screen, style, title, x, y, width, height) {}

UIKnobWidget::~UIKnobWidget() {}

void
UIKnobWidget::OnMouseDrag(double x, double y, double fineDragFactor)
{
    //UICanvas::OnMouseDrag(x, y);
    
#define SENSIVITY 3.0
    
    if (_leftMouseDown)
    {
        double dx = x - _prevDragX;
        double dy = y - _prevDragY;

        dx *= fineDragFactor;
        dy *= fineDragFactor;
        
        double whRatio = _screen->GetWidth()/_screen->GetHeight();
        double d = fabs(dx) > fabs(dy) ? whRatio*dx : -dy;
        
        _prevDragX = x;
        _prevDragY = y;

        double value = _param->GetNormalizedValue();
        value += d*SENSIVITY;
        _param->SetNormalizedValue(value);
    }
}

void
UIKnobWidget::DrawWidget()
{
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable

    double radius = _width*_screen->GetWidth()*0.5;
    
    glTranslatef(_x*_screen->GetWidth() + radius,
                 (1.0 - _y)*_screen->GetHeight() - radius, 0.0);

    double lineWidth = _style->GetParamLineWidth();
    
    // Draw a circle
    
    GLUquadric *quad = gluNewQuadric();
    gluQuadricDrawStyle(quad, GLU_FILL);
    gluDisk(quad, radius - lineWidth, radius, 64, 1);
    gluDeleteQuadric(quad);
    
    // Draw the knob mark
    double t = _param->GetNormalizedValue();
    double angled = MIN_ANGLE + t*(MAX_ANGLE - MIN_ANGLE) + 90.0;
    angled = 180.0 - angled;
    double angle = angled*M_PI/180.0;
        
    double dir[2] = { cos(angle), sin(angle) };

    glLineWidth(lineWidth);

    double knobMarkSize = radius*KNOB_MARK_SIZE;
    
    glBegin(GL_LINES);
    glVertex3f(dir[0]*(radius - knobMarkSize),
               dir[1]*(radius - knobMarkSize),
               0.0);
    glVertex3f(dir[0]*radius,
               dir[1]*radius,
               0.0);
    glEnd();
    
    glPopMatrix();
}
