/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "Time.h"

#include "Serializer.h"
#include "Parameter.h"
#include "UIParameter.h"
#include "UIScreen.h"
#include "UIStyle.h"
#include "UIWidget.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define PARAM_METER_MIN_LENGTH 0.004
#define PARAM_METER_MAX_LENGTH 0.25
#define METER_OFFSET 0.015f
#define METER_ALPHA_COEFF 0.5

#define LINK_ALPHA_COEFF 0.75

#define PARAMETER_RECT_WIDTH 256.0f
#define PARAMETER_RECT_HEIGHT 10.0f
#define PARAMETER_RECT_MARGIN 4.0f

UIParameter::UIParameter(Parameter *param,
                         UIScreen *screen, UIStyle *style)
{
    memset(_name, 0, 255);

    _x = 0.0;
    _y = 0.0;

    memset(_valueStr, 0, 255);

    _param = param;
    _screen = screen;
    _style = style;
}

UIParameter::~UIParameter() {}

void
UIParameter::Draw()
{
    double a = 0.8f + 0.2*sin(1.6*M_PI*Time::Instance().GetSeconds());

    double paramColor[4];
    _style->GetParamColor(paramColor);
    glColor4f(paramColor[0], paramColor[1],
              paramColor[2], a*paramColor[3]);

    double paramLineWidth = _style->GetParamLineWidth();
    glLineWidth(paramLineWidth);

    DrawBackground();
    DrawBorder();
    DrawLabel();
    DrawMeter();
    //DrawPieMenuLink();
}

void
UIParameter::SetPosition(double x, double y)
{
    _x = x;
    _y = y;
}

bool
UIParameter::Pick(double x, double y)
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    double rect[2][2] =
    {
        { _x - PARAMETER_RECT_MARGIN/w,
          _y - (PARAMETER_RECT_HEIGHT + PARAMETER_RECT_MARGIN)/h },
        { _x + (PARAMETER_RECT_WIDTH + PARAMETER_RECT_MARGIN)/w,
          _y + PARAMETER_RECT_MARGIN/h}
    };
    
    bool inside = ((x > rect[0][0]) &&
                   (y > rect[0][1]) &&
                   (x < rect[1][0]) &&
                   (y < rect[1][1]));

    return inside; 
}

void
UIParameter::SetName(const char *name)
{
    sprintf(_name, name);
}

const char *
UIParameter::GetName() const
{
    return _name;
}

void
UIParameter::SetValueStr(const char *valueStr)
{
    sprintf(_valueStr, valueStr);
}

const char *
UIParameter::GetValueStr() const
{
    return _valueStr;
}

void
UIParameter::DrawLabel()
{
    char label[255];
    sprintf(label, "%s: %s", _name, _valueStr);
            
    double scale = 100.0;
    
    GLint len = strlen(label);
    GLint i;

    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    glPushMatrix();
    glTranslatef(_x*w, (1.0 - _y)*h, 0.0);
    glScalef(.001f*scale, .001f*scale, .001f*scale);
    
    for(i = 0; i < len; i++)
        glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, label[i]);
    
    glPopMatrix();
}

void
UIParameter::DrawMeter()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    glPushMatrix();
    
    glTranslatef(_x*w, (1.0 - _y)*h, 0.0);

    double norm = _param->GetNormalizedValue();
    double lineLength = PARAM_METER_MIN_LENGTH +
        norm*(PARAM_METER_MAX_LENGTH - PARAM_METER_MIN_LENGTH);

    double a = 0.8f + 0.2*sin(1.6*M_PI*Time::Instance().GetSeconds());

    a *= METER_ALPHA_COEFF;
    
    double paramMeterLineWidth = _style->GetParamMeterLineWidth();
    double paramMeterColor[4];
    _style->GetParamMeterColor(paramMeterColor);
    
    glColor4f(paramMeterColor[0],
              paramMeterColor[1],
              paramMeterColor[2],
              a*paramMeterColor[3]);
    
    glLineWidth(paramMeterLineWidth);
    
    glBegin(GL_LINES);
    glVertex2f(0.0, -METER_OFFSET*h);
    glVertex2f(lineLength*w, -METER_OFFSET*h);
    glEnd();

    glPopMatrix();
}

void
UIParameter::DrawPieMenuLink(double menuCenterX,
                             double menuCenterY,
                             double menuCenterOffset)
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
        
    // Draw a line from the parameter label to the voice
    glPushMatrix();

    double paramLinkWidth = _style->GetParamLinkLineWidth();
    double paramLinkColor[4];
    _style->GetParamLinkColor(paramLinkColor);

    glLineWidth(paramLinkWidth);

    double a = 0.8f + 0.2*sin(1.6*M_PI*Time::Instance().GetSeconds());

    a*= LINK_ALPHA_COEFF;
    
    glColor4f(paramLinkColor[0],
              paramLinkColor[1],
              paramLinkColor[2],
              a*paramLinkColor[3]);
    
    double p0[2] = { _x*w, ((1.0f - _y) - METER_OFFSET)*h };
    double p1[2] = { menuCenterX*w, menuCenterY*h };
    double vec[2] = { p1[0] - p0[0], p1[1] - p0[1] };
    double norm = sqrt(vec[0]*vec[0] + vec[1]*vec[1]);
    double nvec[2] = { vec[0]/norm, vec[1]/norm };
    double p[2] =
        {
            p0[0] + nvec[0]*(norm - menuCenterOffset),
            p0[1] + nvec[1]*(norm - menuCenterOffset)
        };
    
    
    glBegin(GL_LINES);
    glVertex2f(_x*w, ((1.0 - _y) - METER_OFFSET)*h);
    glVertex2f(p[0], p[1]);
    glEnd();
    
    glPopMatrix();
}

void
UIParameter::UpdateStyle(UIStyle *style)
{
    _style = style;
}

void
UIParameter::Serialize(Serializer *serializer)
{
    serializer->PutDouble(_x);
    serializer->PutDouble(_y);
}
    
void
UIParameter::UnSerialize(Serializer *serializer)
{
    serializer->GetDouble(&_x);
    serializer->GetDouble(&_y);
}

double
UIParameter::GetPosX() const
{
    return _x;
}

double
UIParameter::GetPosY() const
{
    return _y;
}

void
UIParameter::TranslateX(double dx)
{
    _x += dx;
}

void
UIParameter::TranslateY(double dy)
{
    _y += dy;
}

void
UIParameter::DrawBackground()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();

    double corners[4][2];
    UIWidget::
        ComputeCorners(corners,
                       _screen,
                       _x - PARAMETER_RECT_MARGIN/w,
                       _y - (PARAMETER_RECT_HEIGHT +
                        PARAMETER_RECT_MARGIN)/h,
                       (PARAMETER_RECT_WIDTH + PARAMETER_RECT_MARGIN)/w,
                       (PARAMETER_RECT_HEIGHT +
                        PARAMETER_RECT_MARGIN*2)/h);

    glPushMatrix();

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
        
    double col[4];
    _style->GetUIWindowBackgroundColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glBegin(GL_QUADS);
    glVertex3f(corners[0][0], corners[0][1], 0.0);
    glVertex3f(corners[1][0], corners[1][1], 0.0);
    glVertex3f(corners[2][0], corners[2][1], 0.0);
    glVertex3f(corners[3][0], corners[3][1], 0.0);
    glEnd();
    
    glPopMatrix();
}

void
UIParameter::DrawBorder()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();

    double corners[4][2];
    UIWidget::
        ComputeCorners(corners,
                       _screen,
                       _x - PARAMETER_RECT_MARGIN/w,
                       _y - (PARAMETER_RECT_HEIGHT +
                        PARAMETER_RECT_MARGIN)/h,
                       (PARAMETER_RECT_WIDTH + PARAMETER_RECT_MARGIN)/w,
                       (PARAMETER_RECT_HEIGHT +
                        PARAMETER_RECT_MARGIN*2)/h);

    glPushMatrix();

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
        
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    double col[4];
    _style->GetUIWindowBorderColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);
    
    double lineWidth = _style->GetUIWindowBorderWidth();
    glLineWidth(lineWidth);
    
    glBegin(GL_LINE_LOOP);
    glBegin(GL_QUADS);
    glVertex3f(corners[0][0], corners[0][1], 0.0);
    glVertex3f(corners[1][0], corners[1][1], 0.0);
    glVertex3f(corners[2][0], corners[2][1], 0.0);
    glVertex3f(corners[3][0], corners[3][1], 0.0);
    glEnd();
    
    glPopMatrix();
}
