/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */
 
#include "App.h"
#include "ParameterProvider.h"
#include "UIScreen.h"
#include "UIParameter.h"
#include "UIPieMenu.h"

#define MIN_DISTANCE 0.2
#define MAX_DISTANCE 1.2

// Experimental parallax effect on the pie menu
#define USE_PARALLAX 0 //1
#define PARALLAX_FACTOR 0.4

UIPieMenu::UIPieMenu(App *app,
                     UIScreen *screen)
{
    _app = app;
    _screen = screen;
    
    _visible = false;
}

UIPieMenu::~UIPieMenu() {}

bool
UIPieMenu::IsVisible() const
{
    return _visible;
}

void
UIPieMenu::SetVisible(bool flag)
{
    _visible = flag;
}

void
UIPieMenu::Draw()
{
    if (!_visible)
        return;

    ParameterProvider *prov = _app->GetCurrentParamProvider();
    if (prov == NULL)
        return;
    
    if (prov != NULL)
    {
        int numParams = prov->GetNumParams();
        for (int i = 0; i < numParams; i++)
        {
            Parameter *param = prov->GetParameter(i);
            param->UpdateValue();
            
            param->GetUI()->Draw();

            double cx = prov->GetUIX();
            double cy = prov->GetUIY();
            double rad = prov->GetUIRadius();
            param->GetUI()->DrawPieMenuLink(cx, cy, rad);
        }
    }
}

bool
UIPieMenu::OnMouse(int button, int state, double x, double y)
{
    if (!_visible)
        return false;
    
    if (((button == GLUT_LEFT_BUTTON) ||
         (button == GLUT_RIGHT_BUTTON)) &&
        (state == GLUT_DOWN))
    {
        bool picked = _app->PickParameters(x, 1.0 - y);
        
        return picked;
    }

    // Reset param only if left mouse button
    // With right button, we can choose param string to edit
    // even after right button released
    if ((button == GLUT_LEFT_BUTTON) &&
        (state == GLUT_UP))
        _app->ResetCurrentParam();

    return false;
}

void
UIPieMenu::OnMouseDrag(double dx, double dy)
{
    ParameterProvider *prov = _app->GetCurrentParamProvider();
    if (prov == NULL)
        return;
    
    Parameter *param = prov->GetCurrentParameter();
    if (param != NULL)
    {
        // Move parameter display
        dy = -dy;

        param->GetUI()->TranslateX(dx);
        param->GetUI()->TranslateY(dy);

        // Change the parameter value

        // Center
        double cx = prov->GetUIX();
        double cy = prov->GetUIY();
        cy = 1.0 - cy;

        double x = param->GetUI()->GetPosX();
        double y = param->GetUI()->GetPosY();
        
        double whRatio = _screen->GetWidth()/_screen->GetHeight();

        double dir[2] = { (x - cx)*whRatio, y - cy };
        
        double dist = sqrtf(dir[0]*dir[0] + dir[1]*dir[1]);
        double norm = (dist - MIN_DISTANCE)/(MAX_DISTANCE - MIN_DISTANCE);

        norm *= param->GetSensivity();
        
        if (norm < 0.0)
            norm = 0.0;
        if (norm > 1.0)
            norm = 1.0;

        param->SetNormalizedValue(norm);
        param->UpdateValue();

        _app->OnCurrentParamValueUpdated();
    }

#if USE_PARALLAX
    // Parallax
    if (param == NULL)
    {
        int numParams = prov->GetNumParams();
        for (int i = 0; i < numParams; i++)
        {
            Parameter *param0 = prov->GetParameter(i);

            param0->GetUI()->TranslateX(dx*PARALLAX_FACTOR);
            param0->GetUI()->TranslateY(-dy*PARALLAX_FACTOR);
        }
    }
#endif
}

void
UIPieMenu::InitFromParameters()
{
    ParameterProvider *prov = _app->GetCurrentParamProvider();
    if (prov == NULL)
        return;

    double whRatio = _screen->GetWidth()/_screen->GetHeight();
 
    int numParams = prov->GetNumParams();
    for (int i = 0; i < numParams; i++)
    {
        Parameter *param = prov->GetParameter(i);

        // Angle
        double angle = 2.0*M_PI*((double)i)/numParams;

        // Normalized value
        prov->GetParameter(i)->UpdateValue();
        double norm = prov->GetParameter(i)->GetNormalizedValue();

        norm /= param->GetSensivity();

        // Distance
        double dist = MIN_DISTANCE + norm*(MAX_DISTANCE - MIN_DISTANCE);

        // Param relative position
        double paramX = cos(angle)*dist/whRatio;
        double paramY = sin(angle)*dist;

        // Center
        double cx = prov->GetUIX();
        double cy = prov->GetUIY();
        cy = 1.0 - cy;

        // Add center
        paramX += cx;
        paramY += cy;
        
        param->GetUI()->SetPosition(paramX, paramY);
    }
}

void
UIPieMenu::UpdateCurrentParamPos()
{
    ParameterProvider *prov = _app->GetCurrentParamProvider();
    if (prov == NULL)
        return;

    Parameter *param = prov->GetCurrentParameter();
    if (param == NULL)
        return;

    double x = param->GetUI()->GetPosX();
    double y = param->GetUI()->GetPosY();
    
    // Center
    double cx = prov->GetUIX();
    double cy = prov->GetUIY();
    cy = 1.0 - cy;

    // Relative position
    x -= cx;
    y -= cy;

    double whRatio = _screen->GetWidth()/_screen->GetHeight();
        
    // Direction
    double dir[2] = { x*whRatio, y };

    // Distance
    double dist = sqrtf(dir[0]*dir[0] + dir[1]*dir[1]);

    // Norm direction
    if (dist > 0.0)
    {
        dir[0] /= dist;
        dir[1] /= dist;
    }

    param->UpdateValue();
    _app->OnCurrentParamValueUpdated();

    // Normalized value
    double norm = param->GetNormalizedValue();
    norm /= param->GetSensivity();

    // New distance
    double newDist = MIN_DISTANCE + norm*(MAX_DISTANCE - MIN_DISTANCE);

    // New positon
    double newX = cx + dir[0]*newDist/whRatio;
    double newY = cy + dir[1]*newDist;
    
    param->GetUI()->SetPosition(newX, newY);
}
