/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <math.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <GL/freeglut_ext.h> // for glutMouseWheelFunc()
#endif

#include "Globals.h"
#include "SingleGrain.h"
#include "Parameter.h"
#include "SoundRect.h"
#include "UIStyle.h"
#include "UIScreen.h"
#include "UISoundRect.h"
#include "UIParameter.h"
#include "Serializer.h"
#include "Time.h"
#include "UIGrain.h"
#include "UISingleGrain.h"

#define MUTE_CROSS_SIZE 0.01

UISingleGrain::UISingleGrain(SingleGrain *grain,
                             UIStyle *style,
                             UIScreen *screen,
                             double x, double y,
                             vector<SoundRect *> *rects)
{
    _style = style;
    _screen = screen;

    _grain = grain;
    
    _startTime = Time::Instance().GetSeconds();

    _x = x;
    _y = y;

    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    _isSelected = true;
    
    _freq = 1.0f;
    
    _rects = rects;
    
    _uiGrain = new UIGrain(_style, _screen, _x, _y);
    
    _minSelRad = 15.0f;
    _maxSelRad = 19.0f;
    _lambda = 0.997;
    _selRad = _minSelRad;
    _targetRad = _maxSelRad;
}

UISingleGrain::~UISingleGrain() {}

void
UISingleGrain::SetDuration(double dur)
{
    _freq = 1000.0/dur;
}

double
UISingleGrain::GetX()
{
    return _x;
}

double
UISingleGrain::GetY()
{
    return _y;
}

double
UISingleGrain::GetRadius()
{
    return _selRad;
}

void
UISingleGrain::Draw()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
        
    double t_sec = Time::Instance().GetSeconds()  - _startTime ;
    
    glPushMatrix();
    glTranslatef((GLdouble)_x*w, (GLdouble)_y*h, 0.0);
    
    if (_isSelected)
    {
        double grainColorSelected[4];
        _style->GetGrainColorSelected(grainColorSelected);
        glColor4f(grainColorSelected[0], grainColorSelected[1],
                  grainColorSelected[2], grainColorSelected[3]);
    }
    else
    {
        double grainColor[4];
        _style->GetGrainColor(grainColor);
        glColor4f(grainColor[0], grainColor[1],
                  grainColor[2], grainColor[3]);
    }
    
    _selRad = _minSelRad +
        0.5*(_maxSelRad - _minSelRad)*sin(2*M_PI*(_freq*t_sec + 0.125));
    gluDisk(gluNewQuadric(), _selRad, _selRad + 5.0, 128, 2);
    glPopMatrix();
    
    glPushMatrix();

    _uiGrain->Draw();
     
    glPopMatrix();

    if (_grain->IsMuted())
    {
        glLineWidth(3.0);

        if (_isSelected)
        {
            double grainColorSelected[4];
            _style->GetGrainColorSelected(grainColorSelected);
            glColor4f(grainColorSelected[0], grainColorSelected[1],
                      grainColorSelected[2], grainColorSelected[3]);
        }
        else
        {
            double grainColor[4];
            _style->GetGrainColor(grainColor);
            glColor4f(grainColor[0], grainColor[1],
                      grainColor[2], grainColor[3]);
        }
        
        glPushMatrix();

        glBegin(GL_LINES);
        glVertex2f(_x - MUTE_CROSS_SIZE*w,
                   _y - MUTE_CROSS_SIZE*h);
        glVertex2f(_x + MUTE_CROSS_SIZE*w,
                   _y + MUTE_CROSS_SIZE*h);
        glEnd();

        glBegin(GL_LINES);
        glVertex2f(_x + MUTE_CROSS_SIZE*w,
                   _y - MUTE_CROSS_SIZE*h);
        glVertex2f(_x - MUTE_CROSS_SIZE*w,
                   _y + MUTE_CROSS_SIZE*h);
        glEnd();
        
        glPopMatrix();
    }
}

void
UISingleGrain::GetTriggerPos(double *playPos, double *playVol,
                             double dur)
{
    bool trigger = false;
    SoundRect *rect = NULL;
    UpdateGrainPosition(_x, _y);
        
    for (int i = 0; i < _rects->size(); i++)
    {
        rect = (*_rects)[i];
        bool tempTrig = false;
        tempTrig =
            rect->GetUI()->GetNormedPosition(playPos,
                                             playVol,
                                             _uiGrain->GetX(),
                                             _uiGrain->GetY(),
                                             i);
        if (tempTrig)
            trigger = true;
    }
        
    if (trigger)
        _uiGrain->Trigger(dur);
}

void
UISingleGrain::UpdatePosition(double x, double y)
{
    double dx = x - _x;
    double dy = y - _y;
    _x = x;
    _y = y;
    
    double newGrainX = _uiGrain->GetX() + dx;
    double newGrainY = _uiGrain->GetY() + dy;
    _uiGrain->MoveTo(newGrainX, newGrainY);

    for (int i = 0; i < SingleGrain::NUM_PARAMS; i++)
    {
        _grain->GetParameter((SingleGrain::ParamIndex)i)->
            GetUI()->TranslateX(dx);
        _grain->GetParameter((SingleGrain::ParamIndex)i)->
            GetUI()->TranslateY(-dy);
    }
}

void
UISingleGrain::UpdateGrainPosition(double x, double y)
{
    _uiGrain->MoveTo(x, y);
}

void
UISingleGrain::SetMuted(bool flag)
{
    _grain->SetMuted(flag);
}

bool
UISingleGrain::Pick(double x, double y)
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    double xdiff = (x - _x)*w;
    double ydiff = (y - _y)*h;
    
    if (sqrt(xdiff*xdiff + ydiff*ydiff) < _maxSelRad)
        return true;
    else
        return false;
}

void
UISingleGrain::SetSelectState(bool selectState)
{
    _isSelected = selectState;
}

void
UISingleGrain::UpdateStyle(UIStyle *style)
{
    _style = style;
    
    _uiGrain->UpdateStyle(style);

    for (int i = 0; i < SingleGrain::NUM_PARAMS; i++)
        _grain->GetParameter((SingleGrain::ParamIndex)i)->
            GetUI()->UpdateStyle(_style);
}

bool
UISingleGrain::PickParameters(double x, double y)
{
    SingleGrain::ParamIndex idx = SingleGrain::NO_PARAMETER;
    for (int i = 0; i < SingleGrain::NUM_PARAMS; i++)
    {
        if (_grain->GetParameter((SingleGrain::ParamIndex)i)->
            GetUI()->Pick(x, y))
        {
            idx = (SingleGrain::ParamIndex)i;
            _grain->SetCurrentParamIdx(idx);
            
            return true;
        }
    }

    _grain->SetCurrentParamIdx(idx);

    return false;
}

void
UISingleGrain::Serialize(Serializer *serializer)
{
    serializer->PutBool(_isOn);
    serializer->PutBool(_isSelected);
    serializer->PutDouble(_startTime);
    
    serializer->PutDouble(_freq);
    serializer->PutDouble(_x);
    serializer->PutDouble(_y);
    
    serializer->PutDouble(_selRad);
    serializer->PutDouble(_lambda);
    
    serializer->PutDouble(_maxSelRad);
    serializer->PutDouble(_minSelRad);
    serializer->PutDouble(_targetRad);
}

void
UISingleGrain::UnSerialize(Serializer *serializer)
{
    serializer->GetBool(&_isOn);
    serializer->GetBool(&_isSelected);
    serializer->GetDouble(&_startTime);
    
    serializer->GetDouble(&_freq);
    serializer->GetDouble(&_x);
    serializer->GetDouble(&_y);
    
    serializer->GetDouble(&_selRad);
    serializer->GetDouble(&_lambda);
    
    serializer->GetDouble(&_maxSelRad);
    serializer->GetDouble(&_minSelRad);
    serializer->GetDouble(&_targetRad);
}
