/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "Globals.h"
#include "App.h"
#include "Parameter.h"
#include "Serializer.h"
#include "Spectrogram.h"
#include "FilterBank.h"
#include "Scale.h"
#include "WaveformPyramid.h"
#include "UIParameter.h"
#include "UIStyle.h"
#include "UIScreen.h"
#include "UISoundRect.h"


// Waveform
#define MIN_DB_WF -30.0
#define MAX_DB_WF -0.00001

// Spectrogram
#define MIN_DB_SP -50.0
#define MAX_DB_SP -0.00001

#define MAX_SPECTRO_ALPHA 0.7

// Force single channel display?
#define FORCE_DISPLAY_ONE_CHANNEL 1

UISoundRect::UISoundRect(SoundRect *rect,
                         App *app,
                         UIStyle *style,
                         UIScreen *screen)
{
    _rect = rect;
    
    _paramProviderType = ParameterProvider::SOUND_RECT;
    
    double w = screen->GetWidth();
    double h = screen->GetHeight();
    
    _app = app;
    _style = style;
    _screen = screen;

    _width = 0.0;
    _height = 0.0;

    _top = 0.0;
    _bot = 0.0;
    _left = 0.0;
    _right = 0.0;

    _x = 0.0;
    _y = 0.0;
    _z = 0.0;

    _isSelected = false;
    
    _minDim = 0.0;
    _startTime = 0.0;
    
    _orientation = HORIZONTAL;
    
    memset(_name, 0, 255);

    _rectType = WAVEFORM;
    _colorRange = 0.7;
    _colorContrast = 0.99;

    _needUpdatePyramid = true;
    
    _spectroTexId = -1;
    _chromaTexId = -1;
    _needUpdateTextures = true;
    
    Init();

    InitParameters();
    _currentParamIdx = NO_PARAMETER;
    double paramsSensivity[NUM_PARAMS] =
    {
        6.0, // TYPE
        6.0, // RANGE
        6.0, // CONTRAST
        6.0 // ORIENTATION
    };
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->SetSensivity(paramsSensivity[i]);
    
    double xBorder = 100.0/w;
    double yBorder = 50.0/w;
    
    // Translation coordinates
    _x = xBorder + ((double)rand()/RAND_MAX) * (1.0 - xBorder * 2.0);
    _y = yBorder + ((double)rand()/RAND_MAX) * (1.0 - yBorder * 2.0);
    
    // Min w and h dim
    _minDim = 60.0f/w;
    
    // Scale factor for randomization
    double scaleF = 0.5;
    
    // Box corners
    SetDimensions(_minDim + scaleF*((double)rand()/RAND_MAX),
                  _minDim + scaleF*((double)rand()/RAND_MAX));  

#if 0
    // Set orientation
    if (randf() < 0.5)
        _orientation = HORIZONTAL;
    else 
        _orientation = VERTICAL;
#endif

    _filterBank = new FilterBank(Scale::MEL);

    for (int i = 0; i < 2; i++)
        _pyramids[i] = new WaveformPyramid();

    _pulseTimeOffset = (((double)rand())/RAND_MAX)*10.0;
}

UISoundRect::~UISoundRect()
{
    for (int i = 0; i < 2; i++)
        delete _pyramids[i];
}

// Other intialization code
void
UISoundRect::Init()
{    
    // Selection state
    _isSelected = false;
    
    // Get start time
    _startTime = Time::Instance().GetSeconds();
}

bool
UISoundRect::GetSelectState() const
{
    return _isSelected;
}

void
UISoundRect::SetSelectState(bool state)
{
    _isSelected = state;
}

void
UISoundRect::Move(double xDiff, double yDiff)
{
    _x = _x + xDiff;
    _y = _y + yDiff;
    UpdateCorners(_width,_height);
}

void
UISoundRect::ToggleOrientation()
{
    if (_orientation == HORIZONTAL)
        _orientation == VERTICAL;
    else
        _orientation == HORIZONTAL;
    

    if (_orientation == HORIZONTAL)
        SetDimensions(_height, _width);
    else
        SetDimensions(_height, _width);
}

// Set width and height
void
UISoundRect::SetDimensions(double width, double height)
{    
    bool newWidth = false;
    bool newHeight = false;
    if (width >= _minDim)
    { 
        _width = width;
        newWidth = true;
    }
    
    if (height >= _minDim)
    {
        _height = height;
        newHeight = true;
    }
    
    if (newWidth || newHeight)
        UpdateCorners(_width, _height);
}

double
UISoundRect::GetWidth()
{
    return _width;
}

double
UISoundRect::GetHeight()
{
    return _height;
}

// Update box corners with new width values
void UISoundRect::UpdateCorners(double width, double height)
{
    _top = _y + height * 0.5f;
    _bot = _y - height * 0.5f;
    _right = _x + width * 0.5f;
    _left = _x - width * 0.5f;
}

UISoundRect::RectType
UISoundRect::GetRectType() const
{
    return _rectType;
}

void
UISoundRect::SetRectType(RectType type)
{
    _rectType = type;
}

double
UISoundRect::GetColorRange() const
{
    return _colorRange;
}
        
void
UISoundRect::SetColorRange(double range)
{
    if (fabs(range - _colorRange) > 1e-15)
    {
        _colorRange = range;
        _needUpdateTextures = true;
    }
}

double
UISoundRect::GetColorContrast() const
{
    return _colorContrast;
}

void
UISoundRect::SetColorContrast(double contrast)
{
    if (fabs(contrast - _colorContrast) > 1e-15)
    {
        _colorContrast = contrast;
        _needUpdateTextures = true;
    }
}

UISoundRect::Orientation
UISoundRect::GetOrientation() const
{
    return _orientation;
}

void
UISoundRect::SetOrientation(Orientation orientation)
{
    _orientation = orientation;
}

void
UISoundRect::UpdateStyle(UIStyle *style)
{
    _style = style;
}

void
UISoundRect::Serialize(Serializer *serializer)
{
    serializer->PutDouble(_width);
    serializer->PutDouble(_height);

    serializer->PutDouble(_top);
    serializer->PutDouble(_bot);
    serializer->PutDouble(_left);
    serializer->PutDouble(_right);

    serializer->PutDouble(_x);
    serializer->PutDouble(_y);
    serializer->PutDouble(_z);

    serializer->PutBool(_isSelected);
    
    serializer->PutDouble(_minDim);
    
    serializer->PutDouble(_startTime);
    
    serializer->PutInt(_orientation);
    
    SerializeParameters(serializer);
    serializer->PutInt(_currentParamIdx);

    serializer->PutInt(_rectType);
    serializer->PutDouble(_colorRange);
    serializer->PutDouble(_colorContrast);
}

void
UISoundRect::UnSerialize(Serializer *serializer)
{
    serializer->GetDouble(&_width);
    serializer->GetDouble(&_height);

    serializer->GetDouble(&_top);
    serializer->GetDouble(&_bot);
    serializer->GetDouble(&_left);
    serializer->GetDouble(&_right);

    serializer->GetDouble(&_x);
    serializer->GetDouble(&_y);
    serializer->GetDouble(&_z);

    serializer->GetBool(&_isSelected);
    
    serializer->GetDouble(&_minDim);
    
    serializer->GetDouble(&_startTime);
    
    serializer->GetInt((int *)&_orientation);

    UnSerializeParameters(serializer);
    serializer->GetInt((int *)&_currentParamIdx);

    serializer->GetInt((int *)&_rectType);
    serializer->GetDouble(&_colorRange);
    serializer->GetDouble(&_colorContrast);
}

Parameter *
UISoundRect::GetCurrentParameter()
{
    if (_currentParamIdx == NO_PARAMETER)
        return NULL;

    return _params[_currentParamIdx];
}

bool
UISoundRect::PickParameters(double x, double y)
{
    UISoundRect::ParamIndex idx = UISoundRect::NO_PARAMETER;
    for (int i = 0; i < UISoundRect::NUM_PARAMS; i++)
    {
        if (GetParameter((UISoundRect::ParamIndex)i)->
            GetUI()->Pick(x, y))
        {
            idx = (UISoundRect::ParamIndex)i;
            _currentParamIdx = idx;
            
            return true;
        }
    }

    _currentParamIdx = idx;

    return false;
}

void
UISoundRect::InitParameters()
{
    _params[TYPE] = new ParameterSoundRectType(_app, TYPE);
    _params[RANGE] = new ParameterColorRange(_app, RANGE);
    _params[CONTRAST] = new ParameterColorContrast(_app, CONTRAST);
    _params[ORIENTATION] =
        new ParameterSoundRectOrientation(_app, ORIENTATION);
}

void
UISoundRect::UpdatePyramid()
{
    if (_needUpdatePyramid)
    {
        int nFrames;
        int nChans;
        vector<SAMPLE> *samples = _rect->GetSamples(&nFrames, &nChans);
        
        SetupPyramid(samples, nFrames, nChans);

        _needUpdatePyramid = false;
    }
}

void
UISoundRect::UpdateTextures()
{
    if (_needUpdateTextures)
    {
        if (_spectroTexId != -1)
            glDeleteTextures(1, (unsigned int *)&_spectroTexId);
        CreateSpectroTexture();

        if (_chromaTexId != -1)
            glDeleteTextures(1, (unsigned int *)&_chromaTexId);
        CreateChromaTexture();

        _needUpdateTextures = false;
    }
}

void
UISoundRect::SerializeParameters(Serializer *serializer)
{
    int size = NUM_PARAMS;
    serializer->PutInt(size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
UISoundRect::UnSerializeParameters(Serializer *serializer)
{
    int size;
    serializer->GetInt(&size);
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);
}

Parameter *
UISoundRect::GetParameter(int paramIdx)
{
    if (paramIdx == NO_PARAMETER)
        return NULL;
    return _params[paramIdx];
}

int
UISoundRect::GetNumParams() const
{
    return (int)NUM_PARAMS;
}

double
UISoundRect::GetUIX() const
{
    return _x;
}

double
UISoundRect::GetUIY() const
{
    return _y;
}

double
UISoundRect::GetUIRadius() const
{
    return _height*0.5;
}

void
UISoundRect::ResetCurrentParameterIndex()
{
    _currentParamIdx = NO_PARAMETER;
}

void
UISoundRect::UpdateRectPosition(double x, double y)
{
    double dx = x - _x;
    double dy = y - _y;
    _x = x;
    _y = y;

    UpdateCorners(_width, _height);
    
    for (int i = 0; i < UISoundRect::NUM_PARAMS; i++)
    {
        GetParameter((UISoundRect::ParamIndex)i)->
            GetUI()->TranslateX(dx);
        GetParameter((UISoundRect::ParamIndex)i)->
            GetUI()->TranslateY(-dy);
    }
}

void
UISoundRect::SetupPyramid(vector<SAMPLE> *buf, int nFrames, int nChan)
{
    // Pyramid
    for (int k = 0; k < nChan; k++)
    {
#if FORCE_DISPLAY_ONE_CHANNEL
        if (k > 0)
            break;
#endif
        vector<SAMPLE> samples;
        samples.resize(nFrames);
        for (int i = 0; i < nFrames; i++)
            samples[i] = (*buf)[i*nChan + k];
            
        _pyramids[k]->SetValues(samples);
    }
}

void
UISoundRect::Draw()
{
    DrawRectangle();

    UpdatePyramid();
    
    UpdateTextures();

    if (_rectType == WAVEFORM)
        DrawWaveform();
    else if (_rectType == SPECTROGRAM)
        DrawSpectro();
    else if (_rectType == CHROMAGRAM)
        DrawChroma();

    DrawName();
}

void
UISoundRect::DrawRectangle()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    glPushMatrix();

    double col[4];
    _style->GetSoundRectColor(col);

    // Pulsation
    double tsec = Time::Instance().GetSeconds();
    double freq = 0.25;
    double colorCoeff = 1.0 + 0.5*sin(2*M_PI*(freq*tsec + _pulseTimeOffset));
    for (int i = 0; i < 4; i++)
        col[i] *= colorCoeff;

    glColor4f(col[0], col[1], col[2], col[3]);
        
    // Draw rectangle
    glBegin(GL_QUADS);
    glVertex3f(_left*w, _top*h, 0.0f);
    glVertex3f(_right*w,  _top*h, 0.0f);
    glVertex3f(_right*w, _bot*h, 0.0f);
    glVertex3f(_left*w, _bot*h, 0.0f);
    glEnd();
    
    if (_isSelected)
    {
        double soundRectBorderColorSelected[4];
        _style->GetSoundRectBorderColorSelected(soundRectBorderColorSelected);
        glColor4f(soundRectBorderColorSelected[0],
                  soundRectBorderColorSelected[1],
                  soundRectBorderColorSelected[2],
                  soundRectBorderColorSelected[3]);
    }
    else
    {
        double soundRectBorderColor[4];
        _style->GetSoundRectBorderColor(soundRectBorderColor);
        glColor4f(soundRectBorderColor[0], soundRectBorderColor[1],
                  soundRectBorderColor[2], soundRectBorderColor[3]);
    }

    double lineWidth = _style->GetSoundRectLineWidth();
    glLineWidth(lineWidth);
        
    glBegin(GL_LINE_STRIP);
    glVertex3f(_left*w, _top*h, 0.0f);
    glVertex3f(_right*w, _top*h, 0.0f);
    glVertex3f(_right*w, _bot*h, 0.0f);
    glVertex3f(_left*w, _bot*h, 0.0f);
    glVertex3f(_left*w, _top*h, 0.0f);
    glEnd();

    glPopMatrix();
}

void
UISoundRect::DrawName()
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();

    glLineWidth(1.0);
    
    // Draw name
    glPushMatrix();

    double col[4];

    if (_isSelected)
        _style->GetSoundRectNameColorSelected(col);
    else
        _style->GetSoundRectNameColor(col);
    
    glColor4f(col[0], col[1], col[2], col[3]);

#define FONT_HEIGHT 0.02
#define TEXT_OFFSET 0.01
    
    double stringLengthPix =
        glutStrokeLength(GLUT_STROKE_MONO_ROMAN,
                         (const unsigned char *)_name);

    double strScale = 1.0/(stringLengthPix/w);
    if (strScale > 1.0)
        strScale = 1.0;
    
    double x0 = (_x + TEXT_OFFSET - _width*0.5)*w;
    double x1 = (x0/w + _width - TEXT_OFFSET)*w;
    double y = ((1.0 - _y) + TEXT_OFFSET + FONT_HEIGHT - _height*0.5)*h;

    double scale = strScale*0.5*(x1 - x0)/w;
    if (scale > 0.1)
        scale = 0.1;
    
    _app->DrawString2(x0, y, 0.0f, _name, scale);
    
    glPopMatrix();
}

void
UISoundRect::DrawWaveform()
{
#define NUM_SAMPLES_COEFF 4.0
    
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();
    
    // Compute the number of samples we want from the pyramid
    double rectScale = (_orientation == HORIZONTAL) ? _width : _height;
    int numSamplesToDisplay = w*_app->GetZoom()*rectScale;
    numSamplesToDisplay *= NUM_SAMPLES_COEFF;
    
    // Draw audio buffer using pyramid
    double lineWidth = _style->GetWaveLineWidth();
    glLineWidth(lineWidth);
    
    double waveColor[4];
    _style->GetWaveColor(waveColor);
    glColor4f(waveColor[0], waveColor[1],
              waveColor[2], waveColor[3]);
    
    double pointSize = _style->GetWavePointSize();
    glPointSize(pointSize);

    int nChans = 1;
    for (int k = 0; k < nChans; k++)
    {
        //#if FORCE_DISPLAY_ONE_CHANNEL
        //if (k > 0)
        //    break;
        //#endif
        vector<SAMPLE> samples;
        _pyramids[k]->GetValues(0.0, 1.0, numSamplesToDisplay,
                                &samples);
        
        glPushMatrix();
        
        glTranslatef(_x*w, _y*h, 0.0);
        
        if ((nChans == 2) && !FORCE_DISPLAY_ONE_CHANNEL)
        {
            if (_orientation == HORIZONTAL)
            {
                if (k == 0)
                {
                    glTranslatef(0.0, -0.25*_height*h, 0.0);
                    glScalef(1.0, 0.5, 1.0);
                }
                else
                {
                    glTranslatef(0.0, 0.25*_height*h, 0.0);
                    glScalef(1.0, 0.5, 1.0);
                }
            }
            else if (_orientation == VERTICAL)
            {
                if (k == 0)
                {
                    glTranslatef(-0.25*_width*w, 0.0, 0.0);
                    glScalef(0.5, 1.0, 1.0);
                }
                else
                {
                    glTranslatef(0.25*_width*w, 0.0, 0.0);
                    glScalef(0.5, 1.0, 1.0);
                }
            }
        }
            
        glBegin(GL_LINE_STRIP);
        for (int i = 0; i < samples.size(); i++)
        {
            double t = ((double)i)/(samples.size() - 1);
            
            SAMPLE s = samples[i];
            
            // Makes artifacts on the waveform near 0 
#if WAVEFORM_DISPLAY_DB
            s = Scale::AmpToDbNorm(s, MIN_DB_WF, MAX_DB_WF);
#endif
            
            if (_orientation == HORIZONTAL)
                glVertex3f((t - 0.5)*_width*w,
                           s*0.5f*_height*h,
                           0.0f);
            else if (_orientation == VERTICAL)
                glVertex3f(s*0.5f*_width*w,
                           (t - 0.5)*_height*h,
                           0.0f);
        }
        glEnd();
        
        glPopMatrix();
    }
}

void
UISoundRect::DrawSpectro()
{
    DrawSpectro(_spectroTexId);
}

void
UISoundRect::DrawChroma()
{
    DrawSpectro(_chromaTexId);
}

void
UISoundRect::DrawSpectro(int texId)
{
    double w = _screen->GetWidth();
    double h = _screen->GetHeight();

    // TODO: keep this?
    glColor4f(1.0, 1.0, 1.0, 1.0);
    
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texId);
    
    // Draw textured rectangle
    if (_orientation == HORIZONTAL)
    {
        glBegin(GL_QUADS);
        glTexCoord2f(1.0, 0.0);
        glVertex3f(_left*w, _top*h, 0.0f);
        
        glTexCoord2f(1.0, 1.0);
        glVertex3f(_right*w,  _top*h, 0.0f);
        
        glTexCoord2f(0.0, 1.0);
        glVertex3f(_right*w, _bot*h, 0.0f);
        
        glTexCoord2f(0.0, 0.0);
        glVertex3f(_left*w, _bot*h, 0.0f);
        glEnd();
    }
    else
    {
        glBegin(GL_QUADS);
        glTexCoord2f(0.0, 1.0);
        glVertex3f(_left*w, _top*h, 0.0f);
        
        glTexCoord2f(1.0, 1.0);
        glVertex3f(_right*w,  _top*h, 0.0f);
        
        glTexCoord2f(1.0, 0.0);
        glVertex3f(_right*w, _bot*h, 0.0f);
        
        glTexCoord2f(0.0, 0.0);
        glVertex3f(_left*w, _bot*h, 0.0f);
        glEnd();
    }
    
    glDisable(GL_TEXTURE_2D);
}

// Check to see if a coordinate is inside this rectangle
bool
UISoundRect::Pick(double x, double y)
{
    if ((x > _left) && (x < _right))
    {
        if ((y > _bot) && (y < _top))
            return true;
    }
    return false;
}

// Return normalized position values in x and y
bool
UISoundRect::GetNormedPosition(double *positionsX,
                               double *positionsY,
                               double x, double y,
                               unsigned int idx)
{
    bool trigger = false;
    if (Pick(x, y))
    {
        trigger = true;
        if (_orientation == HORIZONTAL)
        {
            positionsX[idx] = (double)((x - _left) / _width);
            positionsY[idx] = (double)((y - _bot) / _height);
        }
        else
        {
            positionsY[idx] = (double)((x - _left) / _width);
            positionsX[idx] = (double)((y - _bot) / _height);
        }
    }
    
    return trigger;
}

double
UISoundRect::GetX() const
{
    return _x;
}

double
UISoundRect::GetY() const
{
    return _y;
}

void
UISoundRect::SetName(const char *name)
{
    strcpy(_name, name);
}

void
UISoundRect::CreateSpectroTexture()
{
    Spectrogram *spectro = _rect->GetSpectrogram(0);
    
    spectro->NormalizeMagnsDisp();
    
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, (unsigned int *)&_spectroTexId);
    
    CreateTexture(spectro, _spectroTexId, _filterBank);

    glDisable(GL_TEXTURE_2D);
}

void
UISoundRect::CreateChromaTexture()
{
    Spectrogram *chroma = _rect->GetChromagram(0);
    chroma->NormalizeMagnsDisp();
    
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, (unsigned int *)&_chromaTexId);
    
    CreateTexture(chroma, _chromaTexId);

    glDisable(GL_TEXTURE_2D);
}

void
UISoundRect::CreateTexture(Spectrogram *spectro, int texId,
                           FilterBank *filterBank)
{
    int width = spectro->GetFFTSize()/2;
    int height = spectro->GetNumCols();
    const vector<SAMPLE> *data = spectro->GetMagnDataDisp();

    _tmpBuf.resize(width);
    _tmpBuf2.resize(width*height);
    for (int i = 0; i < height; i++)
    {
        memcpy(_tmpBuf.data(), &data->data()[i*width], width*sizeof(SAMPLE));

        if (filterBank != NULL)
        {
            filterBank->
                HzToTargetScale(&_tmpBuf, SRATE,
                                SRATE/width, SRATE*0.5, width);
        }
        
        memcpy(&_tmpBuf2.data()[i*width], _tmpBuf.data(), width*sizeof(SAMPLE));
    }
             
    // Convert the data
    unsigned char *texData = new unsigned char[width*height*4];
    for (int i = 0; i < width*height; i++)
    {
        double d = _tmpBuf2[i];
        
        d = Scale::AmpToDbNorm(d, MIN_DB_SP, MAX_DB_SP);

        d *= _colorRange;
        d = Scale::Sigmoid(d, _colorContrast);

        if (d > 1.0)
            d = 1.0;
        
        int c = d*255;
        int alpha = (d < MAX_SPECTRO_ALPHA) ?
            d*255 : MAX_SPECTRO_ALPHA*255;
        texData[i*4] = c;
        texData[i*4 + 1] = c;
        texData[i*4 + 2] = c;
        texData[i*4 + 3] = alpha;
    }

    glBindTexture(GL_TEXTURE_2D, texId);
    
    // TODO: keep mipmaps?
#if 0
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                 width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, texData);
#endif
    
#if 1
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                    GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                    GL_LINEAR_MIPMAP_LINEAR);	
    gluBuild2DMipmaps(GL_TEXTURE_2D, 4, width, height,
                      GL_RGBA, GL_UNSIGNED_BYTE, texData);
#endif

    delete []texData;
}
