/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef UI_SOUNDRECT_H
#define UI_SOUNDRECT_H

#include <vector>
using namespace std;

#ifdef _MSC_VER
#include <windows.h>
#endif

#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include <iostream>
#include <math.h>
#include <algorithm>

#include "Globals.h"
#include "ParameterProvider.h"

#include "Time.h"
#include "Stk.h"

using namespace std;

class SoundRect;
class UIStyle;
class UIScreen;
class App;
class Serializer;
class Spectrogram;
class FilterBank;
class WaveformPyramid;
class UISoundRect : public ParameterProvider
{ 
public:
    enum ParamIndex
    {
        NO_PARAMETER = -1,
        TYPE, // = 0,
        RANGE,
        CONTRAST,
        ORIENTATION,
        NUM_PARAMS
    };

    enum RectType
    {
        WAVEFORM = 0,
        SPECTROGRAM,
        CHROMAGRAM
    };

    enum Orientation
    {
        HORIZONTAL = 0,
        VERTICAL
    };
    
    UISoundRect(SoundRect *rect,
                App *app, UIStyle *style,
                UIScreen *screen);

    virtual ~UISoundRect();
    
    // Other object initialization code
    void Init();
    
    // Draw functions
    void Draw();
    
    // Properties
    void SetDimensions(double width, double height);

    double GetHeight();
    double GetWidth();
    
    // Process mouse drag
    void Move(double xDiff, double yDiff);

    bool Pick(double x, double y);
    
    // Set selection (highlight)
    bool GetSelectState() const;
    void SetSelectState(bool state);
    
    bool GetNormedPosition(double *positionsX,
                           double *positionsY,
                           double x, double y,
                           unsigned int idx);

    double GetX() const;
    double GetY() const;

    void UpdateRectPosition(double x, double y);
        
    // Change from vertical to horizontal
    void ToggleOrientation();
    
    // Set name
    void SetName(const char *name);

    void UpdateStyle(UIStyle *style);

    RectType GetRectType() const;
    void SetRectType(RectType type);

    double GetColorRange() const;
    void SetColorRange(double range);

    double GetColorContrast() const;
    void SetColorContrast(double contrast);

    Orientation GetOrientation() const;
    void SetOrientation(Orientation orientation);
                        
    void Serialize(Serializer *serializer);
    void UnSerialize(Serializer *serializer);

    // ParameterProvider
    Parameter *GetParameter(int paramIdx) override;
    Parameter *GetCurrentParameter() override;
    int GetNumParams() const override;
    double GetUIX() const override;
    double GetUIY() const override;
    double GetUIRadius() const override;

    void ResetCurrentParameterIndex();
    bool PickParameters(double x, double y);
        
protected:
    void SerializeParameters(Serializer *serializer);
    void UnSerializeParameters(Serializer *serializer);

    void InitParameters();

    void UpdatePyramid();
        
    void UpdateTextures();

    void DrawRectangle();
    void DrawName();
    
    void DrawWaveform();
    void DrawSpectro();
    void DrawChroma();
    void DrawSpectro(int texId);
    
    // Set upsampling for waveform display
    // (based on rect size/orientation)
    void SetupPyramid(vector<SAMPLE> *buf, int nFrames, int nChan);

    // Update information used for vertices with new width and height
    void UpdateCorners(double width, double height);

    void CreateSpectroTexture();
    void CreateChromaTexture();
    void CreateTexture(Spectrogram *spectro, int texId,
                       FilterBank *filterBank = NULL);

    SoundRect *_rect;

    double _width;
    double _height;

    double _top;
    double _bot;
    double _left;
    double _right;

    double _x;
    double _y;
    double _z;

    bool _isSelected;
    
    double _minDim;
    
    double _startTime;
    
    Orientation _orientation;
    
    App *_app;
    UIStyle *_style;
    UIScreen *_screen;
    
    char _name[255];

    RectType _rectType;
    double _colorRange;
    double _colorContrast;

    bool _needUpdatePyramid;
    int _spectroTexId;
    int _chromaTexId;
    bool _needUpdateTextures;
    
    FilterBank *_filterBank;
    vector<SAMPLE> _tmpBuf;
    vector<SAMPLE> _tmpBuf2;

    Parameter *_params[NUM_PARAMS];
    ParamIndex _currentParamIdx;

    WaveformPyramid *_pyramids[2];

    double _pulseTimeOffset;
};

#endif
