/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

// Graphics includes
#ifdef __MACOSX_CORE__
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif

#include "Time.h"

#include "UIScreen.h"
#include "UIStyle.h"
#include "UIWidget.h"
#include "UICanvas.h"
#include "UIWindow.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

#define TITLE_BOX_HEIGHT 0.0434
#define TITLE_X_MARGIN 0.0086

#define CONTENTS_Y_MARGIN 0.015

// Draw a box enclosing the title?
#define DRAW_TITLE_BOX_SEPARATOR 1 //0

UIWindow::UIWindow(UIScreen *screen,
                   UIStyle *style,
                   const char *title,
                   double x, double y,
                   double width, double height)
{
    _screen = screen;
    _style = style;

    _visible = true;
    
    sprintf(_title, title);
    
    _x = x;
    _y = y;
    
    _width = width;
    _height = height;
}

UIWindow::~UIWindow() {}

void
UIWindow::SetTitle(const char *title)
{
    sprintf(_title, title);
}

void
UIWindow::SetVisible(bool flag)
{
    _visible = flag;

    for (int i = 0; i < _canvas.size(); i++)
        _canvas[i]->SetVisible(flag);
}

void
UIWindow::Draw()
{
    if (!_visible)
        return;
    
    DrawBackground();
    DrawCanvas();
    DrawBorder();
    DrawTitle();
}

void
UIWindow::UpdateStyle(UIStyle *style)
{
    _style = style;

    for (int i = 0; i < _canvas.size(); i++)
        _canvas[i]->UpdateStyle(style);
}

void
UIWindow::GetContentsRect(double *x, double *y,
                          double *width, double *height)
{
    double titleHeight = TITLE_BOX_HEIGHT;
    
    *x = _x;
    *y = _y + titleHeight + CONTENTS_Y_MARGIN;
    *width = _width;
    *height = _height - titleHeight - CONTENTS_Y_MARGIN*2.0;
}

void
UIWindow::AddCanvas(UICanvas *canvas)
{
    _canvas.push_back(canvas);
}

void
UIWindow::DrawBackground()
{
    double corners[4][2];
    UIWidget::ComputeCorners(corners, _screen, _x, _y, _width, _height);

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
    
    double col[4];
    _style->GetUIWindowBackgroundColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);
    
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable
    
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    glBegin(GL_QUADS);
    glVertex3f(corners[0][0], corners[0][1], 0.0);
    glVertex3f(corners[1][0], corners[1][1], 0.0);
    glVertex3f(corners[2][0], corners[2][1], 0.0);
    glVertex3f(corners[3][0], corners[3][1], 0.0);
    glEnd();
    
    glPopMatrix();
}

void
UIWindow::DrawBorder()
{
    double corners[4][2];
    UIWidget::ComputeCorners(corners, _screen, _x, _y, _width, _height);

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
    
    double col[4];
    _style->GetUIWindowBorderColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);
    
    double lineWidth = _style->GetUIWindowBorderWidth();
    glLineWidth(lineWidth);
    
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable
    
    glBegin(GL_LINE_LOOP);
    glVertex3f(corners[0][0], corners[0][1], 0.0);
    glVertex3f(corners[1][0], corners[1][1], 0.0);
    glVertex3f(corners[2][0], corners[2][1], 0.0);
    glVertex3f(corners[3][0], corners[3][1], 0.0);
    glEnd();
  
    glPopMatrix();
}

void
UIWindow::DrawTitle()
{
    // Draw title label

    double a = 0.8f + 0.2*sin(0.5*M_PI*Time::Instance().GetSeconds());
    
    double col[4];
    _style->GetUIWindowTitleColor(col);
    glColor4f(col[0], col[1], col[2], a*col[3]);
    
    GLint len = strlen(_title);
    
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable

    glTranslatef(_x*_screen->GetWidth() +
                 TITLE_X_MARGIN*_screen->GetHeight(),
                 (1.0 - _y - TITLE_BOX_HEIGHT*0.75)*
                 _screen->GetHeight(),
                 0.0);
    glScalef(0.1, 0.1, 0.1);
    
    for(int i = 0; i < len; i++)
        glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, _title[i]);
    
    glPopMatrix();

#if DRAW_TITLE_BOX_SEPARATOR
    // Draw title box separator
    double lineWidth = _style->GetUIWindowTitleBoxBorderWidth();
    glLineWidth(lineWidth);
        
    glPushMatrix();

    glLoadIdentity(); // for non-zoomable
    
    glBegin(GL_LINES);
    glVertex3f(_x*_screen->GetWidth(),
               (1.0 - _y - TITLE_BOX_HEIGHT)*_screen->GetHeight(),
               0.0);
    glVertex3f((_x + _width)*_screen->GetWidth(),
               (1.0 - _y - TITLE_BOX_HEIGHT)*_screen->GetHeight(),
               0.0);
    glEnd();

    glPopMatrix();
#endif
}

void
UIWindow::DrawCanvas()
{
    for (int i = 0; i < _canvas.size(); i++)
        _canvas[i]->Draw();
}
