/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "Scale.h"
#include "Parameter.h"
#include "VolumeEffect.h"

VolumeEffect::VolumeEffect()
{
    _volume = 1.0;
    _mute = false;

    _params[ON_PARAM] = new ParameterEffectOn(NULL, ON_PARAM);
    _params[VOLUME_PARAM] = new ParameterVolumeEffectVolume(NULL, VOLUME_PARAM);
    _params[MUTE_PARAM] = new ParameterVolumeEffectMute(NULL, MUTE_PARAM);
}

VolumeEffect::~VolumeEffect()
{
    for (int i = 0; i < NUM_PARAMS; i++)
        delete _params[i];
}

void
VolumeEffect::Process(vector<SAMPLE> *buffer, int nFrames, int nChan)
{
    if (nChan < 2)
        return;

    UpdateParameters();

    double on = _params[ON_PARAM]->GetValue();
    if (on < 0.5)
        return;
    
    for (int i = 0; i < nFrames*nChan; i++)
    {
        double s = (*buffer)[i];
        s *= _volume;
        if (_mute)
            s = 0.0;
        (*buffer)[i] = s;
    }
}

void
VolumeEffect::SetVolume(double volumeDB)
{
    _volume = Scale::DbToAmp(volumeDB);
}

void
VolumeEffect::SetMute(bool mute)
{
    _mute = mute;
}

Parameter *
VolumeEffect::GetParameter(ParamIdx paramIdx)
{
    return _params[paramIdx];
}

void
VolumeEffect::Serialize(Serializer *serializer)
{
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->Serialize(serializer);
}

void
VolumeEffect::UnSerialize(Serializer *serializer)
{
    for (int i = 0; i < NUM_PARAMS; i++)
        _params[i]->UnSerialize(serializer);

    UpdateParameters();
}

void
VolumeEffect::UpdateParameters()
{
    double volumeDB = _params[VOLUME_PARAM]->GetValue();
    _volume = Scale::DbToAmp(volumeDB);
    
    _mute = _params[MUTE_PARAM]->GetValue();
}
