/* Copyright (C) 2022 Nicolas Dittlo <deadlab.plugins@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <math.h>

#include "WaveformPyramid.h"

#define VALUE_UNDEFINED -1e16

// Bufferize input and output for pushing and popping buffers
// of size power of two
#define PUSH_POP_POW_TWO 1

// Minimum samples to push
#define MIN_SAMPLES_PUSH 512

WaveformPyramid::WaveformPyramid(int maxLevel)
{
    _maxPyramidLevel = maxLevel;
    
    _waveformPyramid.resize(_maxPyramidLevel + 1);

    _remainToPop = 0;
    
    ResetTmpBuffers();
}

WaveformPyramid::~WaveformPyramid() {}

void
WaveformPyramid::Reset()
{
#if PUSH_POP_POW_TWO
    _pushBuf.clear();
    
    _remainToPop = 0;
#endif
    
    _waveformPyramid.clear();

    _waveformPyramid.resize(_maxPyramidLevel + 1);

    ResetTmpBuffers();
}

void
WaveformPyramid::SetValues(const vector<SAMPLE> &samples0)
{    
#if PUSH_POP_POW_TWO
    _pushBuf.clear();
    
    _remainToPop = 0;
#endif

    // Compute the size
    int pyrLevel = 0;
    int pyrLevelSize = samples0.size();
    while(true)
    {
        if (pyrLevelSize < 2)
            break;

        if (pyrLevel >= _maxPyramidLevel)
            break;
        
        pyrLevel++;
        
        pyrLevelSize /= 2;
    }

    vector<SAMPLE> samples = samples0;
        
    // Level0
    VectorToQueue(samples, &_waveformPyramid[0]);
    
    int pyramidLevel = 1;
    while(true)
    {        
        vector<SAMPLE> prevLevel;
        QueueToVector(_waveformPyramid[pyramidLevel - 1], &prevLevel);
        
        if (prevLevel.size() < 2)
            break;

        if (pyramidLevel > _maxPyramidLevel)
            break;
        
        vector<SAMPLE> newLevel;
        DecimateSamples(&newLevel, prevLevel, (SAMPLE)0.5);

        VectorToQueue(newLevel, &_waveformPyramid[pyramidLevel]);
        
        if (pyramidLevel >= _maxPyramidLevel)
            break;

        pyramidLevel++;
    }
}

long
WaveformPyramid::PushValues(const vector<SAMPLE> &inSamples)
{
    vector<SAMPLE> samples = inSamples;
    
#if PUSH_POP_POW_TWO
    // Bufferize pushed buffers to manage only power of 2 buffers
    _pushBuf.resize(_pushBuf.size() + samples.size());
    for (int i = _pushBuf.size() - samples.size();
             i < _pushBuf.size(); i++)
        _pushBuf[i] = samples[i - (_pushBuf.size() - samples.size())];
    
    int numToAdd = _pushBuf.size();
    numToAdd = NextPowerOfTwo(numToAdd);
    
    // Check if we were exactly power of two at the beginning
    // And divide only if were not
    if (numToAdd > _pushBuf.size())
        numToAdd /= 2;

    if (numToAdd < MIN_SAMPLES_PUSH)
        // We don't have enough buffered input samples 
        return 0;
    
    vector<SAMPLE> &samples0 = _tmpBuf6;
    QueueToVector(_pushBuf, 0, &samples0, numToAdd);

    _pushBuf.erase(_pushBuf.begin(), _pushBuf.begin() + numToAdd);
#else
    vector<SAMPLE> &samples0 = _tmpBuf6;
    samples0 = samples;
#endif

    _waveformPyramid[0].
        resize(_waveformPyramid[0].size() + samples0.size());
    for (int i = 0; i < samples0.size(); i++)
        _waveformPyramid[0][i + _waveformPyramid[0].size() -
                            samples0.size()] = samples0[i];
    
    int pyramidLevel = 0;
    
    int numSamplesOverlap = 0;
    int numSamplesAdd = samples0.size();
    
    while(true)
    {
        if (_waveformPyramid[pyramidLevel].size() < 2)
            // Top of the pyramid
            break;
        
        pyramidLevel++;

        if (pyramidLevel + 1 > _maxPyramidLevel)
            break;
        
        vector<SAMPLE> &currentLevel = _tmpBuf2[pyramidLevel];
        
        QueueToVector(_waveformPyramid[pyramidLevel - 1],
                      &currentLevel);
        
        // Take twice the size of the input buffer
        // for avoiding discontinuities
        int start = currentLevel.size() - numSamplesAdd -
            numSamplesOverlap;
        if (start < 0)
            start = 0;
        int end = currentLevel.size();
        int size = end - start;

        vector<SAMPLE> &samplesCurrentLevel =
            _tmpBuf3[pyramidLevel];

        int numToCopy = size;
        if (numToCopy + start > currentLevel.size())
            numToCopy = currentLevel.size() - start;
        if (numToCopy > 0)
        {
            samplesCurrentLevel.resize(numToCopy);
            memcpy(samplesCurrentLevel.data(),
                   &currentLevel.data()[0], numToCopy*sizeof(SAMPLE));
        }
        
        vector<SAMPLE> &samplesNextLevel = _tmpBuf4[pyramidLevel];
        DecimateSamples(&samplesNextLevel,
                              samplesCurrentLevel, (SAMPLE)0.5);
                                       
        numSamplesOverlap /= 2;
        
        // Take only the second half of the buffer,
        // because we previously took twice the buffer size.
        vector<SAMPLE> &samplesNextLevel0 = _tmpBuf5[pyramidLevel];

        samplesNextLevel0.resize(samplesNextLevel.size() -
                                 numSamplesOverlap);
        memcpy(samplesNextLevel0.data(),
               &samplesNextLevel.data()[numSamplesOverlap],
               (samplesNextLevel.size() - numSamplesOverlap)*
               sizeof(SAMPLE));

        _waveformPyramid[pyramidLevel].
            resize(_waveformPyramid[pyramidLevel].size() +
                   samplesNextLevel0.size());
        for (int i = 0; i < samplesNextLevel0.size(); i++)
            _waveformPyramid[pyramidLevel]
                [i + _waveformPyramid[pyramidLevel].size() -
                 samplesNextLevel0.size()] = 
                samplesNextLevel0[i];
                                          
        numSamplesAdd /= 2;
        
        if (pyramidLevel > _maxPyramidLevel)
            break;
    }

    return numToAdd;
}

void
WaveformPyramid::PopValues(long numSamples)
{
#if PUSH_POP_POW_TWO
    // Because we bufferize when pushing, we must
    // bufferize when poping too, to avoid shifting
    // (Otherwise in Ghost-X view mode as plugins, the waveform
    // will not scroll correclly).
    long numSamples0 = numSamples + _remainToPop;
    
    numSamples0 = NextPowerOfTwo((int)numSamples0);
    
    // Check if we were exactly power of 2 at the beginning
    // and divide only if were not
    if (numSamples0 > numSamples + _remainToPop)
        numSamples0 /= 2;
    
    if (_waveformPyramid.empty() ||
        (_waveformPyramid[0].size() < numSamples0))
    {
        _remainToPop += numSamples;
        
        return;
    }
    
    _remainToPop += numSamples - numSamples0;
    
    numSamples = numSamples0;
#endif
    
    SAMPLE numSamplesD = numSamples;
    for (int i = 0; i < _waveformPyramid.size(); i++)
    {
        // Consume left
        deque<SAMPLE> copy = _waveformPyramid[i];
        _waveformPyramid[i].resize(_waveformPyramid[i].size()
                                   - numSamplesD);
        for (int k = 0; k < copy.size() - numSamplesD; k++)
            _waveformPyramid[i][k] = copy[k + numSamplesD];
            
        numSamplesD *= 0.5;
        
        if (_waveformPyramid[i].size() < 2)            
            break;
    }
}

void
WaveformPyramid::ReplaceValues(long start,
                               const vector<SAMPLE> &samples)
{
    int pyramidLevel = 0;
    
    vector<SAMPLE> samples0 = samples;
    
    SAMPLE startD = start;
    
    // Iterate over the pyramid levels and replace
    while(true)
    {
        if (pyramidLevel >= _waveformPyramid.size())
            break;
        
        // Replace
        int numToAssign = samples0.size();
        if (startD + numToAssign >
            _waveformPyramid[pyramidLevel].size())
            numToAssign =
                _waveformPyramid[pyramidLevel].size() - startD;
        for (int k = 0; k < numToAssign; k++)
            _waveformPyramid[pyramidLevel][k] = samples0[k + startD];
        
        vector<SAMPLE> tmp = samples0;
        DecimateSamples(&samples0, tmp, (SAMPLE)0.5);
        
        startD /= 2.0;
        pyramidLevel++;
    }
}

void
WaveformPyramid::GetValues(SAMPLE start, SAMPLE end, long numValues,
                           vector<SAMPLE> *samples)

{
    if (_waveformPyramid.empty())
        return;
    
    start *= _waveformPyramid[0].size();
    end *= _waveformPyramid[0].size();
        
    // First, check the bounds and possibly return early
    if (!_waveformPyramid.empty())
    {
        int numTotalSamples = _waveformPyramid[0].size();
            
        // Check if we are totally out of bounds
        if (((start <= 0.0) && (end <= 0.0)) ||
            ((start >= numTotalSamples) && (end >= numTotalSamples)))
            // Everything out of view
        {
            // Simply return a buffer of zeros
            samples->resize(numValues);
            memset(samples->data(), 0, samples->size()*sizeof(SAMPLE));
            
            return;
        }
    }
    
    // Check if we must add zeros at the beginning
    SAMPLE numZerosBegin = 0.0;
    if (start < 0)
    {
        numZerosBegin = -start;
        start = 0;
    }
    
    // Check if we must add zeros at the end
    SAMPLE numZerosEnd = 0.0;
    if (!_waveformPyramid.empty() &&
        (end > _waveformPyramid[0].size()))
    {
        numZerosEnd = end - _waveformPyramid[0].size();
        end = _waveformPyramid[0].size();
    }

    // Fixes negative size...
    if (end - start < 0)
    {
        samples->resize(0);
        return;
    }
    
    // Get the data
    int pyramidLevel = 0;
    
    // Find the correct pyramid level
    while(true)
    {
        long size = end - start;
        
        // We must start at the first bigger interval
        if ((pyramidLevel >= _waveformPyramid.size()) ||
            (size <= numValues))
            break;
        
        start /= 2;
        end /= 2;
        
        numZerosBegin /= 2;
        numZerosEnd /= 2;
        
        pyramidLevel++;

        if (pyramidLevel >= _maxPyramidLevel)
            break;
    }

    int size = ceil(end - start) + 1;
    
    vector<SAMPLE> &level = _tmpBuf9;
    QueueToVector(_waveformPyramid[pyramidLevel], start, &level, size);
    
    // Add zeros if necessary
    if (numZerosBegin > 0)
    {
        PadZerosLeft(&level, (int)numZerosBegin);
    }
    
    if (numZerosEnd > 0)
    {
        PadZerosRight(&level, (int)numZerosEnd);
    }
    
    // Update the size with the zeros added
    size = level.size();

    // Decimate
    SAMPLE decimFactor = ((SAMPLE)numValues)/size;

    // Final step decimate only if the level size
    // is significantly bigger
    vector<SAMPLE> samples0;
    samples0.resize(samples->size());
    if (level.size() > numValues*2)
    {
        DecimateSamples(&samples0, level, decimFactor);
    }
    else
    {
        if (level.size() > numValues*0.5)
        {    
            // We don't have too many samples at the and => simple copy
            samples0 = level;
        }
        else
        {
            // We have too few samples
            // Must increase the number of samples,
            // and adjust very accurately to double start and end

            // Take the two border values
            SAMPLE start0 = floor(start);
            SAMPLE end0 = ceil(end);
            SAMPLE size0 = end0 - start0;

            // Extract again, with the addtion to left and right values
            vector<SAMPLE> &level0 = _tmpBuf10;
            QueueToVector(_waveformPyramid[pyramidLevel],
                          start0,
                          &level0, size0);

            // Must pad again because we just extracted the level
            if (numZerosBegin > 0)
            {
                PadZerosLeft(&level0, (int)numZerosBegin);
            }
            
            if (numZerosEnd > 0)
            {
                PadZerosRight(&level0, (int)numZerosEnd);
            }
            
            SAMPLE leftT = start - start0;
            SAMPLE rightT = end0 - end;

            // Adjust with zeros positions if necessary
            leftT -= (numZerosBegin - (int)numZerosBegin);
            rightT -= (numZerosEnd - (int)numZerosEnd);
            
            // Upsample
            UpsampleResult(numValues, &samples0, level0, leftT, rightT);
        }
    }

    *samples = samples0;
}

void
WaveformPyramid::ResetTmpBuffers()
{
    _tmpBuf2.resize(0);
    _tmpBuf2.resize(_maxPyramidLevel + 1);
    
    _tmpBuf3.resize(0);
    _tmpBuf3.resize(_maxPyramidLevel + 1);

    _tmpBuf4.resize(0);
    _tmpBuf4.resize(_maxPyramidLevel + 1);

    _tmpBuf5.resize(0);
    _tmpBuf5.resize(_maxPyramidLevel + 1);

    _tmpBuf7.resize(0);
    _tmpBuf7.resize(_maxPyramidLevel + 1);
    
    _tmpBuf8.resize(0);
    _tmpBuf8.resize(_maxPyramidLevel + 1);
}

void
WaveformPyramid::UpsampleResult(int numValues,
                                vector<SAMPLE> *result,
                                const vector<SAMPLE> &vec,
                                SAMPLE leftT, SAMPLE rightT)
{
    if (vec.size() < 3)
        return;
    
    // Init with undefined
    result->resize(numValues);
    for (int i = 0; i < result->size(); i++)
        (*result)[i] = VALUE_UNDEFINED;

    // Left value
    SAMPLE leftVal = (1.0 - leftT)*vec.data()[0] +
        leftT*vec.data()[1];
    result->data()[0] = leftVal;
    
    // Right value
    SAMPLE rightVal =
        rightT*vec.data()[vec.size() - 2] +
        (1.0 - rightT)*vec.data()[vec.size() - 1];
    result->data()[result->size() - 1] = rightVal;
    
    // Intermediate values
    for (int i = 1; i < vec.size() - 1; i++)
    {
        SAMPLE val = vec.data()[i];

        SAMPLE t = ((SAMPLE)i - leftT)/(vec.size() - 1 -
                                        (leftT + rightT));
        
        int targetIdx = t*result->size();
        if ((targetIdx > 0) && (targetIdx < result->size() - 1))
            result->data()[targetIdx] = val;
    }
    
    bool extendBounds = false;
    Interpolate(result, extendBounds, (SAMPLE)VALUE_UNDEFINED);
}

void
WaveformPyramid::DecimateSamples(vector<SAMPLE> *result,
                                 const vector<SAMPLE> &vec,
                                 SAMPLE decFactor)
{
    if (vec.size() == 0)
        return;
    
    if (decFactor >= 1.0)
    {
        *result = vec;
        
        return;
    }

    // Adjust the result size to ceil(), and adjust decFactor,
    // in case of odd input buffer
    int newSize = ceil(vec.size()*decFactor);
    // Adjust decimation factor
    decFactor = ((SAMPLE)newSize)/vec.size();

    int prevSize = result->size();
    result->resize(newSize);
    if (newSize - prevSize > 0)
        memset(&result->data()[prevSize], 0,
               (newSize - prevSize)*sizeof(SAMPLE));

    int vecSize = vec.size();
    const SAMPLE *vecData = vec.data();
    int resultSize = result->size();
    SAMPLE *resultData = result->data();
    
    int resultIdx = 0;
    
    // Keep the maxima when decimating
    SAMPLE count = 0.0;
    
    SAMPLE maxSample = vec.data()[0];
    SAMPLE prevSampleUsed = vec.data()[0];
        
    for (int i = 0; i < vecSize; i++)
    {
        SAMPLE samp = vecData[i];

        // Take the maximum spacing (whatever the sign)
        if (fabs(samp - prevSampleUsed) >
            fabs(maxSample - prevSampleUsed))
            maxSample = samp;
        
        count += decFactor;

        // In case buf size is odd, and result size is even,
        // be sure to manage the last buf value
        if ((count >= 1.0) || // Enough src samples visited
            (i == vecSize - 1)) // Last src sample
            {
                resultData[resultIdx] = maxSample;

                prevSampleUsed = maxSample;

                // Be sure to keep the first sample value
                // e.g: Series of 10 impulses, if first sample
                // is an impulse peak

                if (resultIdx == 0)
                {
                    // Be sure that the first sample will be the same
                    // in th origin buf and in the decimated buf
                    resultData[resultIdx] = vec.data()[0];
                }

                resultIdx++;
            
                count -= 1.0;

                prevSampleUsed = maxSample;
            }

        if (resultIdx >=  resultSize)
            break;
    }
}

int
WaveformPyramid::NextPowerOfTwo(int value)
{
    int result = 1;
    
    while(result < value)
        result *= 2;
    
    return result;
}

void
WaveformPyramid::VectorToQueue(const vector<SAMPLE> &vec,
                               deque<SAMPLE> *q)
{
    int numToAdd = vec.size() - q->size();
    if (numToAdd > 0)
        q->insert(q->end(), numToAdd, 0.0);
    else if (numToAdd < 0)
    {
        q->clear();
        q->insert(q->begin(), vec.size(), 0.0);
    }

    for (int k = 0; k < vec.size(); k++)
        (*q)[k] = vec[k];
}

void
WaveformPyramid::QueueToVector(const deque<SAMPLE> &q,
                               vector<SAMPLE> *vec)
{
    vec->resize(q.size());
    for (int k = 0; k < vec->size(); k++)
        (*vec)[k] = q[k];
}

void
WaveformPyramid::QueueToVector(const deque<SAMPLE> &q, int queueOffset,
                               vector<SAMPLE> *vec, int numToCopy)
{
    if (numToCopy == -1)
        numToCopy = (q.size() - queueOffset);

    if (numToCopy + queueOffset > q.size())
        numToCopy = q.size() - queueOffset;
    
    vec->resize(numToCopy);

    for (int k = 0; k < numToCopy; k++)
        (*vec)[k] = q[k + queueOffset];
}

void
WaveformPyramid::PadZerosLeft(vector<SAMPLE> *vec, int size)
{
    vector<SAMPLE> copy = *vec;
    vec->resize(vec->size() + size);
    memset(vec->data(), 0, size*sizeof(SAMPLE));

    for (int k = 0; k < copy.size(); k++)
        (*vec)[k] = copy[k];
}

void
WaveformPyramid::PadZerosRight(vector<SAMPLE> *vec, int size)
{
    vec->resize(vec->size() + size);
    memset(&vec->data()[vec->size() - size], 0, size*sizeof(SAMPLE));
}

void
WaveformPyramid::Interpolate(vector<SAMPLE> *values,
                             bool extendBounds,
                             SAMPLE undefinedValue)
{
    if (extendBounds)
        // Extend the last value to the end
    {
        // Find the last max
        int lastIndex = values->size() - 1;
        SAMPLE lastValue = 0.0;
        
        int valuesSize = values->size();
        SAMPLE *valuesData = values->data();
        
        for (int i = valuesSize - 1; i > 0; i--)
        {
            SAMPLE val = valuesData[i];
            if (val > undefinedValue)
            {
                lastValue = val;
                lastIndex = i;
                
                break;
            }
        }
        
        // Fill the last values with last max
        for (int i = valuesSize - 1; i > lastIndex; i--)
        {
            valuesData[i] = lastValue;
        }
    }
    
    // Fill the holes by linear interpolation
    SAMPLE startVal = 0.0;
    
    // First, find start val
    
    int valuesSize = values->size();
    SAMPLE *valuesData = values->data();
    
    for (int i = 0; i < valuesSize; i++)
    {
        SAMPLE val = valuesData[i];
        if (val > undefinedValue)
        {
            startVal = val;
        }
    }
    
    int loopIdx = 0;
    int startIndex = 0;
    
    // Then start the main loop
    while(loopIdx < valuesSize)
    {
        SAMPLE val = valuesData[loopIdx];
        
        if (val > undefinedValue)
            // Defined
        {
            startVal = val;
            startIndex = loopIdx;
            
            loopIdx++;
        }
        else
            // Undefined
        {
            // Start at 0
            if (!extendBounds &&
                (loopIdx == 0))
                startVal = 0.0;
            
            // Find how many missing values we have
            int endIndex = startIndex + 1;
            SAMPLE endVal = 0.0;
            bool defined = false;
            
            while(endIndex < valuesSize)
            {
                if (endIndex < valuesSize)
                    endVal = valuesData[endIndex];
                
                defined = (endVal > undefinedValue);
                if (defined)
                    break;
                
                endIndex++;
            }
            
            // Fill the missing values with lerp
            for (int i = startIndex; i < endIndex; i++)
            {
                SAMPLE t =
                    ((SAMPLE)(i - startIndex))/(endIndex - startIndex);
                
                SAMPLE newVal = (1.0 - t)*startVal + t*endVal;
                
                valuesData[i] = newVal;
            }
            
            startIndex = endIndex;
            loopIdx = endIndex;
        }
    }
}
